#include <antiglut.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "mvlib.h"

/*
 * Uncomment this if you use software rasterizer (or 'HW' one from ATI under *NIX ... SIC!).
 * Or, better, go to the nearest computer shop and upgrade.
 * ... actually, it was: #ifndef _HAVE_NVIDIA_ ...
 * 18.05.2009: _WHY_DONT_U_HAVE_NVDA_ ???
 * and still... hmm... under all of *nix i've seen: ATI in great DUPE... hahaha... !!!
#define	AA_DISABLED
 */

/*
 * Following define are meaningful only if !AA_DISABLED
 */
#define	SRC_ALPHA		0.4f

/*
 * -----------------------------------------------------------------------
 * Part 1: physical model
 */

#define MON_TIME        (vec_t)(1000)
#define SAVE_INTERVAL   (vec_t)(1.0)

#define STEP            (vec_t)(0.0000005)
#define	AUTOPAUSE		0

#define MAX_POINT_SIZE	10
#define MIN_POINT_SIZE	3

#define NUM_OBJECTS     4

typedef struct ph_obj_s
{
	vec3_t  pos;
	vec3_t  vel;
	vec3_t  force; /* force acting on this object */
	vec_t   mass, rm; /* gravital mass and reciprocal inertial mass */
	vec_t	draw_size; /* point size; computed on the base of mass */
}   ph_obj_t;

static ph_obj_t    obj[NUM_OBJECTS];

/*
 * F = r12 * m1 * m2 / abs(r12) ^ 3
 */

static void
gravity_force(ph_obj_t *subj,ph_obj_t *obj,vec3_t force_to_subj)
{
	vec3_t dr;
	vec_t rrq, f;
	vec3Sub(dr,obj->pos,subj->pos);
	rrq = 1/vec3Dot(dr,dr);
	vec3Scale(dr,dr,sqrt(rrq));
	f = /* GRAVITY_CONST * */ subj->mass * obj->mass * rrq;
	vec3Scale(force_to_subj,dr,f);
}

/*
 * For each pairs of objects, compute force and accumulate it
 */
static void
compute_forces(void)
{
	vec3_t force;
	int i, j, lastobj;
	for(i = 0;i < NUM_OBJECTS;i++)
		vec3Zero(obj[i].force);
	lastobj = NUM_OBJECTS - 1;
	for(i = 0;i < lastobj;i++)
	{
		for(j = i + 1;j < NUM_OBJECTS;j++)
		{
			gravity_force(&obj[i],&obj[j],force);
			vec3Add(obj[i].force,obj[i].force,force);
			vec3Sub(obj[j].force,obj[j].force,force);
		}
	}
}

/*
 * Monitoring: computes full energy of system.
 */
static vec_t
full_energy(void)
{
	int i, j;
	vec_t vel_squared, r, potential, energy = 0;
	vec3_t r_ji;
	/* Kinetic energy: summ_i(m_i * v_i ^ 2) / 2 */
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		vel_squared = vec3Dot(obj[i].vel,obj[i].vel);
		energy += obj[i].mass * vel_squared;
	}
	/* Potential energy */
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		potential = 0;
		for(j = 0;j<NUM_OBJECTS;j++)
		{
			if(i==j)
				continue;
			vec3Sub(r_ji,obj[j].pos,obj[i].pos);
			r = vec3Abs(r_ji);
			potential += obj[j].mass / r;
		}
		energy -= potential * obj[i].mass;
	}
	energy *= (vec_t)0.5;
	return energy;
}

/*
 * -----------------------------------------------------------------------
 * Part 2: intial conditons
 */

static vec_t   initial_energy;

static void
initial_conditions_postprocess(void)
{
	int i;
	vec_t m_min = MAXVEC, m_max = -MAXVEC, m_range, m_scale;
	/* setup reciprocal inertial masses */
	for(i = 0;i < NUM_OBJECTS;i++)
		obj[i].rm = 1.0 / obj[i].mass;
	/* setup visible size of objects */
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		if(m_max < obj[i].mass)
			m_max = obj[i].mass;
		if(m_min > obj[i].mass)
			m_min = obj[i].mass;
	}
	m_range = m_max - m_min;
	if(m_range <= FLOAT_ZERO_EPSILON)
		m_scale = (MAX_POINT_SIZE - MIN_POINT_SIZE)/2.0;
	else
		m_scale = (MAX_POINT_SIZE - MIN_POINT_SIZE)/m_range;
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		obj[i].draw_size = MIN_POINT_SIZE + obj[i].mass * m_scale;
	}
	/* compute Hamiltonian for the initial state of system */
	initial_energy = full_energy();
}

/*
 * Be prepared!
 * Almost all of U can see defined HERE!
 */
 
static void
initial_conditions(void)
{
	int cur_obj = 0;
#define NEXT_OBJ	cur_obj++
#define	MASS(m)		obj[cur_obj].mass = m
#define	POS(x,y,z)	vec3Set(obj[cur_obj].pos,x,y,z)
#define	VEL(x,y,z)	vec3Set(obj[cur_obj].vel,x,y,z)


/*
 * Initial values for masses, positions and velocities.
 * The only place to be changed.
 */

	MASS(1);
	POS(1,0,0);
	VEL(0,1,0.11);
	NEXT_OBJ;

	MASS(1);
	POS(0,1,0);
	VEL(-1,0,0.01);
	NEXT_OBJ;

	MASS(1);
	POS(-1,0,0);
	VEL(0,-1,-0.1);
	NEXT_OBJ;

	MASS(1);
	POS(0,-1,0);
	VEL(1,0,-0.01);
	NEXT_OBJ;

	/* assert(cur_obj >= NUM_OBJECTS); */
/*
	MASS(.1);
	POS(0,0,0);
	VEL(0,0,1);
*/

#undef	NEXT_OBJ
#undef	MASS
#undef	POS
#undef	VEL

	initial_conditions_postprocess();
}

/*
 * -----------------------------------------------------------------------
 * Part 3: SODE solver
 * Dumb implementation of Stoermer's rule on non-adaptive time-domain grid.
 */

static vec_t   simulation_time = 0;
static unsigned steps_total = 0;

static vec3_t  prev_r[NUM_OBJECTS];

static void
stoer_first(vec_t timediff)
{
	int i;
	vec3_t v;
	compute_forces();
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		vec3Scale(v,obj[i].force,obj[i].rm * timediff / 2);
		vec3Add(obj[i].vel,obj[i].vel,v);
		vec3Scale(v,obj[i].vel,timediff);
		vec3Copy(prev_r[i],obj[i].pos);
		vec3Add(obj[i].pos,obj[i].pos,v);
	}
	simulation_time += timediff;
	steps_total++;
}

/*
 * The kernel...
 */
static void
stoer_next(vec_t timediff)
{
	int i;
	vec3_t v, v1;
	compute_forces();
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		/* note: we need velocities for consistency check only ! */
		vec3Sub(v,obj[i].pos,prev_r[i]);
		vec3Scale(v,v,1/timediff);
		vec3Scale(v1,obj[i].force,obj[i].rm * timediff * (vec_t)0.5);
		vec3Add(obj[i].vel,v,v1);

		/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
		/* 3-layer Adams-like SODE solver */
		vec3Scale(v,obj[i].force,obj[i].rm * timediff * timediff);
		vec3Scale(v1,obj[i].pos,2);
		vec3Sub(v1,v1,prev_r[i]);
		vec3Add(v1,v1,v);
		vec3Copy(prev_r[i],obj[i].pos); /* store curlayer as prev one */
		vec3Copy(obj[i].pos,v1);
		/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
		/* ...that is all, falks! ;*) */


	}
	simulation_time += timediff;
	steps_total++;
}

/*
 * -----------------------------------------------------------------------
 * Part 4: Monitoring
 */

static vec_t max_diff = 0;
static vec_t energy, diff;

static void
step_status()
{
	energy = full_energy();
	diff = energy - initial_energy;
	if(max_diff < fabs(diff))
		max_diff = fabs(diff);
}

/*
 * -----------------------------------------------------------------------
 * Part 5: drawing stuff
 */

static void
draw_xOy_grid(vec3_t origin,vec_t width,vec_t height,unsigned num_lines)
{
	unsigned i;
	vec3_t p, dp;
	vec3Copy(p,origin);
	p[0] -= (vec_t)0.5 * width;
	p[1] -= (vec_t)0.5 * height;
	vec3Set(dp,width,height,0);
	vec3Scale(dp,dp,(vec_t)1 / (num_lines-1));
	glBegin(GL_LINES);
		/* grid along Oy direction */
		for(i=0;i<num_lines;i++)
		{
			glVertex3d(p[0] + i * dp[0],p[1],p[2]);
			glVertex3d(p[0] + i * dp[0],p[1] + height,p[2]);
		}
		/* grid along Ox direction */
		for(i=0;i<num_lines;i++)
		{
			glVertex3d(p[0],p[1] + i * dp[1],p[2]);
			glVertex3d(p[0] + width,p[1] + i * dp[1],p[2]);
		}
	glEnd(); /* GL_LINES */
}

static void
draw_2_xOy_grids(vec3_t origin,vec3_t center,vec_t width,vec_t height,unsigned major_lines,unsigned minor_lines)
{
	vec_t dw, dh;
	vec3_t neworg;
	if(minor_lines < 2 || major_lines < 2)
		return;
	dw = width / (major_lines - 1);
	dh = height / (major_lines - 1);
	vec3Copy(neworg,center);
	neworg[0] -= (vec_t)fmod(center[0] - origin[0],dw);
	neworg[1] -= (vec_t)fmod(center[1] - origin[1],dh);
	glLineWidth(1.5f);
	glColor4f(0,0.4f,0,SRC_ALPHA);
	draw_xOy_grid(neworg,width,height,minor_lines);
	glLineWidth(3.5f);
	glColor4f(0.2f,0.2f,0,SRC_ALPHA);
	draw_xOy_grid(neworg,width,height,major_lines);
}

static GLfloat alpha = 0, beta = 0;
static GLfloat deltax = 0, deltay = 0;
static GLfloat fmin_ = 1 * 4.4f, fcenter = 15 * 4.4f, fmax_ = 29 * 4.4f, ext_scale = 1 * 4 * 1.15f;
static GLboolean aw = GL_FALSE;

static void
draw(void)
{
	int i;
	vec3_t mins, maxs, origin, center, v;
	vec_t scale, se;

	/* Compute bounding box */
	bboxClear(mins,maxs);
	for(i = 0;i < NUM_OBJECTS;i++)
		bboxAddPoint(mins,maxs,obj[i].pos);

	/* translate to center of all objects */
	vec3Zero(origin);
	vec3Sub(center,mins,maxs);
	scale = 4.0f / vec3Abs(center);
	vec3Add(center,mins,maxs);
	vec3Scale(center,center,0.5f);

	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	glTranslatef(0,0,-fcenter);
	se = scale * ext_scale; /* preserve 'scale' */
	glScaled(se,se,se); 

	glTranslatef(deltax,deltay,0);
	glRotatef(alpha,0,1,0);
	glRotatef(beta,1,0,0);

	glTranslated(-center[0],-center[1],-center[2]);
	/* Draw coordsys-aligned grid on xOy plane */
	draw_2_xOy_grids(origin,center,8,8,5,25);
	glColor4f(0.5f,0,0,SRC_ALPHA);
	/* Origin */
	glBegin(GL_POINTS);
		glVertex3f(0,0,0);
	glEnd();
	/* Draw masses */
	glColor4f(0,0,0.5f,SRC_ALPHA);
	/* We can't just loop over all points; the points' size must be changed */
	for(i = 0;i < NUM_OBJECTS;i++)
	{
		glPointSize((GLfloat)obj[i].draw_size);
		glBegin(GL_POINTS);
			glVertex3d(obj[i].pos[0],obj[i].pos[1],obj[i].pos[2]);
		glEnd();
	}
	/* Draw ticks */
	glColor4f(0.3f,0,0,SRC_ALPHA);
	glLineWidth(1);
	glBegin(GL_LINES);
		for(i = 0;i < NUM_OBJECTS;i++)
		{
			glVertex3d(obj[i].pos[0],obj[i].pos[1],obj[i].pos[2]);
			glVertex3d(obj[i].pos[0],obj[i].pos[1],0);
		}
	glEnd();
	/* Draw velocities */
	glColor4f(0,0.2f,0.2f,SRC_ALPHA);
	glLineWidth(1.5f);
	glBegin(GL_LINES);
		for(i = 0;i < NUM_OBJECTS;i++)
		{
			vec_t va = vec3Abs(obj[i].vel);
			if(va >= FLOAT_ZERO_EPSILON)
			{
				va = 0.2f / (va * scale);
				glVertex3d(obj[i].pos[0],obj[i].pos[1],obj[i].pos[2]);
				vec3Scale(v,obj[i].vel,va);
				vec3Add(v,v,obj[i].pos);
				glVertex3d(v[0],v[1],v[2]);
			}
		}
	glEnd();
	/* Draw endpoints of velocities (fixme - accum prev tests?) */
	glPointSize(4);
	glBegin(GL_POINTS);
		for(i = 0;i < NUM_OBJECTS;i++)
		{
			vec_t va = vec3Abs(obj[i].vel);
			if(va >= FLOAT_ZERO_EPSILON)
			{
				va = 0.2f / (va * scale);
				vec3Scale(v,obj[i].vel,va);
				vec3Add(v,v,obj[i].pos);
				glVertex3d(v[0],v[1],v[2]);
			}
		}
	glEnd();
	/* Draw accels */
	glColor4f(0.5f,0,0,SRC_ALPHA);
	glLineWidth(1.2f);
	glBegin(GL_LINES);
		for(i = 0;i < NUM_OBJECTS;i++)
		{
			vec_t va = vec3Abs(obj[i].force);
			if(va >= FLOAT_ZERO_EPSILON)
			{
				va = 0.3f / (va * scale);
				glVertex3d(obj[i].pos[0],obj[i].pos[1],obj[i].pos[2]);
				vec3Scale(v,obj[i].force,va);
				vec3Add(v,v,obj[i].pos);
				glVertex3d(v[0],v[1],v[2]);
			}
		}
	glEnd();
	agSwap();
}

/*
 * -----------------------------------------------------------------------
 * Part 6: Camera Control
 */
static int paused = 0;
static int autopause = 0;

static void
correct_frustum(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1,1,-1,1,fmin_,fmax_);
	glMatrixMode(GL_MODELVIEW);
}

static int
handle_event_check_break(ui_event_p in)
{
	switch(in->type)
	{
		case ui_key_event:
			switch(in->key.scan)
			{
				case sc_escape:
					return 1;
				case sc_wheelup:
				case sc_minus:
				case sc_kp_subtract:
					fmin_ *= 0.9f;
					fmax_ *= 0.9f;
					fcenter = (fmin_ + fmax_) * 0.5f;
					ext_scale *= 0.85f;
					correct_frustum();
					break;
				case sc_wheeldown:
				case sc_equal:
				case sc_kp_add:
					fmin_ *= 1.1f;
					fmax_ *= 1.1f;
					ext_scale *= 1.15f;
					fcenter = (fmin_ + fmax_) * 0.5f;
					correct_frustum();
					break;
				case sc_space:
				case sc_kp_begin:
					if(in->key.press)
					{
						alpha = beta = deltax = deltay = 0;
					}
					break;
				case sc_up:
					deltay -= 0.1f;
					break;
				case sc_kp_up:
					beta += 3;
					break;
				case sc_down:
					deltay += 0.1f;
					break;
				case sc_kp_down:
					beta -= 3;
					break;
				case sc_left:
					deltax += 0.1f;
					break;
				case sc_kp_left:
					alpha -= 3;
					break;
				case sc_right:
					deltax -= 0.1f;
					break;
				case sc_kp_right:
					alpha += 3;
					break;
				case sc_w:
					if(in->key.press)
						uiAutoWarp(aw = (GLboolean)(aw ? GL_FALSE : GL_TRUE));
					break;
				case sc_a:
					if(in->key.press)
						autopause = !autopause;
					break;
				case sc_p:
					if(in->key.press)
						paused = !paused;
					break;
				case sc_s:
					step_status();
					printf("Time %4.3g\tHamiltonian %g\n",simulation_time,energy);
					break;
			}
			break;
		case ui_wm_event:
			switch(in->configure.wm_event)
			{
				case ui_wm_close:
					return 1;
			}
			break;
		case ui_pointer_motion_event:
			if((in->button & UI_LBUTTON_MASK) || 
				(aw && !(in->button & UI_RBUTTON_MASK)))
			{
				alpha += in->client_dx;
				beta += in->client_dy;
			}
			if(in->button & UI_RBUTTON_MASK)
			{
				deltax += 0.05f * in->client_dx / ext_scale;
				deltay -= 0.05f * in->client_dy / ext_scale;
			}
			break;
		default:
			break;
	}
	return 0;
}

static void
print_status()
{
	int i;
	static vec_t stime = 0;
	if(simulation_time - stime > SAVE_INTERVAL)
	{
		stime = simulation_time;
		step_status();
		printf("Time %4.3g\tHamiltonian %g\nPositions:\n",simulation_time,energy);
		for(i = 0;i < NUM_OBJECTS;i++)
		{
			printf("Particle #%i\t(%4.3g;%4.3g;%4.3g)\n",i+1,
				obj[i].pos[0],obj[i].pos[1],obj[i].pos[2]);
		}
		puts("Velocities:");
		for(i = 0;i < NUM_OBJECTS;i++)
		{
			printf("Particle #%i\t(%4.3g;%4.3g;%4.3g)\n",i+1,
				obj[i].vel[0],obj[i].vel[1],obj[i].vel[2]);
		}
		if(autopause && !paused)
			paused = !paused;
		puts("");
	}
}

static	vec_t   timelimit = MON_TIME;

static void
main_loop(void)
{
	static unsigned computation_time = 20; /* 50 FPS max */
	static unsigned lastdraw_time = 0;
	int i;
	ui_event_t in;
	stoer_first(STEP);
	do
	{
		while(uiPending())
		{
			uiNextEvent(&in);
			if(handle_event_check_break(&in))
				goto good_bye;
		}
		if(!paused)
		{
			for(i = 0;i < 10;i++)
				stoer_next(STEP);
		}
		else
			agSleep(computation_time);
		if(agTickCount() - lastdraw_time > computation_time)
		{
			draw();
			lastdraw_time = agTickCount();
		}
		print_status();
	} while(simulation_time < timelimit + SAVE_INTERVAL);
good_bye:;
}

/*
 * -----------------------------------------------------------------------
 * Part 7: main()
 */

static int     nodraw = 0;
static vec_t   diff_en_limit = 1e6;

int main(int argc,char *argv[])
{
	int i;
	for(i = 1;i < argc;i++)
	{
		char *cmd = argv[i];
		if(!strcmp(cmd,"nodraw"))
			nodraw = 1;
		else if(!strcmp(cmd,"diff_en_limit"))
			diff_en_limit = atof(argv[++i]);
		else if(!strcmp(cmd,"timelimit"))
			timelimit = atof(argv[++i]);
	}
	initial_conditions();
	printf("Diff. en. limit %g\nTimelimit %g\n",diff_en_limit,timelimit);
	if(nodraw)
	{
		int i,st;
		stoer_first(STEP);
		for(st = 0;;st++)
		{
			for(i = 0;i < 10;i++)
				stoer_next(STEP);
			print_status(); /* ignore 'paused' state completely */
			if(simulation_time > timelimit)
				break;
			if(max_diff > diff_en_limit)
				break;
		}
	}
	else
	{
		agWindowConfig("N-body problem: Stoermer's rule",
			768, 768, GL_FALSE);
		agInit(AG_REQUIRED_1_1,GL_FALSE,GL_FALSE,GL_FALSE);
		glClearColor(1,1,1,0.5f);
		correct_frustum();
#ifndef AA_DISABLED
		glEnable(GL_POINT_SMOOTH);
		glEnable(GL_LINE_SMOOTH);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
#endif  /* !AA_DISABLED */
		main_loop();
		agShutdown();
	}
	return 0;
}

