#include <antiglut.h>
#include <math.h>
#include <stdio.h>
#include <assert.h>
#include "mvlib.h"

#define NP	2

vec_t	xne[NP];
vec_t	yne[NP];
vec_t	zne[NP];

vec_t	xc[NP];
vec_t	yc[NP];
vec_t	zc[NP];

vec_t	xp[NP];
vec_t	yp[NP];
vec_t	zp[NP];

vec_t	mass[NP];

#define	dt		0.0005

#define	_sqr_(x)	((x)*(x))

vec_t force(int i,int ncoord)
{
	int j = i ? 0 : 1;
	vec_t drx = xc[j] - xc[i];
	vec_t dry = yc[j] - yc[i];
	vec_t drz = zc[j] - zc[i];
	vec_t dr = sqrt(_sqr_(drx)+_sqr_(dry)+_sqr_(drz));
	vec_t fa = mass[i] * mass[j] / (dr * dr);
	switch(ncoord)
	{
		case 0: return fa * (drx) / dr;
		case 1: return fa * (dry) / dr;
		case 2: return fa * (drz) / dr;
	}
	return 0;
}

void dump_all()
{
	int i;
	for(i=0;i<NP;i++)
	{
		printf("particle[%d] = (%g,%g,%g)->(%g,%g,%g)->(%g,%g,%g)\n",i,
			xp[i],yp[i],zp[i],
			xc[i],yc[i],zc[i],
			xne[i],yne[i],zne[i]);
	}
	puts("");
}
	
void vnext()
{
#define	VNEXT(pref,ncord) \
		pref##ne[i] = 2*pref##c[i] - pref##p[i] + (force(i,ncord) / mass[i]) * dt * dt
		
	int i;
	for(i=0;i<NP;i++)
	{
		VNEXT(x,0);
		VNEXT(y,1);
		VNEXT(z,2);
	}

	/* shift coords */
	for(i=0;i<NP;i++)
	{
		xp[i] = xc[i];
		xc[i] = xne[i];

		yp[i] = yc[i];
		yc[i] = yne[i];

		zp[i] = zc[i];
		yc[i] = yne[i];
	}
}

void initcond(vec_t *x,vec_t *y,vec_t *z,vec_t *ivx,vec_t *ivy,vec_t *ivz)
{
	int i;
	for(i=0;i<NP;i++)
	{
		mass[i] = 1;
		xc[i] = x[i];
		yc[i] = y[i];
		zc[i] = z[i];
		/*
		 * Euler's sceme to calc prev: center - v * dt = prev 
		*/
		xp[i] = xc[i] - ivx[i] * dt;
		yp[i] = yc[i] - ivy[i] * dt;
		zp[i] = zc[i] - ivz[i] * dt;
	}
}

void draw()
{
	int i;

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

#define GS	1.9f
	glScalef(3.f/4.f * GS,GS,GS);
	
	glClear(GL_COLOR_BUFFER_BIT);
	glColor3f(0,0,0.3f);
	glBegin(GL_QUADS);
		glVertex3f(1,1,0);
		glVertex3f(-1,1,0);
		glVertex3f(-1,-1,0);
		glVertex3f(1,-1,0);
	glEnd();
	glColor3f(1,1,1);
	glPointSize(3);

	glBegin(GL_POINTS);
	for(i=0;i<NP;i++)
	{
		glVertex3d(xc[i],yc[i],zc[i]);
	}
	glEnd();

	agSwap();

	agSleep(5);
}

void gr_init()
{
	agInit(AG_REQUIRED_1_1,GL_FALSE,GL_FALSE,GL_FALSE);
	glClearColor(0,0,0,0.5f);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-2,2,-2,2,-2,2);
}

void dump()
{
	int i;
	for(i = 0;i < NP;i++)
		printf("particle[%d] = (%g,%g,%g)\n",i,xc[i],yc[i],zc[i]);
	puts("");
}

int main()
{
	int i;
	int nsteps = 100000;
	int monsteps = 10, stepcount = 0;
	vec_t ix[NP] = {-1,1}, iy[NP] = {0,0}, iz[NP] = {0,0};
	vec_t ivx[NP] = {0,0}, ivy[NP] = {0.3,-0.3}, ivz[NP] = {0,0};
	initcond(ix,iy,iz,ivx,ivy,ivz);
	gr_init();
	for(i = 1;i <= nsteps;i++)
	{
		vnext();
		if(++stepcount == monsteps)
		{
			stepcount = 0;
			draw();
			/* dump(); */
			
		}
	}
	agShutdown();
	return 0;
}
