#include "mvlib.h"

/*
 *	mat3Rotate - computes arbitrary axis rotation matrix.
 *	Original code: MesaGL
 *	NOTES.
 *		1. Assuming that 'dir' vector is normalized.
 *		2. Matrix rotates obects COUNTERCLOCKWISE, if look towards origin
 *		of 'dir'.
 */
void	V_GLOBAL
mat3Rotate(matrix3_t mat,vec3_t n,vec_t angle)
{
	vec_t s, c;
	vec_t xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;
	s = sin(angle);
	c = cos(angle);
	/* Arbitrary axis rotation matrix.
	 * Author: Erich Boleyn (erich@uruk.org); see MesaGL code. */
	xx = n[0]*n[0];
	yy = n[1]*n[1];
	zz = n[2]*n[2];
	xy = n[0]*n[1];
	yz = n[1]*n[2];
	zx = n[2]*n[0];
	xs = n[0]*s;
	ys = n[1]*s;
	zs = n[2]*s;
	one_c = 1.0F - c;
	mat[0][0] = (one_c * xx) + c;
	mat[0][1] = (one_c * xy) - zs;
	mat[0][2] = (one_c * zx) + ys;
	mat[1][0] = (one_c * xy) + zs;
	mat[1][1] = (one_c * yy) + c;
	mat[1][2] = (one_c * yz) - xs;
	mat[2][0] = (one_c * zx) - ys;
	mat[2][1] = (one_c * yz) + xs;
	mat[2][2] = (one_c * zz) + c;
}


/* Same as mat3FromDirAndUp */
void	V_GLOBAL
mat3x4FromDirAndUp(matrix3x4_t mat,vec3_t dir,vec3_t up)
{
	vec3_t left;
	vec3Cross(left,up,dir);
	mat3x4SetRows(mat,left,up,dir);
}

void	V_GLOBAL
mat3x4FromPointDirAndUp(matrix3x4_t mat,vec3_t pt,vec3_t dir,vec3_t up)
{
	mat3x4FromDirAndUp(mat,dir,up);
	mat[0][3] = - mat[0][0] * pt[0] - mat[0][1] * pt[1] - mat[0][2] * pt[2];
	mat[1][3] = - mat[1][0] * pt[0] - mat[1][1] * pt[1] - mat[1][2] * pt[2];
	mat[2][3] = - mat[2][0] * pt[0] - mat[2][1] * pt[1] - mat[2][2] * pt[2];
}
