#ifndef	MVLIB_H
#define	MVLIB_H

#include <math.h>

#ifndef M_PI2
#define	M_PI2 1.5707963267949f
#endif

#ifndef M_PI
#define	M_PI 3.14159265358979f
#endif

#ifndef M_2PI
#define	M_2PI 6.28318530717959f
#endif

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#define	V_GLOBAL
#define V_LOCAL	static

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

typedef	int	s32_t;
typedef unsigned u32_t;

typedef int	vboolean;
#ifndef __cplusplus
#define false	0
#define true	(!false)
#endif

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

typedef	GLfloat	vec_t, *vec_p;
typedef vec_t	vec2_t[2];
typedef	vec_t	vec3_t[3];

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#ifndef	SQRT_3
#define	SQRT_3	VEC(1.73205080756887729352745)
#endif

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#define	MAXVEC	((vec_t)(1E+38))
#define	FLOAT_ZERO_EPSILON	((vec_t)1E-24)
#define IsZero(v)	(fabs(v) < FLOAT_ZERO_EPSILON)

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/*
 * The equation of plane is
 * $$\mathb{r} \cdot normal - dist = 0$$
 */
typedef struct	plane3_s
{
	vec3_t  normal;
	vec_t   dist;
} plane3_t, *plane3_p;

/*
 * Parametric equation of the line.
 * 'dir' is the direction vector.
 */
typedef struct	line3_s
{
	vec3_t origin;	/*	any point on line */
	vec3_t dir;		/*	direction vector */
} line3_t, *line3_p;


/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#define	vec2Zero(dst) \
{ \
	(dst)[0] = 0; \
	(dst)[1] = 0; \
}

#define	vec2Set(dst, x, y, z) \
{ \
	(dst)[0] = x; \
	(dst)[1] = y; \
}

#define	vec2Copy(dst, src) \
{ \
	(dst)[0] = (src)[0]; \
	(dst)[1] = (src)[1]; \
}

#define	vec2Add(dst, v1, v2) \
{ \
	(dst)[0] = (v1)[0] + (v2)[0]; \
	(dst)[1] = (v1)[1] + (v2)[1]; \
}

#define	vec2Sub(dst, v1, v2) \
{ \
	(dst)[0] = (v1)[0] - (v2)[0]; \
	(dst)[1] = (v1)[1] - (v2)[1]; \
}

#define	vec2Scale(dst, v, scale) \
{ \
	vec_t local_scale = (vec_t)(scale); \
	(dst)[0] = (v)[0] * local_scale; \
	(dst)[1] = (v)[1] * local_scale; \
}

#define	vec2Dot(v1, v2) \
	((v1)[0] * (v2)[0] + (v1)[1] * (v2)[1])

#define	vec2Abs(v) \
	((vec_t)sqrt(vec2Dot(v,v)))

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#define	vec3Zero(dst) \
{ \
	(dst)[0] = 0; \
	(dst)[1] = 0; \
	(dst)[2] = 0; \
}

#define	vec3Set(dst, x, y, z) \
{ \
	(dst)[0] = x; \
	(dst)[1] = y; \
	(dst)[2] = z; \
}

#define	vec3Copy(dst, src) \
{ \
	(dst)[0] = (src)[0]; \
	(dst)[1] = (src)[1]; \
	(dst)[2] = (src)[2]; \
}

#define	vec3Add(dst, v1, v2) \
{ \
	(dst)[0] = (v1)[0] + (v2)[0]; \
	(dst)[1] = (v1)[1] + (v2)[1]; \
	(dst)[2] = (v1)[2] + (v2)[2]; \
}

#define	vec3Sub(dst, v1, v2) \
{ \
	(dst)[0] = (v1)[0] - (v2)[0]; \
	(dst)[1] = (v1)[1] - (v2)[1]; \
	(dst)[2] = (v1)[2] - (v2)[2]; \
}

#define	vec3Scale(dst, v, scale) \
	{ \
		vec_t local_scale = (vec_t)(scale); \
		(dst)[0] = (v)[0] * local_scale; \
		(dst)[1] = (v)[1] * local_scale; \
		(dst)[2] = (v)[2] * local_scale; \
	}

#define	vec3Dot(v1, v2) \
	((v1)[0] * (v2)[0] + (v1)[1] * (v2)[1] + (v1)[2] * (v2)[2])

#define	vec3Abs(v) \
	((vec_t)sqrt(vec3Dot(v,v)))

#define vec3Unit(dst,v) \
	{ \
		vec_t rlen = 1 / vec3Abs(v); \
		(dst)[0] = (v)[0] * rlen; \
		(dst)[1] = (v)[1] * rlen; \
		(dst)[2] = (v)[2] * rlen; \
	}
		
#define	bboxClear(mins,maxs) \
{ \
	vec3Set(mins,MAXVEC,MAXVEC,MAXVEC); \
	vec3Set(maxs,-MAXVEC,-MAXVEC,-MAXVEC); \
}

#define	bboxAddPoint(mins,maxs,p) \
{ \
	if((mins)[0] > (p)[0]) (mins)[0] = (p)[0]; \
	if((maxs)[0] < (p)[0]) (maxs)[0] = (p)[0]; \
	if((mins)[1] > (p)[1]) (mins)[1] = (p)[1]; \
	if((maxs)[1] < (p)[1]) (maxs)[1] = (p)[1]; \
	if((mins)[2] > (p)[2]) (mins)[2] = (p)[2]; \
	if((maxs)[2] < (p)[2]) (maxs)[2] = (p)[2]; \
}

#endif	/* !MVLIB_H */

void bboxDraw(vec3_t mins,vec3_t maxs);
void line3FromPoints(line3_p dest,vec3_t p1,vec3_t p2);
void line3CommonVertical(line3_p l1, line3_p l2,
	vec3_t l1base_ret, vec3_t l2base_ret,
	vec_p mindist_ret);
