/*
 * Common config macro(s)
 */
#define	DISCARD_DUPLICATE_MOUSEMOVE

#ifndef	__LIBV__

#include "../include/antiglut.h"
#include <stdio.h>
#include <stdlib.h>
#ifndef __STDC__
#include <malloc.h>
#endif /* !__STDC__ */
#include <string.h>
#include <assert.h>

/*
 * Global data (should be stored in TLS).
 */

static struct thread_window_config_s
{
	char title[512];
	GLuint width, height;
	GLuint left, top;
	GLboolean resizeable;
	GLboolean fullscreen;
}	thread_wc = {"AntiGlut",
		0, 0, 		/* Use desktop dimensions */
		0, 0,		/* Upper left corner */
		GL_FALSE,	/* Fixed window size */
		GL_TRUE};	/* Fullscreen */

static void	*thread_glctx = NULL;

#define	get_thread_window_config_addr()	(&thread_wc)
#define	get_thread_context_addr()	(&thread_glctx)

/*
 * Wrappers for LibV memory management routines.
 * Macro MALLOC_VERBOSE may be set in antiglut.h.
 */

static GLuint num_mem_blocks = 0;

#define	memZero(ptr,size)	memset(ptr,0,size)

static void	*memAllocZ(GLuint size)
{
	void *ptr;
	assert(size);
	ptr = malloc(size);
#ifdef	MALLOC_VERBOSE
	printf("malloc(): %u bytes: %X\n",size,ptr);
#endif	/* MALLOC_VERBOSE */
	if(ptr)
	{
		memset(ptr,0,size);
		num_mem_blocks++;
		return ptr;
	}
	TRACE("Not enought memory.");
	assert(0);
	return NULL; /* make compiler happy */
}

#define	memAlloc	memAllocZ

static void	memFree(void *ptr)
{
	assert(ptr);
#ifdef	MALLOC_VERBOSE
	printf("free(): %X; ",ptr);
#endif	/* MALLOC_VERBOSE */
	if(num_mem_blocks)
	{
		num_mem_blocks--;
		free(ptr);
#ifdef	MALLOC_VERBOSE
		printf("(%u block(s) left)\n",num_mem_blocks);
#endif	/* MALLOC_VERBOSE */
		return;
	}
	TRACE("Extra free().");
	assert(0);
}

static char *strDup(char *str)
{
	unsigned len;
	char *ptr;
	if(!str)
		return NULL;
	len = strlen(str)+1;
	ptr = (char*)memAlloc(len+1);
	memcpy(ptr,str,len);
	return ptr;
}

/*
 * Template for deque ([F|L]IFO), used for implementation of message queue.
 * Circular doubly-linked list is used.
 */

#define	DEQUE_STRUCT(prefix,storage_class) \
\
typedef struct prefix##_deque_s \
{ \
	storage_class	store; \
	struct prefix##_deque_s *next, *prev; \
}	prefix##_deque_t, *prefix##_deque_p;

#define	IMPLEMENT_DEQUE_API(prefix,calling_convention,storage_class) \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPush(prefix##_deque_p deq,storage_class *store) \
{ \
	prefix##_deque_p node = \
		(prefix##_deque_p)memAllocZ(sizeof(prefix##_deque_t)); \
	memcpy(&node->store,store,sizeof(storage_class)); \
	if(!deq) \
	{ \
		node->next = node; \
		node->prev = node; \
		return node; \
	} \
	/* insert _before_ que */ \
	node->next = deq; \
	node->prev = deq->prev; \
	deq->prev = node; \
	node->prev->next = node; \
	return node; \
} \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPop(prefix##_deque_p deq,storage_class *store_ret) \
{ \
	prefix##_deque_p next, prev; \
	if(!deq) \
		return NULL; \
	if(store_ret) \
		memcpy(store_ret,&deq->store,sizeof(storage_class)); \
	next = deq->next; \
	prev = deq->prev; \
	next->prev = prev; \
	prev->next = next; \
	if(next == deq) \
		next = NULL; \
	memFree(deq); \
	return next; \
} \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPopBack(prefix##_deque_p deq,storage_class *store_ret) \
{ \
	return deq ? prefix##DeqPop(deq->prev,store_ret) : NULL; \
} \
\
static GLboolean	calling_convention \
prefix##DeqPeek(prefix##_deque_p deq, \
	unsigned offset_from_last_push,storage_class **store_ptr_ret) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return GL_FALSE; \
	do \
	{ \
		if(count == offset_from_last_push) \
		{ \
			if(store_ptr_ret) \
				*store_ptr_ret = &ptr->store; \
			return GL_TRUE; \
		} \
		count++; \
		ptr = ptr->next; \
	} while(ptr != deq); \
	return GL_FALSE; \
} \
\
static GLboolean	calling_convention \
prefix##DeqPeekBack(prefix##_deque_p deq, \
	unsigned offset_from_first_push, storage_class **store_ptr_ret) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return GL_FALSE; \
	do \
	{ \
		if(count == offset_from_first_push) \
		{ \
			if(store_ptr_ret) \
				*store_ptr_ret = &ptr->prev->store; \
			return GL_TRUE; \
		} \
		count++; \
		ptr = ptr->prev; \
	} while(ptr != deq); \
	return GL_FALSE; \
} \
\
static prefix##_deque_p calling_convention \
prefix##DeqRemove(prefix##_deque_p deq, \
	unsigned offset_from_last_push) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return NULL; \
	do \
	{ \
		if(count == offset_from_last_push) \
		{ \
			if(ptr == deq) \
				return prefix##DeqPop(ptr,NULL); \
			prefix##DeqPop(ptr,NULL); \
			return deq; \
		} \
		count++; \
		ptr = ptr->next; \
	} while(ptr != deq); \
	return deq; \
} \
\
static prefix##_deque_p calling_convention \
prefix##DeqRemoveBack(prefix##_deque_p deq, \
	unsigned offset_from_first_push) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return NULL; \
	do \
	{ \
		ptr = ptr->prev; \
		if(count == offset_from_first_push) \
		{ \
			if(ptr == deq) \
				return prefix##DeqPop(ptr,NULL); \
			prefix##DeqPop(ptr,NULL); \
			return deq; \
		} \
		count++; \
	} while(ptr != deq); \
	return deq; \
} \
\
static void	calling_convention \
prefix##DeqFree(prefix##_deque_p deq) \
{ \
	prefix##_deque_p head, next; \
	if(!deq) \
		return; \
	head = deq; \
	do \
	{ \
		next = deq->next; \
		memFree(deq); \
		deq = next; \
	} while(deq != head); \
}

#define	DEQ_PUSH(prefix,deq,store) \
	((deq) = prefix##DeqPush(deq,store))

#define	DEQ_POP(prefix,deq,store_ret) \
	((deq) = prefix##DeqPop(deq,store_ret))

#define	DEQ_POP_BACK(prefix,deq,store_ret) \
	((deq) = prefix##DeqPopBack(deq,store_ret))

#define	DEQ_PEEK(prefix,deq,offset_from_last_push,store_ret) \
	prefix##DeqPeek(deq,offset_from_last_push,store_ret)

#define	DEQ_PEEK_BACK(prefix,deq,offset_from_first_push,store_ret) \
	prefix##DeqPeekBack(deq,offset_from_first_push,store_ret)

#define	DEQ_REMOVE(prefix,deq,offset_from_last_push) \
	((deq) = prefix##DeqRemove(deq,offset_from_last_push))

#define	DEQ_REMOVE_BACK(prefix,deq,offset_from_first_push) \
	((deq) = prefix##DeqRemoveBack(deq,offset_from_first_push))

#define	DEQ_FREE(prefix,deq)	prefix##DeqFree(deq)

/*
 * Event que declaration and implementation
 */

#define	DEFAULT_CALLING_CONVENTION

DEQUE_STRUCT(event,ui_event_t)

IMPLEMENT_DEQUE_API(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#ifdef	__UNIX__

#define	IMPLEMENT_DEQUE_PB(prefix,calling_convention,storage_class) \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPushBack(prefix##_deque_p deq,storage_class *store) \
{ \
	prefix##_deque_p new_head = prefix##DeqPush(deq,store); \
	return deq ? deq : new_head; \
} \

#define	DEQ_PUSH_BACK(prefix,deq,store) \
	((deq) = prefix##DeqPushBack(deq,store))

IMPLEMENT_DEQUE_PB(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#else	/* !__UNIX__ */

#define	IMPLEMENT_DEQUE_SCANS(prefix,calling_convention,storage_class) \
\
static prefix##_deque_p calling_convention \
prefix##DeqScanForward(prefix##_deque_p deq, \
	int (calling_convention *reject)(storage_class*,void*),void *arg) \
{ \
	prefix##_deque_p head, ptr, tail; \
	if(!deq) \
		return NULL; \
	deq = deq->prev; \
	head = NULL; \
	tail = deq; \
	do \
	{ \
		ptr = deq->prev; \
		if(reject(&deq->store,arg)) \
			memFree(deq); \
		else \
		{ \
			deq->next = head; \
			head = deq; \
		} \
		deq = ptr; \
	} while(deq != tail); \
	if(!head) \
		return NULL; \
	for(tail = head, ptr = head->next;ptr;ptr = ptr->next) \
	{ \
		ptr->prev = tail; \
		tail = ptr; \
	} \
	head->prev = tail; \
	tail->next = head; \
	return head; \
} \
\
static prefix##_deque_p calling_convention \
prefix##DeqScanBackward(prefix##_deque_p deq, \
	int (calling_convention *reject)(storage_class*,void*),void *arg) \
{ \
	prefix##_deque_p head, ptr, tail; \
	if(!deq) \
		return NULL; \
	head = deq; \
	tail = NULL; \
	do \
	{ \
		ptr = deq->next; \
		if(reject(&deq->store,arg)) \
			memFree(deq); \
		else \
		{ \
			deq->prev = tail; \
			tail = deq; \
		} \
		deq = ptr; \
	} while(deq != head); \
	if(!tail) \
		return NULL; \
	for(head = tail, ptr = tail->prev;ptr;ptr = ptr->prev) \
	{ \
		ptr->next = head; \
		head = ptr; \
	} \
	head->prev = tail; \
	tail->next = head; \
	return head; \
}

#define	DEQ_SCAN_FORWARD(prefix,deq,reject,arg) \
	((deq) = prefix##DeqScanForward(deq,reject,arg))

#define	DEQ_SCAN_BACKWARD(prefix,deq,reject,arg) \
	((deq) = prefix##DeqScanBackward(deq,reject,arg))

IMPLEMENT_DEQUE_SCANS(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#endif	/* !__UNIX__ */

#else	/* __LIBV__ */

/*
 * #include "antiglut.h"
 */

#define	get_thread_window_config_addr() \
	(&(GetThreadData()->window_config))

#define	get_thread_context_addr() \
	(&(GetThreadData()->gl_context))

#define	DEFAULT_CALLING_CONVENTION

DEQUE_STRUCT(event,ui_event_t)

IMPLEMENT_DEQUE_API(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#endif	/* __LIBV__ */

#define	alloc_gl_context()	\
	(gl_context_p)memAllocZ(sizeof(gl_context_t))

/*
 * Some WM/GL data is common for both platforms...
 */

typedef struct vidmode_s
{
	GLuint	width;
	GLuint	height;
	GLuint	bpp;		/* set correctly but not used under X11 */
	GLuint	refresh;	/* set correctly but not used under X11 */
#ifdef	ARCH_HAVE_XF86_VIDMODE
	GLuint	modeline;
#endif	/* ARCH_HAVE_XF86_VIDMODE */
}	vidmode_t, *vidmode_p;

#define	MAX_WINDOW_TITLE	128

#define	ARCH_COMMON_WINDOW_FIELDS \
	char	title[MAX_WINDOW_TITLE]; /* caption */ \
	GLuint	x, y; /* [left,top] */ \
	GLuint	width, height; \
	GLuint	root_width, root_height; \
	/* X11 - time of 1st input msg; Win32 - time of wnd creation */ \
	GLuint start_time; \
	GLboolean focused; /* affected by winsystem | archWinFocus() */ \
	GLboolean visible; /* affected by winsystem | archWinShow() */ \
	GLuint pointer_hide_count; /* affected by archWinShowPointer() */ \
	GLuint ms_grab_count; /* affected by archWinGrabMouse() */ \
	GLuint kb_grab_count; /* affected by archWinGrabKeyboard() */ \
	GLint  current_vidmode; /* affected by archWinSetVidMode() */ \
	GLuint saved_pointer_hide_count; /* affected by ui_win_loose_control() */ \
	GLuint saved_ms_grab_count; /* affected by ui_win_loose_control() */ \
	GLuint saved_kb_grab_count; /* affected by ui_win_loose_control() */ \
	GLint  saved_vidmode; /* affected by ui_win_loose_control() */ \
	/* last ms coords, client space, as returned by windowing system */ \
	GLuint last_mx, last_my; \
	/* last ms coords, client space, as tracked by library */ \
	GLuint last_vmx, last_vmy; \
	GLuint last_key_state; /* keyboard modifiers of last input msg */ \
	GLuint last_button_state; /* mouse btns of last mouse msg */ \
	GLboolean auto_repeat_disabled; /* !filter autorepeated keys */ \
	/* mouse pointer hidden, 'infinite' mouse mode */ \
	GLboolean auto_warp_enabled; \
	GLuint auto_warp_saved_x, auto_warp_saved_y; \
	GLboolean zap_next_mousemove; /* eat next MouseMove msg */ \
	GLubyte keyboard_state[sc_maxcode]; /* 0 if sc_XXX key is released */ \
	event_deque_p msg_que; /* message deque */

#define	INIT_COMMON_WINDOW_FIELDS(winobj) \
	(winobj)->title[0] = 0; \
	(winobj)->x = (winobj)->y = 0; \
	(winobj)->width = (winobj)->height = 256; \
	(winobj)->root_width = (winobj)->root_height = 0; \
	(winobj)->start_time = 0; \
	(winobj)->visible = GL_FALSE; \
	(winobj)->focused = GL_FALSE; \
	(winobj)->pointer_hide_count = 0; \
	(winobj)->ms_grab_count = 0; \
	(winobj)->kb_grab_count = 0; \
	(winobj)->current_vidmode = -1; \
	(winobj)->saved_pointer_hide_count = 0; \
	(winobj)->saved_ms_grab_count = 0; \
	(winobj)->saved_kb_grab_count = 0; \
	(winobj)->saved_vidmode = -1; \
	(winobj)->auto_warp_enabled = GL_FALSE; \
	(winobj)->last_mx = (winobj)->last_my = 0; \
	(winobj)->last_key_state = (winobj)->last_button_state = 0; \
	(winobj)->auto_repeat_disabled = GL_FALSE; \
	(winobj)->zap_next_mousemove = GL_FALSE; \
	memZero((winobj)->keyboard_state,sc_maxcode); \
	(winobj)->msg_que = NULL;

#define	CLEAR_COMMON_WINDOW_FIELDS(winobj) \
	eventDeqFree((winobj)->msg_que); \
	(winobj)->msg_que = NULL; \

#define ARCH_COMMON_GL_CONTEXT_FIELDS \
	arch_window_p wnd; \
	char *vendor, *renderer, *version, *extensions;

#define ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx) \
	if((ctx)->vendor) { \
		memFree((ctx)->vendor); \
		(ctx)->vendor = NULL; \
	} \
	if((ctx)->renderer) { \
		memFree((ctx)->renderer); \
		(ctx)->renderer = NULL; \
	} \
	if((ctx)->version) { \
		memFree((ctx)->version); \
		(ctx)->version = NULL; \
	} \
	if((ctx)->extensions) { \
		memFree((ctx)->extensions); \
		(ctx)->extensions = NULL; \
	} \
	if((ctx)->wnd) { \
		archWinClose((ctx)->wnd); \
		(ctx)->wnd = NULL; \
	}

#define	IF_NO_GLCONTEXT(ctx) \
	(ctx) = (gl_context_p)(*get_thread_context_addr()); \
	if((ctx) == NULL)

/*
 * ... and other are not.
 */

#ifdef	__WINDOWS__

#include <windows.h>

#define	MAX_CLASS_NAME	32

typedef struct arch_window_s
{
	ARCH_COMMON_WINDOW_FIELDS
	GLuint		num_vidmodes;
	vidmode_p	vidmodes;
	char		class_name[MAX_CLASS_NAME];
	WNDCLASS	wc;
	HWND		wnd;
	HDC			dc;
	DWORD		delta_w, delta_h;
	int			wheel_delta;
	BYTE		oem_kbstate[256];
	UINT		auto_capture_button;
}	arch_window_t, *arch_window_p;

typedef struct gl_context_s
{
	ARCH_COMMON_GL_CONTEXT_FIELDS
	HGLRC wgl_context;
}	gl_context_t, *gl_context_p;

#else	/* !__WINDOWS__ */

/*
 * X11 configuration macro(s).
 */
#include <locale.h>
#include <dlfcn.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <GL/glx.h>
#include <sys/time.h>

#ifdef	ARCH_HAVE_XF86_VIDMODE
#include <X11/extensions/xf86vmode.h>
#endif	/* ARCH_HAVE_XF86_VIDMODE */

typedef struct x11_display_connection_s
{
	Window root;
	GLuint scrnum;
	GLuint screen_width, screen_height;
	char *lang;
	Display *dpy;
	XIM im;
	Atom wm_protocols_atom;
	Atom wm_delete_window_atom;
	Cursor null_cursor;
	GLboolean grab_keyboard;
	GLuint connection_fd;
	GLuint		num_vidmodes;
	vidmode_p	vidmodes;
#ifdef	ARCH_HAVE_XF86_VIDMODE
	GLboolean vidmode_active;
	XF86VidModeModeInfo **xf86_vidmodes;
#endif	/* ARCH_HAVE_XF86_VIDMODE */
}	x11_display_connection_t, *x11_display_connection_p;

typedef struct arch_window_s
{
	ARCH_COMMON_WINDOW_FIELDS
	x11_display_connection_p cn;
	Window		parent, wnd;
	XIC			ic;
        GLuint	event_mask;
	GLuint	keycode_to_scan[256];
	Colormap	cmap;
}	arch_window_t, *arch_window_p;

typedef struct gl_context_s
{
	ARCH_COMMON_GL_CONTEXT_FIELDS
	XVisualInfo *visual_info;
	GLXContext glx_context;
}	gl_context_t, *gl_context_p;

static x11_display_connection_p x11_open_connection();
static void x11_close_connection(x11_display_connection_p cn);

#endif	/* !__WINDOWS__ */

/*
 * archXXX routines
 */
static void
archWinSetVidMode(arch_window_p aw,GLint new_mode);

static void
archWinGrabMouse(arch_window_p aw,GLboolean grab);

static GLboolean
archWinWarpPointer(arch_window_p aw,GLint x,GLint y);

#ifndef	__UNIX__
static void
archWinFocus(arch_window_p aw);
#endif	/* !__UNIX__ */

static void
archWinClose(arch_window_p aw);

/*
 * ui_default_wnd_proc()
 * Returns 'GL_FALSE' if message must _not_ be stored in queue.
 */
static GLboolean
ui_default_wnd_proc(arch_window_p aw,ui_event_p ev)
{
	GLuint x, y, cx, cy, dx, dy;
	switch(ev->type)
	{
		case ui_pointer_motion_event:
                	if(aw->zap_next_mousemove)
			{
				aw->zap_next_mousemove = GL_FALSE;
				return GL_FALSE;
			}
			x = ev->client_x;
			y = ev->client_y;
#ifdef	DISCARD_DUPLICATE_MOUSEMOVE
			if(aw->last_mx == x && aw->last_my == y)
				return GL_FALSE;
#endif	/* DISCARD_DUPLICATE_MOUSEMOVE */
			dx = x - aw->last_mx;
			dy = y - aw->last_my;
			if(aw->auto_warp_enabled)
			{
				cx = aw->width >> 1;
				cy = aw->height >> 1;
				aw->last_mx = cx;
				aw->last_my = cy;
				aw->last_vmx += dx;
				aw->last_vmy += dy;
				ev->client_x = aw->last_vmx;
				ev->client_y = aw->last_vmy;
				aw->zap_next_mousemove = GL_TRUE;
				archWinWarpPointer(aw, cx, cy);
				ev->client_dx = dx;
				ev->client_dy = dy;
			}
			else
			{
				ev->client_dx = x - aw->last_mx;
				ev->client_dy = y - aw->last_my;
				aw->last_vmx = aw->last_mx = x;
				aw->last_vmy = aw->last_my = y;
			}
			return GL_TRUE;
		default:
			break;
	}
	if(aw->auto_warp_enabled)
	{
		ev->client_x = aw->last_vmx;
		ev->client_y = aw->last_vmy;
	}
	ev->client_dx = 0;
	ev->client_dy = 0;
	return GL_TRUE;
}

#ifndef	__WINDOWS__
static void
archWinGrabKeyboard(arch_window_p aw,GLboolean grab);
#endif	/* !__WINDOWS__ */

static void
ui_win_loose_control(arch_window_p aw,GLboolean hiding)
{
	if(hiding)
	{
		/* check if we need to uninstall grabs / restore vidmode */
		if(aw->ms_grab_count > 0)
		{
			aw->saved_ms_grab_count = aw->ms_grab_count;
			aw->ms_grab_count = 1;
			archWinGrabMouse(aw, GL_FALSE);
		}
		if(aw->kb_grab_count > 0)
		{
			aw->saved_kb_grab_count = aw->kb_grab_count;
			aw->kb_grab_count = 1;
#ifdef	__UNIX__
			archWinGrabKeyboard(aw, GL_FALSE);
#endif	/* __UNIX__ */
		}
		if(aw->current_vidmode != -1)
		{
			aw->saved_vidmode = aw->current_vidmode;
			archWinSetVidMode(aw, -1);
		}
		/*
		 * On exit:
		 *  ms_grab_count == kb_grab_count == 0;
		 *  current_vidmode == 0;
		 */
	}
	else
	{
		/* check if we need to reinstall grabs / set vidmode */
		if(aw->saved_ms_grab_count > 0)
		{
			aw->ms_grab_count = 0;
			archWinGrabMouse(aw, GL_TRUE);
			aw->ms_grab_count = aw->saved_ms_grab_count;
			aw->saved_ms_grab_count = 0;
		}
		if(aw->saved_kb_grab_count > 0)
		{
			aw->kb_grab_count = 0;
#ifdef	__UNIX__
			archWinGrabKeyboard(aw, GL_TRUE);
#endif	/* __UNIX__ */
			aw->kb_grab_count = aw->saved_kb_grab_count;
			aw->saved_kb_grab_count = 0;
		}
		if(aw->saved_vidmode != -1)
		{
			archWinSetVidMode(aw, aw->saved_vidmode);
			aw->saved_vidmode = -1;
		}
		/*
		 * On exit:
		 *  saved_ms_grab_count == saved_kb_grab_count == 0;
		 *  saved_vidmode == -1;
		 */
	}
}

static void
ui_make_default_key_event(ui_key_event_p ev)
{
	ev->press = GL_FALSE;
	ev->scan = sc_unknown;
	ev->acode = 0;
	ev->ucode = 0;
	ev->utf8_len = 0;
	ev->utf8[0] = 0;
}

static void
ui_set_configure_from_aw(arch_window_p aw,ui_configure_event_p ev)
{
	ev->focused = aw->focused;
	ev->visible = aw->visible;
	if(aw->visible)
	{
		ev->x = aw->x;
		ev->y = aw->y;
		ev->width = aw->width;
		ev->height = aw->height;
	}
	else
	{
		ev->x = ev->y = 0;
		ev->width = ev->height = 0;
	}
	ev->exposed_left = ev->exposed_top = 0;
	ev->exposed_right = ev->exposed_bottom = 0;
	ev->wm_event = ui_wm_unknown;
}

static GLfloat
ui_make_message_time(GLuint *start,GLuint current)
{
	GLuint delta1, delta2;
	if(!*start)
		*start = current;
	if(*start <= current)
	{
		delta1 = current - *start;
		delta2 = 0;
	}
	else
	{
		/* overflow ! */
		GLuint last;
		last = 0xFFFFFFFF;
		delta1 = last - *start;
		delta2 = *start = current;
	}
	return (GLfloat)0.001 * ((GLfloat)delta1 + (GLfloat)delta2);
}

/*
 * Windows - specific stuff
 */

#ifdef __WINDOWS__

static char *LIBV_HANDLE_PROP = "LIBV_HANDLE";

/*
 * Translates wParam and lParam into LibV scan code.
 */
static GLuint
win_make_scan_code(WPARAM wParam,LPARAM lParam)
{
	GLubyte scan = (GLubyte)((lParam >> 16) & 0xFF);
	GLubyte ext = (GLubyte)((lParam & (1 << 24)) != 0);

	switch(wParam)
	{
		case VK_ESCAPE:	return sc_escape;

		case VK_F1: return sc_f1;
		case VK_F2: return sc_f2;
		case VK_F3: return sc_f3;
		case VK_F4: return sc_f4;
		case VK_F5: return sc_f5;
		case VK_F6: return sc_f6;
		case VK_F7: return sc_f7;
		case VK_F8: return sc_f8;
		case VK_F9: return sc_f9;
		case VK_F10: return sc_f10;
		case VK_F11: return sc_f11;
		case VK_F12: return sc_f12;
		case VK_SCROLL: return sc_scroll_lock;

		case 192: return sc_grave;
		case '1': return sc_1;
		case '2': return sc_2;
		case '3': return sc_3;
		case '4': return sc_4;
		case '5': return sc_5;
		case '6': return sc_6;
		case '7': return sc_7;
		case '8': return sc_8;
		case '9': return sc_9;
		case '0': return sc_0;
		case 189: return sc_minus;
		case 187: return sc_equal;
		case 220: return sc_backslash;
		case VK_BACK: return sc_backspace;

		case VK_TAB: return sc_tab;
		case 'Q': return sc_q;
		case 'W': return sc_w;
		case 'E': return sc_e;
		case 'R': return sc_r;
		case 'T': return sc_t;
		case 'Y': return sc_y;
		case 'U': return sc_u;
		case 'I': return sc_i;
		case 'O': return sc_o;
		case 'P': return sc_p;
		case 219: return sc_lbracket;
		case 221: return sc_rbracket;
		case VK_RETURN:	return ext ? sc_kp_enter : sc_enter;

		case VK_CAPITAL: return sc_caps_lock;
		case 'A': return sc_a;
		case 'S': return sc_s;
		case 'D': return sc_d;
		case 'F': return sc_f;
		case 'G': return sc_g;
		case 'H': return sc_h;
		case 'J': return sc_j;
		case 'K': return sc_k;
		case 'L': return sc_l;
		case 186: return sc_semicolon;
		case 222: return sc_apostrophe;

		case 16: return (scan == 42) ? sc_lshift : sc_rshift;
		case 'Z': return sc_z;
		case 'X': return sc_x;
		case 'C': return sc_c;
		case 'V': return sc_v;
		case 'B': return sc_b;
		case 'N': return sc_n;
		case 'M': return sc_m;
		case 188: return sc_comma;
		case 190: return sc_period;
		case 191: return sc_slash;

		case 17: return ext ? sc_rcontrol : sc_lcontrol;
		case VK_LWIN: return sc_lsuper;
		case 18: return ext ? sc_ralt : sc_lalt;
		case VK_SPACE: return sc_space;
		case VK_RWIN: return sc_rsuper;
		case VK_APPS: return sc_menu;
		case VK_NUMLOCK: return sc_num_lock;
		case VK_DIVIDE: return sc_kp_divide;
		case 106: return sc_kp_multiply;
		case 109: return sc_kp_subtract;
		case 107: return sc_kp_add;
		/* case VK_CLEAR: return sc_kp_begin; */
	}
	switch(scan)
	{
		case 71: return ext ? sc_home : sc_kp_home;
		case 72: return ext ? sc_up : sc_kp_up;
		case 73: return ext ? sc_prior : sc_kp_prior;
		case 75: return ext ? sc_left : sc_kp_left;
		case 76: return sc_kp_begin;
		case 77: return ext ? sc_right : sc_kp_right;
		case 79: return ext ? sc_end : sc_kp_end;
		case 80: return ext ? sc_down : sc_kp_down;
		case 81: return ext ? sc_next : sc_kp_next;
		case 82: return ext ? sc_insert : sc_kp_insert;
		case 83: return ext ? sc_delete : sc_kp_delete;
		case 69: return sc_pause;
		case 55: return sc_print;
	}
	return sc_unknown;
}

static void
win_get_key_modifiers(arch_window_p aw,ui_event_p ev)
{
	BYTE *keys = aw->oem_kbstate;
	GetKeyboardState(keys);
	ev->state = ev->button = 0;
	if(keys[VK_NUMLOCK] & 1)
		ev->state |= UI_NUM_LOCK_MASK;
	if(keys[VK_CAPITAL] & 1)
		ev->state |= UI_CAPS_LOCK_MASK;
	if(keys[VK_SCROLL] & 1)
		ev->state |= UI_SCROLL_LOCK_MASK;

	if(keys[VK_LCONTROL] & 0x80 || keys[VK_RCONTROL] & 0x80)
		ev->state |= UI_CONTROL_MASK;
	if(keys[VK_LSHIFT] & 0x80 || keys[VK_RSHIFT] & 0x80)
		ev->state |= UI_SHIFT_MASK;
	if(keys[VK_LWIN] & 0x80 || keys[VK_RWIN] & 0x80)
		ev->state |= UI_SUPER_MASK;
	if(keys[18] & 0x80 /*keys[VK_MENU] & 0x80 || keys[VK_RENU]*/)
		ev->state |= UI_ALT_MASK;
	if(keys[VK_LBUTTON] & 0x80)
		ev->button |= UI_LBUTTON_MASK;
	if(keys[VK_MBUTTON] & 0x80)
		ev->button |= UI_MBUTTON_MASK;
	if(keys[VK_RBUTTON] & 0x80)
		ev->button |= UI_RBUTTON_MASK;
}

static void
win_get_mouse_pos(HWND wnd,POINT *psc,ui_event_p ev)
{
	POINT pcc;
	pcc.x = ev->screen_x = psc->x;
	pcc.y = ev->screen_y = psc->y;
	if(ScreenToClient(wnd,&pcc))
	{
		ev->client_x = pcc.x;
		ev->client_y = pcc.y;
	}
	else
	{
		ev->client_x = psc->x;
		ev->client_y = psc->y;
	}
}

/*
 * Should be called _after_ win_get_key_modifiers().
 */
#ifndef	CP_UTF8
#define CP_UTF8                   65001       /* UTF-8 translation */
#endif

static void
win_get_char_codes(arch_window_p aw,WPARAM wParam,LPARAM lParam,ui_event_p ev)
{
	int ulen;
	WORD chars[4];
	WCHAR uchars[4];
	UINT oem_scan = (lParam >> 16) & 0xFF;
	ev->key.acode = ev->key.ucode = 0;
	if(ToAscii(wParam,oem_scan,aw->oem_kbstate,chars,0) == 1)
		ev->key.acode = chars[0] & 0xFF;
	ulen = ToUnicode(wParam,oem_scan,aw->oem_kbstate,uchars,4,0);
	if(ulen  == 1)
		ev->key.ucode = uchars[0];
	if(ev->key.acode < 0x80)
	{
		ev->key.utf8_len = 1;
		ev->key.utf8[0] = (unsigned char)ev->key.acode;
		ev->key.utf8[1] = 0;
	}
	else if(ulen)
	{
		int retval = WideCharToMultiByte(CP_UTF8,0,uchars,ulen,
			(LPSTR)ev->key.utf8,AG_UTF8_MAXLEN,NULL,NULL);
		if(retval > 0)
		{
			if(retval >= AG_UTF8_MAXLEN)
				retval = AG_UTF8_MAXLEN - 1;
			ev->key.utf8_len = (unsigned)retval;
			ev->key.utf8[retval] = 0;
		}
	}
}

#define	SIGNED_LOWORD(src) \
	(((src) & 0x8000) ? LOWORD(src) | 0xFFFF0000 : LOWORD(src))

#define	SIGNED_HIWORD(src) \
	(((src) & 0x80000000) ? HIWORD(src) | 0xFFFF0000 : HIWORD(src))

static void
win_window_proc(arch_window_p aw,
	UINT msg,WPARAM wParam,LPARAM lParam,DWORD time,POINT *psc)
{
	POINT def_psc;
	RECT rc;
	PAINTSTRUCT ps;
	ui_event_t dst;
	GLboolean zap_message = GL_FALSE;
	unsigned long u32bit;
	if(!time)
		time = GetTickCount();
	if(!psc)
	{
		psc = &def_psc;
		GetCursorPos(psc);
	}
	dst.time = ui_make_message_time(&aw->start_time,time);
	dst.type = ui_no_event;
	ui_make_default_key_event(&dst.key);
	switch(msg)
	{
		case WM_SHOWWINDOW:
			if(wParam)
			{
				aw->visible = GL_TRUE;
				dst.type = ui_window_map_event;
			}
			else
			{
				aw->visible = GL_FALSE;
				dst.type = ui_window_unmap_event;
			}
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_SYSCOMMAND:
			switch(wParam)
			{
				case SC_MINIMIZE:
					if(!aw->visible)
						zap_message = GL_TRUE;
					else
					{
						ui_set_configure_from_aw(aw,&dst.configure);
						aw->visible = GL_FALSE;
						dst.type = ui_window_unmap_event;
					}
					break;

				case SC_MAXIMIZE:
				case SC_RESTORE:
					if(aw->visible)
						zap_message = GL_TRUE;
					else
					{
						ui_set_configure_from_aw(aw,&dst.configure);
						aw->visible = GL_TRUE;
						dst.type = ui_window_map_event;
					}
					break;

				case SC_CLOSE:
					dst.type = ui_wm_event;
					ui_set_configure_from_aw(aw,&dst.configure);
					dst.configure.wm_event = ui_wm_close;
					break;

				default:
					zap_message = GL_TRUE;
	
			}
			if(!zap_message)
			{
				win_get_mouse_pos(aw->wnd,psc,&dst);
				win_get_key_modifiers(aw,&dst);
			}
			break;

		case WM_SETFOCUS:
		case WM_KILLFOCUS:
			if(msg == WM_SETFOCUS)
			{
				aw->focused = GL_TRUE;
				dst.type = ui_window_focus_in_event;
			}
			else
			{
				aw->focused = GL_FALSE;
				dst.type = ui_window_focus_out_event;
			}
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_PAINT:
			memZero(&rc, sizeof(RECT));
			if(!GetUpdateRect(aw->wnd,&rc,FALSE))
				return;
			BeginPaint(aw->wnd,&ps);
			EndPaint(aw->wnd,&ps);
			dst.type = ui_window_expose_event;
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.configure.exposed_left = rc.left;
			dst.configure.exposed_top = rc.top;
			dst.configure.exposed_right = rc.right;
			dst.configure.exposed_bottom = rc.bottom;
			break;

		case WM_MOVE:
			if(GetUpdateRect(aw->wnd,&rc,FALSE))
				PatBlt(aw->dc,rc.left,rc.top,
					rc.right - rc.left,rc.bottom-rc.top,BLACKNESS);
			dst.type = ui_window_move_event;
			/* win_get_window_geometry */
			u32bit = lParam;
			aw->x = (GLuint)SIGNED_LOWORD(u32bit);
			aw->y = (GLuint)SIGNED_HIWORD(u32bit);
			if(aw->x == -32000)
				aw->x = 0;
			if(aw->y == -32000)
				aw->y = 0;
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_SIZE:
			if(GetUpdateRect(aw->wnd,&rc,FALSE))
				PatBlt(aw->dc,rc.left,rc.top,
					rc.right - rc.left,rc.bottom-rc.top,BLACKNESS);
			dst.type = ui_window_size_event;
			aw->width = (GLuint)LOWORD(lParam);
			aw->height = (GLuint)HIWORD(lParam);
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		/*
		 * Caller is _not_ a win_sys_window_proc().
		 */
		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYUP:
			if(msg == WM_KEYDOWN || msg == WM_SYSKEYDOWN)
				dst.key.press = GL_TRUE;
			else
				dst.key.press = GL_FALSE;
			if(aw->auto_repeat_disabled && dst.key.press &&
				(lParam & (1 << 30)))
			{
				zap_message = GL_TRUE;
				break;
			}
			dst.type = ui_key_event;
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			dst.key.scan = win_make_scan_code(wParam,lParam);
			win_get_char_codes(aw,wParam,lParam,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			if(dst.key.scan != sc_unknown)
				aw->keyboard_state[dst.key.scan] = (GLubyte)(dst.key.press ? 1 : 0);
			break;

		case WM_MOUSEMOVE:
			dst.type = ui_pointer_motion_event;
			u32bit = lParam;
			dst.client_x = SIGNED_LOWORD(u32bit);
			dst.client_y = SIGNED_HIWORD(u32bit);
			dst.screen_x = psc->x;
			dst.screen_y = psc->y;
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MBUTTONUP:
		case WM_RBUTTONUP:
		case 522: /* WM_MOUSEWHEEL */
			dst.type = ui_key_event;
			switch(msg)
			{
				case WM_LBUTTONUP:
				case WM_MBUTTONUP:
				case WM_RBUTTONUP:
					dst.key.press = GL_FALSE;
					if(msg == aw->auto_capture_button)
					{
						archWinGrabMouse(aw, GL_FALSE);
						aw->auto_capture_button = 0;
						aw->zap_next_mousemove = GL_TRUE;
					}
					break;
				case WM_LBUTTONDOWN:
					dst.key.press = GL_TRUE;
					if(!aw->auto_capture_button)
						aw->auto_capture_button = WM_LBUTTONUP;
					break;
				case WM_MBUTTONDOWN:
					dst.key.press = GL_TRUE;
					if(!aw->auto_capture_button)
						aw->auto_capture_button = WM_MBUTTONUP;
					break;
				case WM_RBUTTONDOWN:
					dst.key.press = GL_TRUE;
					if(!aw->auto_capture_button)
						aw->auto_capture_button = WM_RBUTTONUP;
					break;
				default:
					dst.key.press = GL_TRUE;
			}
			if(aw->auto_capture_button)
				archWinGrabMouse(aw, GL_TRUE);
			win_get_mouse_pos(aw->wnd,psc,&dst);
			win_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.key.acode = dst.key.ucode = 0; 
			if(msg == 522) /* WM_MOUSEWHEEL */
			{
				u32bit = wParam;
				aw->wheel_delta += SIGNED_HIWORD(u32bit);
				if(abs(aw->wheel_delta) < 120)
				{
					zap_message = GL_TRUE;
					break;
				}
				if(aw->wheel_delta > 0)
				{
					dst.key.scan = sc_wheelup;
					aw->keyboard_state[sc_wheeldown] = 0;
					aw->keyboard_state[sc_wheelup] = 1;
				}
				else
				{
					dst.key.scan = sc_wheeldown;
					aw->keyboard_state[sc_wheeldown] = 1;
					aw->keyboard_state[sc_wheelup] = 0;
				}
				aw->wheel_delta = 0;
			}
			else
			{
				if(msg == WM_LBUTTONDOWN || msg == WM_LBUTTONUP)
					dst.key.scan = sc_lbutton;
				else if(msg == WM_RBUTTONDOWN || msg == WM_RBUTTONUP)
					dst.key.scan = sc_rbutton;
				else
					dst.key.scan = sc_mbutton;
				aw->keyboard_state[dst.key.scan] = (GLubyte)(dst.key.press ? 1 : 0);
			}
			break;

		default:
			zap_message = GL_TRUE;
			break;
	}
	if(!zap_message && ui_default_wnd_proc(aw,&dst))
		DEQ_PUSH(event,aw->msg_que,&dst);
	return;
}

/*
 * Window callback. All system-generated messages and messages
 * unknown to win_window_proc() have to be processed here.
 */
static LRESULT CALLBACK
win_sys_window_proc(HWND wnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	arch_window_p aw = (arch_window_p)GetProp(wnd,LIBV_HANDLE_PROP);
	if(!aw || aw->wnd != wnd)
		aw = NULL;
	switch(msg)
	{
		case WM_CREATE:
			return 0;

		case WM_DESTROY:
			return 0;
			
		case WM_CLOSE:
			return 0;

		case WM_ERASEBKGND:
			return 1;

		case WM_SYSCOMMAND:
		case WM_SHOWWINDOW:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
		case WM_PAINT:
		case WM_MOVE:
		case WM_SIZE:
			if(aw)
				win_window_proc(aw,msg,wParam,lParam,0,NULL);
			if(msg != WM_SYSCOMMAND)
				return 0;

		default:
			break;
	}
	return DefWindowProc(wnd,msg,wParam,lParam);
}

static event_deque_p
ui_collapse_configure_events(event_deque_p msg_que);

static void
win_poll_events(arch_window_p aw)
{
	MSG msg;
	while(PeekMessage(&msg,NULL,0,0,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		if(msg.hwnd == aw->wnd)
		{
			win_window_proc(aw,msg.message,msg.wParam,msg.lParam,
				msg.time, &msg.pt);
		}
		else
		{
			arch_window_p foreign = 
				(arch_window_p)GetProp(msg.hwnd,LIBV_HANDLE_PROP);
			if(foreign && foreign->wnd == msg.hwnd)
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
				win_window_proc(foreign,msg.message,msg.wParam,msg.lParam,
					msg.time, &msg.pt);
				foreign->msg_que = 
					ui_collapse_configure_events(foreign->msg_que);
			}
		}
	}
	aw->msg_que = ui_collapse_configure_events(aw->msg_que);
}

static void
win_query_vid_modes(arch_window_p aw)
{
	DWORD i, count = 0;
	DEVMODE dm;
	memZero(&dm, sizeof(DEVMODE));
	dm.dmSize = sizeof(dm);
	for(i = 0;EnumDisplaySettings(NULL, i, &dm);i++)
	{
		if(dm.dmBitsPerPel >= 15) /* Hi/TrueColor */
			count++;
	}
	aw->num_vidmodes = count;
	if(aw->vidmodes)
		memFree(aw->vidmodes);
	if(count)
	{
		vidmode_p ptr;
		aw->vidmodes = (vidmode_p)memAllocZ(count * sizeof(vidmode_t));
		ptr = aw->vidmodes;
		for(i = 0;EnumDisplaySettings(NULL, i, &dm);i++)
		{
			if(dm.dmBitsPerPel < 15)
				continue;
			ptr->width = dm.dmPelsWidth;
			ptr->height = dm.dmPelsHeight;
			ptr->bpp = dm.dmBitsPerPel;
			ptr->refresh = dm.dmDisplayFrequency;
			if(ptr->refresh == 1)
				ptr->refresh = 0;
			ptr++;
		}
	}
	else
		aw->vidmodes = NULL;
}

static LONG
win_change_video_mode(vidmode_p new_mode)
{
	LONG result;
	if(new_mode)
	{
		DEVMODE dm;
		HDC refdc = GetDC(NULL);
		GLuint old_bpp = GetDeviceCaps(refdc,BITSPIXEL);
		GLuint old_refresh = GetDeviceCaps(refdc,VREFRESH);
		GLuint old_width = GetSystemMetrics(SM_CXSCREEN);
		GLuint old_height = GetSystemMetrics(SM_CYSCREEN);
		ReleaseDC(NULL,refdc);
		memZero(&dm, sizeof(DEVMODE));
		dm.dmSize = sizeof(dm);
		if(!new_mode->width)
		{
			TRACE("Using current desktop width.");
			dm.dmPelsWidth = old_width;
		}
		else
			dm.dmPelsWidth = new_mode->width;
		if(!new_mode->height)
		{
			TRACE("Using current desktop height.");
			dm.dmPelsHeight = old_height;
		}
		else
			dm.dmPelsHeight = new_mode->height;
		/* TODO: remove any ability to touch the BPP (unportable) */
		if(!new_mode->bpp)
		{
			TRACE("Using current desktop display depth.");
			dm.dmBitsPerPel = old_bpp;
		}
		else
			dm.dmBitsPerPel = new_mode->bpp;
		dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
		
		if(!new_mode->refresh)
		{
			if(dm.dmPelsWidth > old_width || dm.dmPelsHeight > old_height)
			{
				TRACE("No refresh rate for hi-res mode; using 75 Hz.");
				/* suitable both for CRT and (good) LCD monitors */
				dm.dmDisplayFrequency = 75;
			}
			else
			{
				TRACE("Using current refresh rate.");
				dm.dmDisplayFrequency = old_refresh;
			}
		}
		else
			dm.dmDisplayFrequency = new_mode->refresh;
		dm.dmFields |= DM_DISPLAYFREQUENCY;
		if(new_mode->width != old_width ||
				new_mode->height != old_height ||
				new_mode->bpp != old_bpp ||
				new_mode->refresh != old_refresh)
			{
				TRACE("Calling CDS()...");
				result = ChangeDisplaySettings(&dm,0);
				if(result == DISP_CHANGE_SUCCESSFUL)
				{
					TRACE("... Ok!");
				}
			}
		else
			result = DISP_CHANGE_SUCCESSFUL;
	}
	else
		result = ChangeDisplaySettings(NULL,0);
	if(result != DISP_CHANGE_SUCCESSFUL)
	{
		TRACE("ChangeDisplaySettings() failed.");
		return GL_FALSE;
	}
	return GL_TRUE;
}

static void
archWinSetVidMode(arch_window_p aw,GLint new_mode)
{
	vidmode_p mode;
	if(!aw->visible)
	{
		TRACE("Vidmode change with invisible window.");
		return;
	}
	if(!aw->focused)
	{
		TRACE("Vidmode change with non-focused window.");
		archWinFocus(aw);
	}
	if(new_mode == -1)
	{
		/* restore default mode */
		if(aw->current_vidmode != -1)
		{
			if(aw->auto_capture_button)
				archWinGrabMouse(aw, GL_FALSE);
			win_change_video_mode(NULL);
			if(aw->auto_capture_button)
			{
				archWinGrabMouse(aw, GL_TRUE);
				aw->zap_next_mousemove = GL_TRUE;
			}
			SetWindowPos(aw->wnd,HWND_NOTOPMOST,0,0,0,0,
				SWP_NOSIZE|SWP_NOMOVE);
			aw->current_vidmode = -1;
			aw->root_width = GetSystemMetrics(SM_CXSCREEN);
			aw->root_height = GetSystemMetrics(SM_CYSCREEN);
		}
		return;
	}
	mode = aw->vidmodes + new_mode;
	if(aw->auto_capture_button)
		archWinGrabMouse(aw, GL_FALSE);
	if(win_change_video_mode(mode))
	{
		aw->current_vidmode = new_mode;
		SetWindowPos(aw->wnd,HWND_TOP,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE);
		aw->root_width = mode->width;
		aw->root_height = mode->height;
	}
	if(aw->auto_capture_button)
	{
		archWinGrabMouse(aw, GL_TRUE);
		aw->zap_next_mousemove = GL_TRUE;
	}
}

static GLboolean
win_register_window_class(arch_window_p aw)
{
	unsigned pid = (unsigned)GetCurrentProcessId();
	unsigned tm = (unsigned)GetTickCount();
	sprintf(aw->class_name,"AntiGlut%u%u",pid,tm);
	memZero(&aw->wc,sizeof(WNDCLASS));
	aw->wc.lpszClassName = aw->class_name;
	aw->wc.hInstance = GetModuleHandle(NULL);
	aw->wc.style = CS_CLASSDC;
	aw->wc.lpfnWndProc = win_sys_window_proc;
	aw->wc.hIcon = LoadIcon(NULL,IDI_WINLOGO);
	aw->wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	aw->wc.hbrBackground = GetStockObject(HOLLOW_BRUSH);
	aw->wc.lpszMenuName = NULL;
	if(RegisterClass(&aw->wc))
		return GL_TRUE;
	TRACE("RegisterClass() failed.");
	return GL_FALSE;
}

static arch_window_p
archWinOpenNP(char *title,GLuint left,GLuint top,GLuint width,GLuint height,
	GLboolean caption,GLboolean allow_resize)
{
	RECT rc;
	arch_window_p aw;
	DWORD style = WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
	if(allow_resize)
	{
		if(!caption)
		{
			TRACE("Can't create resizeable window without caption.");
			return NULL;
		}
		style |= WS_OVERLAPPEDWINDOW;
	}
	else
	{
		style |= WS_POPUP;
		if(caption)
			style |= WS_CAPTION;
	}
	aw = (arch_window_p)memAllocZ(sizeof(arch_window_t));
	INIT_COMMON_WINDOW_FIELDS(aw);
	if(!win_register_window_class(aw))
	{
		archWinClose(aw);
		return NULL;
	}
	strncpy(aw->title,title,MAX_WINDOW_TITLE);
	aw->x = left;
	aw->y = top;
	aw->root_width = GetSystemMetrics(SM_CXSCREEN);
	aw->root_height = GetSystemMetrics(SM_CYSCREEN);
	if (width == 0)
		width = aw->root_width - left;
	if (height == 0)
		height = aw->root_height - top;
	aw->width = width;
	aw->height = height;
	aw->wnd = CreateWindow(aw->class_name,
		aw->title, style, left, top,
		width, height,
		NULL,
		NULL,
		GetModuleHandle(NULL),NULL);
	if(!aw->wnd)
	{
		TRACE("CreateWindow() failed.");
		archWinClose(aw);
		return NULL;
	}
	SetProp(aw->wnd,LIBV_HANDLE_PROP,(HANDLE)aw);
	GetClientRect(aw->wnd,&rc);
	aw->delta_w = aw->width - rc.right;
	aw->delta_h = aw->height - rc.bottom;
	SetWindowPos(aw->wnd,HWND_NOTOPMOST,
		0,0,
		2 * aw->width - rc.right,
		2 * aw->height - rc.bottom,
		SWP_NOMOVE | SWP_NOZORDER);
	aw->dc = GetDC(aw->wnd);
	aw->wheel_delta = 0;
	win_query_vid_modes(aw);
	memZero(aw->oem_kbstate,256);
	aw->zap_next_mousemove = GL_FALSE;
	return aw;
}

static GLboolean
archWinShow(arch_window_p aw,GLboolean show)
{
	if(show)
	{
		ShowWindow(aw->wnd,SW_SHOW);
		UpdateWindow(aw->wnd);
		ui_win_loose_control(aw, GL_FALSE);
		aw->visible = GL_TRUE;
	}
	else
	{
		ui_win_loose_control(aw, GL_TRUE);
		aw->visible = GL_FALSE;
		ShowWindow(aw->wnd,SW_HIDE);
	}
	return GL_TRUE;
}

static void
archWinFocus(arch_window_p aw)
{
	if(!aw->visible)
	{
		ShowWindow(aw->wnd,SW_SHOW);
		UpdateWindow(aw->wnd);
	}
	SetFocus(aw->wnd);
	SetActiveWindow(aw->wnd);
}

/*
 * Returns 'GL_TRUE' if cursor is visible at exit.
 */
static GLboolean
archWinShowPointer(arch_window_p aw,GLboolean show)
{
	if(show)
	{
		aw->pointer_hide_count--;
		if(!aw->pointer_hide_count)
			ShowCursor(TRUE);
	}
	else
	{
		if(!aw->pointer_hide_count)
			ShowCursor(FALSE);
		aw->pointer_hide_count++;
	}
	return (GLboolean)((aw->pointer_hide_count > 0) ? GL_FALSE : GL_TRUE);
}

static GLboolean
archWinWarpPointer(arch_window_p aw,GLint x,GLint y)
{
	POINT pt;
	pt.x = x;
	pt.y = y;
	ClientToScreen(aw->wnd, &pt);
	if(pt.x < 0 || pt.y < 0 ||
			(GLuint)pt.x >= aw->root_width ||
			(GLuint)pt.y >= aw->root_height)
	{
		TRACE("Unable to warp pointer outside of root window.");
		return GL_FALSE;
	}
	SetCursorPos(pt.x, pt.y);
	return GL_TRUE;
}

static void
archWinGrabMouse(arch_window_p aw,GLboolean grab)
{
	if(!aw->visible)
	{
		TRACE("Unable to grab mouse: window invisible.");
		return;
	}
	if(grab)
	{
		if(!aw->ms_grab_count++)
			SetCapture(aw->wnd);
	}
	else
	{
		if(!--aw->ms_grab_count)
			ReleaseCapture();
	}
}

static void
archWinClose(arch_window_p aw)
{
	if(aw->pointer_hide_count > 0)
	{
		aw->pointer_hide_count = 1;
		archWinShowPointer(aw, GL_TRUE);
	}
	ui_win_loose_control(aw, GL_TRUE);
	if(aw->wnd)
	{
		if(aw->dc)
			ReleaseDC(aw->wnd,aw->dc);
		RemoveProp(aw->wnd,LIBV_HANDLE_PROP);
		DestroyWindow(aw->wnd);
	}
	if(aw->vidmodes)
		memFree(aw->vidmodes);
	UnregisterClass(aw->class_name,GetModuleHandle(NULL));
	CLEAR_COMMON_WINDOW_FIELDS(aw);
	memFree(aw);
}

struct rejection_arg_s
{
	GLboolean was_size, was_move, was_close;
	ui_event_t last_expose;
	GLuint left, top, right, bottom;
};

static int
mvsz_rejection_proc(ui_event_p ev,void *arg)
{
	struct rejection_arg_s *rj = (struct rejection_arg_s*)arg;
	/* assume scan from last (new) to first (old) */
	switch(ev->type)
	{
		case ui_window_move_event:
			if(rj->was_move)
				return GL_TRUE;
			rj->was_move = GL_TRUE;
			break;
		case ui_window_size_event:
			if(rj->was_size)
				return GL_TRUE;
			rj->was_size = GL_TRUE;
			break;
		default:
			break;
	}
	return GL_FALSE;
}

static int
wmclose_rejection_proc(ui_event_p ev,void *arg)
{
	struct rejection_arg_s *rj = (struct rejection_arg_s*)arg;
	/* assume scan from last (new) to first (old) */
	switch(ev->type)
	{
		case ui_wm_event:
			if(ev->configure.wm_event == ui_wm_close)
			{
				if(rj->was_close)
					return GL_TRUE;
				rj->was_close = GL_TRUE;
			}
			break;
		default:
			break;
	}
	return GL_FALSE;
}

static int
expose_collapse_proc(ui_event_p ev,void *arg)
{
	struct rejection_arg_s *rj = (struct rejection_arg_s*)arg;
	/* assume scan from first (old) to last (new) */
	if(ev->type == ui_window_expose_event)
	{
		memcpy(&rj->last_expose,ev,sizeof(ui_event_t));
		if(rj->left > ev->configure.exposed_left)
			rj->left = ev->configure.exposed_left;
		if(rj->top > ev->configure.exposed_top)
			rj->top = ev->configure.exposed_top;
		if(rj->right < ev->configure.exposed_right)
			rj->right = ev->configure.exposed_right;
		if(rj->bottom < ev->configure.exposed_bottom)
			rj->bottom = ev->configure.exposed_bottom;
		return GL_TRUE;
	}
	return GL_FALSE;
}

static event_deque_p
ui_collapse_configure_events(event_deque_p msg_que)
{
	struct rejection_arg_s rj;
	if(!msg_que)
		return NULL;
	memZero(&rj, sizeof(rj));
	rj.left = rj.top = MAXLONG;
	rj.was_size = rj.was_move = rj.was_close = GL_FALSE;
	DEQ_SCAN_BACKWARD(event,msg_que,mvsz_rejection_proc,&rj);
	DEQ_SCAN_BACKWARD(event,msg_que,wmclose_rejection_proc,&rj);
	DEQ_SCAN_FORWARD(event,msg_que,expose_collapse_proc,&rj);
	if(rj.left != MAXLONG)
	{
		rj.last_expose.configure.exposed_left = rj.left;
		rj.last_expose.configure.exposed_top = rj.top;
		rj.last_expose.configure.exposed_right = rj.right;
		rj.last_expose.configure.exposed_bottom = rj.bottom;
		DEQ_PUSH(event,msg_que,&rj.last_expose);
	}
	return msg_que;
}

static void
archWinPoll(arch_window_p aw,GLboolean block)
{
	if(!block)
	{
		win_poll_events(aw);
		return;
	}
	for(;;)
	{
		win_poll_events(aw);
		if(aw->msg_que)
			break;
		WaitMessage();
	}
}

static GLboolean
vgl_load_proc(char *proc_name,void **ptr_ret, GLboolean *terminate)
{
	void *proc = (void*)wglGetProcAddress((LPCSTR)proc_name);
	if(*ptr_ret && *ptr_ret != proc)
	{
		/*
		 * Different implementation (M$ generic after HW or vice versa).
		 */
		*terminate = GL_TRUE;
		return GL_FALSE;
	}
	if(proc)
	{
		*ptr_ret = (void*)proc;
		return GL_TRUE;
	}
	return GL_FALSE;
}

void
agSwap(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	SwapBuffers(ctx->wnd->dc);
}

void
agShutdown(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	*get_thread_context_addr() = NULL;
	wglMakeCurrent(NULL, NULL);
	if(ctx->wgl_context)
		wglDeleteContext(ctx->wgl_context);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx);
	memFree(ctx);
}

#ifdef __BORLANDC__
#include <float.h>
#endif

static gl_context_p
vglInit(
	char *caption,
	GLuint x, GLuint y, GLuint width, GLuint height,
	GLboolean fullscreen,
	GLboolean can_resize,
	GLboolean need_depth,
	GLboolean need_stencil,
	GLboolean need_dst_alpha)
{
	PIXELFORMATDESCRIPTOR pfd;
	gl_context_p ctx;
	GLuint bpp;
	GLuint i, bestmode = 0xFFFFFFFF, pixfmt;
	GLfloat aspect;
	GLboolean with_caption;
	ctx = alloc_gl_context();
	if(!ctx)
		return NULL;
	if(fullscreen)
		with_caption = GL_FALSE;
	else
		with_caption = GL_TRUE;
	ctx->wnd = archWinOpenNP(caption, x, y, width, height,
		with_caption,
		can_resize);
	if(!ctx->wnd)
		goto err_exit;
	width = ctx->wnd->width;
	height = ctx->wnd->height;
	archWinShow(ctx->wnd, GL_TRUE);
	bpp = GetDeviceCaps(ctx->wnd->dc, BITSPIXEL);
	if(bpp <= 8)
	{
		if(!fullscreen)
		{
			TRACE("True or high color mode required.");
			goto err_exit;
		}
		else
			bpp = 32;
	}
	if(fullscreen)
	{
		for(i = 0; i < ctx->wnd->num_vidmodes; i++)
		{
			vidmode_p mode = ctx->wnd->vidmodes + i;
			if((GLuint)mode->width == width && 
				(GLuint)mode->height == height &&
				(GLuint)mode->bpp == bpp) /* exact match */
			{
				bestmode = i;
				break;
			}
		}
		if(bestmode != 0xFFFFFFFF)
			archWinSetVidMode(ctx->wnd, bestmode);
	}
#ifdef __BORLANDC__
	_control87(MCW_EM,MCW_EM);
#endif
	memZero(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.iLayerType = PFD_MAIN_PLANE;
	pfd.cColorBits = 16;
	if(need_depth)
		pfd.cDepthBits = 16;
	if(need_stencil)
		pfd.cStencilBits = 1;
	if(need_dst_alpha)
		pfd.cAlphaBits = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER
		| PFD_SUPPORT_OPENGL;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pixfmt = ChoosePixelFormat(ctx->wnd->dc, &pfd);
	if(!pixfmt)
	{
		TRACE("ChoosePixelFormat() failed.");
		goto err_exit;
	}
	memZero(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	if(!DescribePixelFormat(ctx->wnd->dc, pixfmt, pfd.nSize, &pfd))
	{
		TRACE("DescribePixelFormat() failed.");
		goto err_exit;
	}
	if(need_depth && !pfd.cDepthBits)
	{
		TRACE("Renderer doesn't support Z-buffering.");
		goto err_exit;
	}
	if(need_dst_alpha && !pfd.cAlphaBits)
	{
		TRACE("Renderer doesn't support dst alpha.");
		goto err_exit;
	}
	if(need_stencil && !pfd.cStencilBits)
	{
		TRACE("Renderer doesn't support stencil.");
		goto err_exit;
	}
	if(!SetPixelFormat(ctx->wnd->dc, pixfmt, &pfd))
	{
		TRACE("SetPixelFormat() failed.");
		goto err_exit;
	}
	ctx->wgl_context = wglCreateContext(ctx->wnd->dc);
	if(!ctx->wgl_context)
	{
		TRACE("wglCreateContext() failed.");
		goto err_exit;
	}
	if(!wglMakeCurrent(ctx->wnd->dc, ctx->wgl_context))
	{
		TRACE("wglMakeCurrent() failed.");
		goto err_exit;
	}
	ctx->vendor = strDup((char*)glGetString(GL_VENDOR));
	ctx->renderer = strDup((char*)glGetString(GL_RENDERER));
	ctx->version = strDup((char*)glGetString(GL_VERSION));
	ctx->extensions = strDup((char*)glGetString(GL_EXTENSIONS));
#ifdef	AG_VERBOSE
	printf("PFD %d, %s OpenGL on %s ver. %s.\n",
		pixfmt,ctx->vendor,ctx->renderer,ctx->version);
#endif	/* AG_VERBOSE */
	archWinPoll(ctx->wnd, GL_FALSE);
	aspect = (GLfloat)ctx->wnd->width /
		(GLfloat)ctx->wnd->height;
	glViewport(0, 0, ctx->wnd->width, ctx->wnd->height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-aspect, aspect, -1, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	return ctx;

err_exit:;
	wglMakeCurrent(NULL, NULL);
	if(ctx->wgl_context)
		wglDeleteContext(ctx->wgl_context);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx);
	memFree(ctx);
	return NULL;
}

/*
 * Helper routines.
 */

void
agWindowTitle(char *title)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	if(!title)
		return;
	strncpy(ctx->wnd->title,title,MAX_WINDOW_TITLE);
	SetWindowText(ctx->wnd->wnd,title);
}

GLuint
agTickCount(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GetTickCount();
	return GetTickCount() - ctx->wnd->start_time;
}

GLvoid
agSleep(GLuint milliseconds)
{
	Sleep(milliseconds);
}

#else	/* __UNIX__ */

/*
 * Uncomment this to use "official" way of getting pointers to functions
 * of GL 1.2 and above.
 *
 * In fact, I do _not_ recommend to do it.
 *
#define	USE_GLX_GET_PROC_ADDRESS
 */

/*
 * Partial copyright by ~ Id Software ~
 */

static char *local_dpy_name = "localhost:0.0";

#ifdef	ARCH_HAVE_XF86_VIDMODE

static void
x11_set_default_mode(x11_display_connection_p cn)
{
	if(!cn->vidmode_active || !cn->xf86_vidmodes)
		return;
	XF86VidModeLockModeSwitch(cn->dpy, cn->scrnum, False);
	XF86VidModeSwitchToMode(cn->dpy, cn->scrnum, cn->xf86_vidmodes[0]);
	cn->vidmode_active = GL_FALSE;
}

static void
x11_free_mode_lines(x11_display_connection_p cn)
{
	GLuint i;
	if(cn->xf86_vidmodes)
	{
		x11_set_default_mode(cn);
		for(i = 0; i < cn->num_vidmodes; i++)
		{
			if(cn->xf86_vidmodes[i]->privsize)
#if	defined(__cplusplus) || defined(c_plusplus)
				XFree(cn->xf86_vidmodes[i]->c_private);
#else
				XFree(cn->xf86_vidmodes[i]->private);
#endif
		}
		XFree(cn->xf86_vidmodes);
		cn->xf86_vidmodes = NULL;
	}
	if(cn->vidmodes)
	{
		memFree(cn->vidmodes);
		cn->vidmodes = NULL;
	}
	cn->num_vidmodes = 0;
}

static void
x11_get_mode_lines(x11_display_connection_p cn)
{
	GLint i, mj, mn;
	vidmode_p ptr;
	GLuint depth;
	XF86VidModeModeInfo *xm;
	x11_free_mode_lines(cn);
	if(!XF86VidModeQueryVersion(cn->dpy, &mj, &mn))
	{
		TRACE("XFree86-VidModeExtension not found.");
		return;
	}
	XF86VidModeGetAllModeLines(cn->dpy, cn->scrnum,
		(int*)&cn->num_vidmodes, &cn->xf86_vidmodes);
	if(cn->num_vidmodes)
	{
		cn->vidmodes = (vidmode_p)memAllocZ(cn->num_vidmodes * sizeof(vidmode_t));
		depth = DefaultDepth(cn->dpy, cn->scrnum);
		for(i = 0; i < cn->num_vidmodes; i++)
		{
			xm = cn->xf86_vidmodes[i];
			ptr = cn->vidmodes + i;
			ptr->width = xm->hdisplay;
			ptr->height = xm->vdisplay;
			ptr->bpp = depth;
			ptr->refresh =
				(GLuint)((double)xm->dotclock * 1000 / xm->htotal / xm->vtotal);
			if(xm->flags & 32) /* doublescan */
				ptr->refresh = ptr->refresh >> 1;
		}
	}
	return;
}

static void
archWinSetVidMode(arch_window_p aw,GLint mode_num)
{
	vidmode_p ptr;
	if(!aw->visible)
	{
		TRACE("Unable to change vidmode: window invisible.");
		return;
	}
	if(mode_num == -1)
	{
		x11_set_default_mode(aw->cn);
		aw->current_vidmode = -1;
		return;
	}
	ptr = aw->cn->vidmodes + mode_num;
	if(aw->cn->screen_width != ptr->width || aw->cn->screen_height != ptr->height)
	{

		XWarpPointer(aw->cn->dpy, None, aw->wnd, 0, 0, 0, 0,
			ptr->width >> 1, ptr->height >> 1);
		XF86VidModeSwitchToMode(aw->cn->dpy, aw->cn->scrnum,
			aw->cn->xf86_vidmodes[ptr->modeline]);
		XF86VidModeSetViewPort(aw->cn->dpy, aw->cn->scrnum, 0, 0);
		XF86VidModeLockModeSwitch(aw->cn->dpy, aw->cn->scrnum, True);
		aw->current_vidmode = ptr->modeline;
		aw->cn->vidmode_active = GL_TRUE;
		aw->root_width = ptr->width;
		aw->root_height = ptr->height;
	}
}

#else	/* !ARCH_HAVE_XF86_VIDMODE */

static void
archWinSetVidMode(arch_window_p aw,GLint mode_num)
{
}

#endif	/* !ARCH_HAVE_XF86_VIDMODE */

/*
 * Partial copyright by ~ Id Software ~
 */
static x11_display_connection_p
x11_open_connection()
{
	char *lang;
	Pixmap cursormask;
	XGCValues xgc;
	GC gc;
	XColor dummycolour;
	x11_display_connection_p cn = 
		(x11_display_connection_p)memAllocZ(sizeof(x11_display_connection_t));
	if(!cn)
		return NULL;
	lang = getenv("LANG");
	if(lang)
	{
		if(!setlocale(LC_ALL,lang))
		{
			TRACE_1ARG("Could not set locale for LANG=\"%s\".",lang);
		}
	}
	else
	{
		TRACE("LANG not set, skipping setlocale().");
	}
	cn->dpy = XOpenDisplay(NULL);
	if(!cn->dpy)
	{
		TRACE_1ARG("XOpenDisplay(NULL) failed; trying \"%s\".",
			local_dpy_name);
		cn->dpy = XOpenDisplay(local_dpy_name);
		if(!cn->dpy)
		{
			TRACE_1ARG("XOpenDisplay(\"%s\") failed.",
				local_dpy_name);
			memFree(cn);
			return NULL;
		}
	}
	cn->scrnum = DefaultScreen(cn->dpy);
	cn->root = RootWindow(cn->dpy, cn->scrnum);
	XSetLocaleModifiers("");
	cn->wm_protocols_atom = XInternAtom(cn->dpy, "WM_PROTOCOLS", False); 
	cn->wm_delete_window_atom = XInternAtom(cn->dpy, "WM_DELETE_WINDOW", False);
	cursormask = XCreatePixmap(cn->dpy, cn->root, 1, 1, 1);
	xgc.function = GXclear;
	gc = XCreateGC(cn->dpy, cursormask, GCFunction, &xgc);
	XFillRectangle(cn->dpy, cursormask, gc, 0, 0, 1, 1);
	dummycolour.pixel = 0;
	dummycolour.red = 0;
	dummycolour.flags = 04;
	cn->null_cursor = XCreatePixmapCursor(cn->dpy, cursormask,
		cursormask, &dummycolour,&dummycolour, 0, 0);
	XFreePixmap(cn->dpy, cursormask);
	XFreeGC(cn->dpy, gc);
	cn->connection_fd = ConnectionNumber(cn->dpy);
	cn->im = XOpenIM(cn->dpy, NULL, NULL, NULL);
	if(!cn->im)
	{
		TRACE("Unable to open input method.");
	}
	cn->screen_width = WidthOfScreen(DefaultScreenOfDisplay(cn->dpy));
	cn->screen_height = HeightOfScreen(DefaultScreenOfDisplay(cn->dpy));
#ifdef	ARCH_HAVE_XF86_VIDMODE
	x11_get_mode_lines(cn);
#endif	/* ARCH_HAVE_XF86_VIDMODE */
	return cn;
}

static void
x11_close_connection(x11_display_connection_p cn)
{
#ifdef	ARCH_HAVE_XF86_VIDMODE
	x11_free_mode_lines(cn);
#endif	/* ARCH_HAVE_XF86_VIDMODE */
	if(cn->dpy)
	{
		XUngrabKeyboard(cn->dpy, CurrentTime);
		if(cn->null_cursor)
			XFreeCursor(cn->dpy, cn->null_cursor);
		if(cn->im)
			XCloseIM(cn->im);
		XCloseDisplay(cn->dpy);
	}
	memFree(cn);
}

#ifndef XK_KP_Home
#define XK_KP_Home              0xFF95
#endif
#ifndef XK_KP_Left
#define XK_KP_Left              0xFF96
#endif
#ifndef XK_KP_Up
#define XK_KP_Up                0xFF97
#endif
#ifndef XK_KP_Right
#define XK_KP_Right             0xFF98
#endif
#ifndef XK_KP_Down
#define XK_KP_Down              0xFF99
#endif
#ifndef XK_KP_Prior
#define XK_KP_Prior             0xFF9A
#endif
#ifndef XK_KP_Next
#define XK_KP_Next              0xFF9B
#endif
#ifndef XK_KP_End
#define XK_KP_End               0xFF9C
#endif
#ifndef XK_KP_Insert
#define XK_KP_Insert            0xFF9E
#endif
#ifndef XK_KP_Delete
#define XK_KP_Delete            0xFF9F
#endif

static void
X11_init_keycode_to_scan(Display *dpy,GLuint *tab)
{
	KeySym ks;
	GLuint scan;
	GLint i, mincode, maxcode;
	XDisplayKeycodes(dpy, &mincode, &maxcode);
	if(mincode < 0)
		mincode = 0;
	if(maxcode > 255)
		maxcode = 255;
	for(i = 0;i < mincode;i++)
		tab[i] = sc_unknown;
	for(i = maxcode + 1;i < 256;i++)
		tab[i] = sc_unknown;
	for(i = mincode;i <= maxcode;i++)
	{
		ks = XKeycodeToKeysym(dpy,i,0);
		switch(ks)
		{
			case XK_KP_Page_Up: 
			case XK_KP_9: scan = sc_kp_prior; break;
			case XK_Page_Up: scan = sc_prior; break;

			case XK_KP_Page_Down: 
			case XK_KP_3: scan = sc_kp_next; break;
			case XK_Page_Down: scan = sc_next; break;

			case XK_KP_Home:
			case XK_KP_7: scan = sc_kp_home; break;
			case XK_Home: scan = sc_home; break;

			case XK_KP_End:
			case XK_KP_1: scan = sc_kp_end; break;
			case XK_End: scan = sc_end; break;

			case XK_KP_Left:
			case XK_KP_4: scan = sc_kp_left; break;
			case XK_Left: scan = sc_left; break;

			case XK_KP_Right:
			case XK_KP_6: scan = sc_kp_right; break;
			case XK_Right: scan = sc_right; break;

			case XK_KP_Down:
			case XK_KP_2: scan = sc_kp_down; break;
			case XK_Down: scan = sc_down; break;

			case XK_KP_Up:   
			case XK_KP_8: scan = sc_kp_up; break;
			case XK_Up: scan = sc_up; break;

			case XK_Escape: scan = sc_escape; break;

			case XK_KP_Enter: scan = sc_kp_enter; break;
			case XK_Return: scan = sc_enter; break;

			case XK_Tab: scan = sc_tab; break;

			case XK_F1: scan = sc_f1; break;
			case XK_F2: scan = sc_f2; break;
			case XK_F3: scan = sc_f3; break;
			case XK_F4: scan = sc_f4; break;
			case XK_F5: scan = sc_f5; break;
			case XK_F6: scan = sc_f6; break;
			case XK_F7: scan = sc_f7; break;
			case XK_F8: scan = sc_f8; break;
			case XK_F9: scan = sc_f9; break;
			case XK_F10: scan = sc_f10; break;
			case XK_F11: scan = sc_f11; break;
			case XK_F12: scan = sc_f12; break;

			case XK_BackSpace: scan = sc_backspace; break;
			case XK_KP_Delete:
			case XK_KP_Decimal: scan = sc_kp_delete; break;
			case XK_Delete: scan = sc_delete; break;

			case XK_Pause: scan = sc_pause; break;

			case XK_Shift_L: scan = sc_lshift; break;
			case XK_Shift_R: scan = sc_rshift; break;

			case XK_Execute:
			case XK_Control_L: scan = sc_lcontrol; break;
			case XK_Control_R: scan = sc_rcontrol; break;

			case XK_Alt_L:  
			case XK_Meta_L: scan = sc_lalt; break;
			case XK_Alt_R:  
			case XK_Meta_R: scan = sc_ralt; break;

			case XK_KP_Begin: scan = sc_kp_begin; break;

			case XK_KP_Insert:
			case XK_KP_0: scan = sc_kp_insert; break;
			case XK_Insert: scan = sc_insert; break;

			case XK_KP_Multiply: scan = sc_kp_multiply; break;
			case XK_KP_Add: scan = sc_kp_add; break;
			case XK_KP_Subtract: scan = sc_kp_subtract; break;
			case XK_KP_Divide: scan = sc_kp_divide; break;

			case XK_exclam: scan = sc_1; break;
			case XK_at: scan = sc_2; break;
			case XK_numbersign: scan = sc_3; break;
			case XK_dollar: scan = sc_4; break;
			case XK_percent: scan = sc_5; break;
			case XK_asciicircum: scan = sc_6; break;
			case XK_ampersand: scan = sc_7; break;
			case XK_asterisk: scan = sc_8; break;
			case XK_parenleft: scan = sc_9; break;
			case XK_parenright: scan = sc_0; break;
			case XK_twosuperior: scan = sc_grave; break;
			case XK_space:
			case XK_KP_Space: scan = sc_space; break;
		
			case XK_Sys_Req:
			case XK_Print: scan = sc_print; break;
			case XK_Scroll_Lock: scan = sc_scroll_lock; break;
			case XK_Caps_Lock: scan = sc_caps_lock; break;
			case XK_Num_Lock: scan = sc_num_lock; break;
			case XK_Hyper_L:
			case XK_Super_L: scan = sc_lsuper; break;
			case XK_Hyper_R:
			case XK_Super_R: scan = sc_rsuper; break;
			case XK_Menu: scan = sc_menu; break;

			case '`':
			case '~': scan = sc_grave; break;
			case '1': scan = sc_1; break;
			case '2': scan = sc_2; break;
			case '3': scan = sc_3; break;
			case '4': scan = sc_4; break;
			case '5': scan = sc_5; break;
			case '6': scan = sc_6; break;
			case '7': scan = sc_7; break;
			case '8': scan = sc_8; break;
			case '9': scan = sc_9; break;
			case '0': scan = sc_0; break;
			case '-':
			case '_': scan = sc_minus; break;
			case '=':
			case '+': scan = sc_equal; break;
			case '\\':
			case '|': scan = sc_backslash; break;
			case 'q':
			case 'Q': scan = sc_q; break;
			case 'w':
			case 'W': scan = sc_w; break;
			case 'e':
			case 'E': scan = sc_e; break;
			case 'r':
			case 'R': scan = sc_r; break;
			case 't':
			case 'T': scan = sc_t; break;
			case 'y':
			case 'Y': scan = sc_y; break;
			case 'u':
			case 'U': scan = sc_u; break;
			case 'i':
			case 'I': scan = sc_i; break;
			case 'o':
			case 'O': scan = sc_o; break;
			case 'p':
			case 'P': scan = sc_p; break;
			case '[':
			case '{': scan = sc_lbracket; break;
			case ']':
			case '}': scan = sc_rbracket; break;
			case 'a':
			case 'A': scan = sc_a; break;
			case 's':
			case 'S': scan = sc_s; break;
			case 'd':
			case 'D': scan = sc_d; break;
			case 'f':
			case 'F': scan = sc_f; break;
			case 'g':
			case 'G': scan = sc_g; break;
			case 'h':
			case 'H': scan = sc_h; break;
			case 'j':
			case 'J': scan = sc_j; break;
			case 'k':
			case 'K': scan = sc_k; break;
			case 'l':
			case 'L': scan = sc_l; break;
			case ';':
			case ':': scan = sc_semicolon; break;
			case '\'':
			case '"': scan = sc_apostrophe; break;
			case 'z':
			case 'Z': scan = sc_z; break;
			case 'x':
			case 'X': scan = sc_x; break;
			case 'c':
			case 'C': scan = sc_c; break;
			case 'v':
			case 'V': scan = sc_v; break;
			case 'b':
			case 'B': scan = sc_b; break;
			case 'n':
			case 'N': scan = sc_n; break;
			case 'm':
			case 'M': scan = sc_m; break;
			case ',':
			case '<': scan = sc_comma; break;
			case '.':
			case '>': scan = sc_period; break;
			case '/':
			case '?': scan = sc_slash; break;
			default: scan = sc_unknown;
		}
		tab[i] = scan;
	}
}

static arch_window_p
archWinOpenNP(x11_display_connection_p cn,XVisualInfo *vi,
	char *title,GLuint left,GLuint top,
	GLuint width,GLuint height,GLboolean caption,GLboolean allow_resize)
{
	arch_window_p aw;
	unsigned long mask;
	XSizeHints sh;
	XSetWindowAttributes attr;
	if (width == 0)
		width = cn->screen_width;
	if (height == 0)
		height = cn->screen_height;
	memZero(&attr,sizeof(XSetWindowAttributes));
	attr.event_mask =
		StructureNotifyMask | ExposureMask |
		KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask | VisibilityChangeMask | FocusChangeMask;
	attr.do_not_propagate_mask = KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask;
	attr.background_pixmap = None;
	attr.background_pixel = None;
	attr.border_pixmap = None;
	attr.border_pixel = None;
	attr.backing_pixel = 0;
	attr.backing_store = NotUseful;
	attr.save_under = False;
	mask = CWBackPixmap | CWBackPixel | 
		CWBorderPixmap | CWBorderPixel | 
		CWBackingPixel | CWBackingStore |
		CWEventMask | CWDontPropagate | CWSaveUnder;
	mask |= CWColormap;
	memZero(&sh, sizeof(sh));
	sh.x = left;
	sh.y = top;
	sh.width = sh.min_width = sh.max_width = sh.base_width = width;
	sh.height = sh.min_height = sh.max_height = sh.base_height = height;
	sh.flags = USPosition | PPosition | PSize;
	if(!allow_resize)
	{
		sh.flags |= PBaseSize | PMinSize | PMaxSize;
		if(!caption)
		{
			mask |= CWOverrideRedirect;
			attr.override_redirect = True;
		}
	}
	else if(!caption)
	{
		TRACE("Can't create resizeable window in fullscreen mode.");
		return NULL;
	}
	aw = (arch_window_p)memAllocZ(sizeof(arch_window_t));
	if(!aw)
	{
		return NULL;
	}
	INIT_COMMON_WINDOW_FIELDS(aw);
	attr.colormap = XCreateColormap(cn->dpy, cn->root, vi->visual, AllocNone);
	aw->cn = cn;
	X11_init_keycode_to_scan(cn->dpy, aw->keycode_to_scan);
	strncpy(aw->title,title,MAX_WINDOW_TITLE);
	aw->x = left;
	aw->y = top;
	aw->width = width;
	aw->height = height;
	aw->parent = cn->root;
	aw->event_mask = attr.event_mask;
	aw->wnd = XCreateWindow(cn->dpy, aw->parent, left, top, width, height, 0,
		vi ? vi->depth : DefaultDepth(cn->dpy, cn->scrnum),
		InputOutput,
		vi ? vi->visual : DefaultVisual(cn->dpy, cn->scrnum), mask, &attr);
	if(!aw->wnd)
	{
		TRACE("XCreateWindow() failed.");
		archWinClose(aw);
		return NULL;
	}
	if(cn->im)
		aw->ic = XCreateIC(cn->im, XNInputStyle,
			XIMPreeditNothing | XIMStatusNothing,
			XNClientWindow, aw->wnd, XNFocusWindow, aw->wnd, NULL);
	else
		aw->ic = 0;
	XStoreName(cn->dpy, aw->wnd, aw->title);
	XSetWMNormalHints(cn->dpy, aw->wnd, &sh);
	XSetWMProtocols(cn->dpy, aw->wnd, &cn->wm_delete_window_atom, 1);
	XMapRaised(cn->dpy, aw->wnd);
	if(!caption)
		XSetInputFocus(cn->dpy, aw->wnd, RevertToParent, CurrentTime);
	aw->auto_repeat_disabled = GL_FALSE;
	aw->cmap = attr.colormap;
	return aw;
}

#define	NUM_XK_SYMS		1264

static struct ks2unitab_s
{
	unsigned	ks, uni;
}
	ks2unitab[NUM_XK_SYMS] =
{
	{0x000001A1,0x0104},{0x000001A2,0x02D8},{0x000001A3,0x0141},{0x000001A5,0x013D},
	{0x000001A6,0x015A},{0x000001A9,0x0160},{0x000001AA,0x015E},{0x000001AB,0x0164},
	{0x000001AC,0x0179},{0x000001AE,0x017D},{0x000001AF,0x017B},{0x000001B1,0x0105},
	{0x000001B2,0x02DB},{0x000001B3,0x0142},{0x000001B5,0x013E},{0x000001B6,0x015B},
	{0x000001B7,0x02C7},{0x000001B9,0x0161},{0x000001BA,0x015F},{0x000001BB,0x0165},
	{0x000001BC,0x017A},{0x000001BD,0x02DD},{0x000001BE,0x017E},{0x000001BF,0x017C},
	{0x000001C0,0x0154},{0x000001C3,0x0102},{0x000001C5,0x0139},{0x000001C6,0x0106},
	{0x000001C8,0x010C},{0x000001CA,0x0118},{0x000001CC,0x011A},{0x000001CF,0x010E},
	{0x000001D0,0x0110},{0x000001D1,0x0143},{0x000001D2,0x0147},{0x000001D5,0x0150},
	{0x000001D8,0x0158},{0x000001D9,0x016E},{0x000001DB,0x0170},{0x000001DE,0x0162},
	{0x000001E0,0x0155},{0x000001E3,0x0103},{0x000001E5,0x013A},{0x000001E6,0x0107},
	{0x000001E8,0x010D},{0x000001EA,0x0119},{0x000001EC,0x011B},{0x000001EF,0x010F},
	{0x000001F0,0x0111},{0x000001F1,0x0144},{0x000001F2,0x0148},{0x000001F5,0x0151},
	{0x000001F8,0x0159},{0x000001F9,0x016F},{0x000001FB,0x0171},{0x000001FE,0x0163},
	{0x000001FF,0x02D9},{0x000002A1,0x0126},{0x000002A6,0x0124},{0x000002A9,0x0130},
	{0x000002AB,0x011E},{0x000002AC,0x0134},{0x000002B1,0x0127},{0x000002B6,0x0125},
	{0x000002B9,0x0131},{0x000002BB,0x011F},{0x000002BC,0x0135},{0x000002C5,0x010A},
	{0x000002C6,0x0108},{0x000002D5,0x0120},{0x000002D8,0x011C},{0x000002DD,0x016C},
	{0x000002DE,0x015C},{0x000002E5,0x010B},{0x000002E6,0x0109},{0x000002F5,0x0121},
	{0x000002F8,0x011D},{0x000002FD,0x016D},{0x000002FE,0x015D},{0x000003A2,0x0138},
	{0x000003A3,0x0156},{0x000003A5,0x0128},{0x000003A6,0x013B},{0x000003AA,0x0112},
	{0x000003AB,0x0122},{0x000003AC,0x0166},{0x000003B3,0x0157},{0x000003B5,0x0129},
	{0x000003B6,0x013C},{0x000003BA,0x0113},{0x000003BB,0x0123},{0x000003BC,0x0167},
	{0x000003BD,0x014A},{0x000003BF,0x014B},{0x000003C0,0x0100},{0x000003C7,0x012E},
	{0x000003CC,0x0116},{0x000003CF,0x012A},{0x000003D1,0x0145},{0x000003D2,0x014C},
	{0x000003D3,0x0136},{0x000003D9,0x0172},{0x000003DD,0x0168},{0x000003DE,0x016A},
	{0x000003E0,0x0101},{0x000003E7,0x012F},{0x000003EC,0x0117},{0x000003EF,0x012B},
	{0x000003F1,0x0146},{0x000003F2,0x014D},{0x000003F3,0x0137},{0x000003F9,0x0173},
	{0x000003FD,0x0169},{0x000003FE,0x016B},{0x0000047E,0x203E},{0x000004A1,0x3002},
	{0x000004A2,0x300C},{0x000004A3,0x300D},{0x000004A4,0x3001},{0x000004A5,0x30FB},
	{0x000004A6,0x30F2},{0x000004A7,0x30A1},{0x000004A8,0x30A3},{0x000004A9,0x30A5},
	{0x000004AA,0x30A7},{0x000004AB,0x30A9},{0x000004AC,0x30E3},{0x000004AD,0x30E5},
	{0x000004AE,0x30E7},{0x000004AF,0x30C3},{0x000004B0,0x30FC},{0x000004B1,0x30A2},
	{0x000004B2,0x30A4},{0x000004B3,0x30A6},{0x000004B4,0x30A8},{0x000004B5,0x30AA},
	{0x000004B6,0x30AB},{0x000004B7,0x30AD},{0x000004B8,0x30AF},{0x000004B9,0x30B1},
	{0x000004BA,0x30B3},{0x000004BB,0x30B5},{0x000004BC,0x30B7},{0x000004BD,0x30B9},
	{0x000004BE,0x30BB},{0x000004BF,0x30BD},{0x000004C0,0x30BF},{0x000004C1,0x30C1},
	{0x000004C2,0x30C4},{0x000004C3,0x30C6},{0x000004C4,0x30C8},{0x000004C5,0x30CA},
	{0x000004C6,0x30CB},{0x000004C7,0x30CC},{0x000004C8,0x30CD},{0x000004C9,0x30CE},
	{0x000004CA,0x30CF},{0x000004CB,0x30D2},{0x000004CC,0x30D5},{0x000004CD,0x30D8},
	{0x000004CE,0x30DB},{0x000004CF,0x30DE},{0x000004D0,0x30DF},{0x000004D1,0x30E0},
	{0x000004D2,0x30E1},{0x000004D3,0x30E2},{0x000004D4,0x30E4},{0x000004D5,0x30E6},
	{0x000004D6,0x30E8},{0x000004D7,0x30E9},{0x000004D8,0x30EA},{0x000004D9,0x30EB},
	{0x000004DA,0x30EC},{0x000004DB,0x30ED},{0x000004DC,0x30EF},{0x000004DD,0x30F3},
	{0x000004DE,0x309B},{0x000004DF,0x309C},{0x000005AC,0x060C},{0x000005BB,0x061B},
	{0x000005BF,0x061F},{0x000005C1,0x0621},{0x000005C2,0x0622},{0x000005C3,0x0623},
	{0x000005C4,0x0624},{0x000005C5,0x0625},{0x000005C6,0x0626},{0x000005C7,0x0627},
	{0x000005C8,0x0628},{0x000005C9,0x0629},{0x000005CA,0x062A},{0x000005CB,0x062B},
	{0x000005CC,0x062C},{0x000005CD,0x062D},{0x000005CE,0x062E},{0x000005CF,0x062F},
	{0x000005D0,0x0630},{0x000005D1,0x0631},{0x000005D2,0x0632},{0x000005D3,0x0633},
	{0x000005D4,0x0634},{0x000005D5,0x0635},{0x000005D6,0x0636},{0x000005D7,0x0637},
	{0x000005D8,0x0638},{0x000005D9,0x0639},{0x000005DA,0x063A},{0x000005E0,0x0640},
	{0x000005E1,0x0641},{0x000005E2,0x0642},{0x000005E3,0x0643},{0x000005E4,0x0644},
	{0x000005E5,0x0645},{0x000005E6,0x0646},{0x000005E7,0x0647},{0x000005E8,0x0648},
	{0x000005E9,0x0649},{0x000005EA,0x064A},{0x000005EB,0x064B},{0x000005EC,0x064C},
	{0x000005ED,0x064D},{0x000005EE,0x064E},{0x000005EF,0x064F},{0x000005F0,0x0650},
	{0x000005F1,0x0651},{0x000005F2,0x0652},{0x000006A1,0x0452},{0x000006A2,0x0453},
	{0x000006A3,0x0451},{0x000006A4,0x0454},{0x000006A5,0x0455},{0x000006A6,0x0456},
	{0x000006A7,0x0457},{0x000006A8,0x0458},{0x000006A9,0x0459},{0x000006AA,0x045A},
	{0x000006AB,0x045B},{0x000006AC,0x045C},{0x000006AD,0x0491},{0x000006AE,0x045E},
	{0x000006AF,0x045F},{0x000006B0,0x2116},{0x000006B1,0x0402},{0x000006B2,0x0403},
	{0x000006B3,0x0401},{0x000006B4,0x0404},{0x000006B5,0x0405},{0x000006B6,0x0406},
	{0x000006B7,0x0407},{0x000006B8,0x0408},{0x000006B9,0x0409},{0x000006BA,0x040A},
	{0x000006BB,0x040B},{0x000006BC,0x040C},{0x000006BD,0x0490},{0x000006BE,0x040E},
	{0x000006BF,0x040F},{0x000006C0,0x044E},{0x000006C1,0x0430},{0x000006C2,0x0431},
	{0x000006C3,0x0446},{0x000006C4,0x0434},{0x000006C5,0x0435},{0x000006C6,0x0444},
	{0x000006C7,0x0433},{0x000006C8,0x0445},{0x000006C9,0x0438},{0x000006CA,0x0439},
	{0x000006CB,0x043A},{0x000006CC,0x043B},{0x000006CD,0x043C},{0x000006CE,0x043D},
	{0x000006CF,0x043E},{0x000006D0,0x043F},{0x000006D1,0x044F},{0x000006D2,0x0440},
	{0x000006D3,0x0441},{0x000006D4,0x0442},{0x000006D5,0x0443},{0x000006D6,0x0436},
	{0x000006D7,0x0432},{0x000006D8,0x044C},{0x000006D9,0x044B},{0x000006DA,0x0437},
	{0x000006DB,0x0448},{0x000006DC,0x044D},{0x000006DD,0x0449},{0x000006DE,0x0447},
	{0x000006DF,0x044A},{0x000006E0,0x042E},{0x000006E1,0x0410},{0x000006E2,0x0411},
	{0x000006E3,0x0426},{0x000006E4,0x0414},{0x000006E5,0x0415},{0x000006E6,0x0424},
	{0x000006E7,0x0413},{0x000006E8,0x0425},{0x000006E9,0x0418},{0x000006EA,0x0419},
	{0x000006EB,0x041A},{0x000006EC,0x041B},{0x000006ED,0x041C},{0x000006EE,0x041D},
	{0x000006EF,0x041E},{0x000006F0,0x041F},{0x000006F1,0x042F},{0x000006F2,0x0420},
	{0x000006F3,0x0421},{0x000006F4,0x0422},{0x000006F5,0x0423},{0x000006F6,0x0416},
	{0x000006F7,0x0412},{0x000006F8,0x042C},{0x000006F9,0x042B},{0x000006FA,0x0417},
	{0x000006FB,0x0428},{0x000006FC,0x042D},{0x000006FD,0x0429},{0x000006FE,0x0427},
	{0x000006FF,0x042A},{0x000007A1,0x0386},{0x000007A2,0x0388},{0x000007A3,0x0389},
	{0x000007A4,0x038A},{0x000007A5,0x03AA},{0x000007A7,0x038C},{0x000007A8,0x038E},
	{0x000007A9,0x03AB},{0x000007AB,0x038F},{0x000007AE,0x0385},{0x000007AF,0x2015},
	{0x000007B1,0x03AC},{0x000007B2,0x03AD},{0x000007B3,0x03AE},{0x000007B4,0x03AF},
	{0x000007B5,0x03CA},{0x000007B6,0x0390},{0x000007B7,0x03CC},{0x000007B8,0x03CD},
	{0x000007B9,0x03CB},{0x000007BA,0x03B0},{0x000007BB,0x03CE},{0x000007C1,0x0391},
	{0x000007C2,0x0392},{0x000007C3,0x0393},{0x000007C4,0x0394},{0x000007C5,0x0395},
	{0x000007C6,0x0396},{0x000007C7,0x0397},{0x000007C8,0x0398},{0x000007C9,0x0399},
	{0x000007CA,0x039A},{0x000007CB,0x039B},{0x000007CB,0x039B},{0x000007CC,0x039C},
	{0x000007CD,0x039D},{0x000007CE,0x039E},{0x000007CF,0x039F},{0x000007D0,0x03A0},
	{0x000007D1,0x03A1},{0x000007D2,0x03A3},{0x000007D4,0x03A4},{0x000007D5,0x03A5},
	{0x000007D6,0x03A6},{0x000007D7,0x03A7},{0x000007D8,0x03A8},{0x000007D9,0x03A9},
	{0x000007E1,0x03B1},{0x000007E2,0x03B2},{0x000007E3,0x03B3},{0x000007E4,0x03B4},
	{0x000007E5,0x03B5},{0x000007E6,0x03B6},{0x000007E7,0x03B7},{0x000007E8,0x03B8},
	{0x000007E9,0x03B9},{0x000007EA,0x03BA},{0x000007EB,0x03BB},{0x000007EB,0x03BB},
	{0x000007EC,0x03BC},{0x000007ED,0x03BD},{0x000007EE,0x03BE},{0x000007EF,0x03BF},
	{0x000007F0,0x03C0},{0x000007F1,0x03C1},{0x000007F2,0x03C3},{0x000007F3,0x03C2},
	{0x000007F4,0x03C4},{0x000007F5,0x03C5},{0x000007F6,0x03C6},{0x000007F7,0x03C7},
	{0x000007F8,0x03C8},{0x000007F9,0x03C9},{0x000008A1,0x23B7},{0x000008A4,0x2320},
	{0x000008A5,0x2321},{0x000008A7,0x23A1},{0x000008A8,0x23A3},{0x000008A9,0x23A4},
	{0x000008AA,0x23A6},{0x000008AB,0x239B},{0x000008AC,0x239D},{0x000008AD,0x239E},
	{0x000008AE,0x23A0},{0x000008AF,0x23A8},{0x000008B0,0x23AC},{0x000008BD,0x2260},
	{0x000008BE,0x2265},{0x000008BF,0x222B},{0x000008C0,0x2234},{0x000008C1,0x221D},
	{0x000008C2,0x221E},{0x000008C5,0x2207},{0x000008C8,0x223C},{0x000008C9,0x2243},
	{0x000008CD,0x21D4},{0x000008CE,0x21D2},{0x000008CF,0x2261},{0x000008D6,0x221A},
	{0x000008DA,0x2282},{0x000008DB,0x2283},{0x000008DC,0x2229},{0x000008DD,0x222A},
	{0x000008DE,0x2227},{0x000008DF,0x2228},{0x000008EF,0x2202},{0x000008F6,0x0192},
	{0x000008FB,0x2190},{0x000008FC,0x2191},{0x000008FD,0x2192},{0x000008FE,0x2193},
	{0x000009E1,0x2592},{0x000009E2,0x2409},{0x000009E3,0x240C},{0x000009E4,0x240D},
	{0x000009E5,0x240A},{0x000009E8,0x2424},{0x000009E9,0x240B},{0x000009EA,0x2518},
	{0x000009EB,0x2510},{0x000009EC,0x250C},{0x000009ED,0x2514},{0x000009EE,0x253C},
	{0x000009EF,0x23BA},{0x000009F0,0x23BB},{0x000009F1,0x2500},{0x000009F2,0x23BC},
	{0x000009F3,0x23BD},{0x000009F4,0x251C},{0x000009F5,0x2524},{0x000009F6,0x2534},
	{0x000009F7,0x252C},{0x000009F8,0x2502},{0x00000AA1,0x2003},{0x00000AA2,0x2002},
	{0x00000AA3,0x2004},{0x00000AA4,0x2005},{0x00000AA5,0x2007},{0x00000AA6,0x2008},
	{0x00000AA7,0x2009},{0x00000AA8,0x200A},{0x00000AA9,0x2014},{0x00000AAA,0x2013},
	{0x00000AAE,0x2026},{0x00000AAF,0x2025},{0x00000AB0,0x2153},{0x00000AB1,0x2154},
	{0x00000AB2,0x2155},{0x00000AB3,0x2156},{0x00000AB4,0x2157},{0x00000AB5,0x2158},
	{0x00000AB6,0x2159},{0x00000AB7,0x215A},{0x00000AB8,0x2105},{0x00000ABB,0x2012},
	{0x00000AC4,0x215C},{0x00000AC5,0x215D},{0x00000AC6,0x215E},{0x00000AC9,0x2122},
	{0x00000AD0,0x2018},{0x00000AD1,0x2019},{0x00000AD2,0x201C},{0x00000AD3,0x201D},
	{0x00000AD4,0x211E},{0x00000AD6,0x2032},{0x00000AD7,0x2033},{0x00000AD9,0x271D},
	{0x00000AEC,0x2663},{0x00000AED,0x2666},{0x00000AEE,0x2665},{0x00000AF0,0x2720},
	{0x00000AF1,0x2020},{0x00000AF2,0x2021},{0x00000AF3,0x2713},{0x00000AF4,0x2717},
	{0x00000AF5,0x266F},{0x00000AF6,0x266D},{0x00000AF7,0x2642},{0x00000AF8,0x2640},
	{0x00000AF9,0x260E},{0x00000AFA,0x2315},{0x00000AFB,0x2117},{0x00000AFC,0x2038},
	{0x00000AFD,0x201A},{0x00000AFE,0x201E},{0x00000BC2,0x22A5},{0x00000BC4,0x230A},
	{0x00000BCA,0x2218},{0x00000BCC,0x2395},{0x00000BCE,0x22A4},{0x00000BCF,0x25CB},
	{0x00000BD3,0x2308},{0x00000BD6,0x0000},{0x00000BDC,0x22A2},{0x00000BFC,0x22A3},
	{0x00000CDF,0x2017},{0x00000CE0,0x05D0},{0x00000CE1,0x05D1},{0x00000CE2,0x05D2},
	{0x00000CE3,0x05D3},{0x00000CE4,0x05D4},{0x00000CE5,0x05D5},{0x00000CE6,0x05D6},
	{0x00000CE7,0x05D7},{0x00000CE8,0x05D8},{0x00000CE9,0x05D9},{0x00000CEA,0x05DA},
	{0x00000CEB,0x05DB},{0x00000CEC,0x05DC},{0x00000CED,0x05DD},{0x00000CEE,0x05DE},
	{0x00000CEF,0x05DF},{0x00000CF0,0x05E0},{0x00000CF1,0x05E1},{0x00000CF2,0x05E2},
	{0x00000CF3,0x05E3},{0x00000CF4,0x05E4},{0x00000CF5,0x05E5},{0x00000CF6,0x05E6},
	{0x00000CF7,0x05E7},{0x00000CF8,0x05E8},{0x00000CF9,0x05E9},{0x00000CFA,0x05EA},
	{0x00000DA1,0x0E01},{0x00000DA2,0x0E02},{0x00000DA3,0x0E03},{0x00000DA4,0x0E04},
	{0x00000DA5,0x0E05},{0x00000DA6,0x0E06},{0x00000DA7,0x0E07},{0x00000DA8,0x0E08},
	{0x00000DA9,0x0E09},{0x00000DAA,0x0E0A},{0x00000DAB,0x0E0B},{0x00000DAC,0x0E0C},
	{0x00000DAD,0x0E0D},{0x00000DAE,0x0E0E},{0x00000DAF,0x0E0F},{0x00000DB0,0x0E10},
	{0x00000DB1,0x0E11},{0x00000DB2,0x0E12},{0x00000DB3,0x0E13},{0x00000DB4,0x0E14},
	{0x00000DB5,0x0E15},{0x00000DB6,0x0E16},{0x00000DB7,0x0E17},{0x00000DB8,0x0E18},
	{0x00000DB9,0x0E19},{0x00000DBA,0x0E1A},{0x00000DBB,0x0E1B},{0x00000DBC,0x0E1C},
	{0x00000DBD,0x0E1D},{0x00000DBE,0x0E1E},{0x00000DBF,0x0E1F},{0x00000DC0,0x0E20},
	{0x00000DC1,0x0E21},{0x00000DC2,0x0E22},{0x00000DC3,0x0E23},{0x00000DC4,0x0E24},
	{0x00000DC5,0x0E25},{0x00000DC6,0x0E26},{0x00000DC7,0x0E27},{0x00000DC8,0x0E28},
	{0x00000DC9,0x0E29},{0x00000DCA,0x0E2A},{0x00000DCB,0x0E2B},{0x00000DCC,0x0E2C},
	{0x00000DCD,0x0E2D},{0x00000DCE,0x0E2E},{0x00000DCF,0x0E2F},{0x00000DD0,0x0E30},
	{0x00000DD1,0x0E31},{0x00000DD2,0x0E32},{0x00000DD3,0x0E33},{0x00000DD4,0x0E34},
	{0x00000DD5,0x0E35},{0x00000DD6,0x0E36},{0x00000DD7,0x0E37},{0x00000DD8,0x0E38},
	{0x00000DD9,0x0E39},{0x00000DDA,0x0E3A},{0x00000DE0,0x0E40},{0x00000DE1,0x0E41},
	{0x00000DE2,0x0E42},{0x00000DE3,0x0E43},{0x00000DE4,0x0E44},{0x00000DE5,0x0E45},
	{0x00000DE6,0x0E46},{0x00000DE7,0x0E47},{0x00000DE8,0x0E48},{0x00000DE9,0x0E49},
	{0x00000DEA,0x0E4A},{0x00000DEB,0x0E4B},{0x00000DEC,0x0E4C},{0x00000DED,0x0E4D},
	{0x00000DF0,0x0E50},{0x00000DF1,0x0E51},{0x00000DF2,0x0E52},{0x00000DF3,0x0E53},
	{0x00000DF4,0x0E54},{0x00000DF5,0x0E55},{0x00000DF6,0x0E56},{0x00000DF7,0x0E57},
	{0x00000DF8,0x0E58},{0x00000DF9,0x0E59},{0x000013BC,0x0152},{0x000013BD,0x0153},
	{0x000013BE,0x0178},{0x000020AC,0x20AC},{0x0100012C,0x012C},{0x0100012D,0x012D},
	{0x01000174,0x0174},{0x01000175,0x0175},{0x01000176,0x0176},{0x01000177,0x0177},
	{0x0100018F,0x018F},{0x0100019F,0x019F},{0x010001A0,0x01A0},{0x010001A1,0x01A1},
	{0x010001AF,0x01AF},{0x010001B0,0x01B0},{0x010001B5,0x01B5},{0x010001B6,0x01B6},
	{0x010001D1,0x01D2},{0x010001D2,0x01D2},{0x010001E6,0x01E6},{0x010001E7,0x01E7},
	{0x01000259,0x0259},{0x01000275,0x0275},{0x01000492,0x0492},{0x01000493,0x0493},
	{0x01000496,0x0496},{0x01000497,0x0497},{0x0100049A,0x049A},{0x0100049B,0x049B},
	{0x0100049C,0x049C},{0x0100049D,0x049D},{0x010004A2,0x04A2},{0x010004A3,0x04A3},
	{0x010004AE,0x04AE},{0x010004AF,0x04AF},{0x010004B0,0x04B0},{0x010004B1,0x04B1},
	{0x010004B2,0x04B2},{0x010004B3,0x04B3},{0x010004B6,0x04B6},{0x010004B7,0x04B7},
	{0x010004B8,0x04B8},{0x010004B9,0x04B9},{0x010004BA,0x04BA},{0x010004BB,0x04BB},
	{0x010004D8,0x04D8},{0x010004D9,0x04D9},{0x010004E2,0x04E2},{0x010004E3,0x04E3},
	{0x010004E8,0x04E8},{0x010004E9,0x04E9},{0x010004EE,0x04EE},{0x010004EF,0x04EF},
	{0x01000531,0x0531},{0x01000532,0x0532},{0x01000533,0x0533},{0x01000534,0x0534},
	{0x01000535,0x0535},{0x01000536,0x0536},{0x01000537,0x0537},{0x01000538,0x0538},
	{0x01000539,0x0539},{0x0100053A,0x053A},{0x0100053B,0x053B},{0x0100053C,0x053C},
	{0x0100053D,0x053D},{0x0100053E,0x053E},{0x0100053F,0x053F},{0x01000540,0x0540},
	{0x01000541,0x0541},{0x01000542,0x0542},{0x01000543,0x0543},{0x01000544,0x0544},
	{0x01000545,0x0545},{0x01000546,0x0546},{0x01000547,0x0547},{0x01000548,0x0548},
	{0x01000549,0x0549},{0x0100054A,0x054A},{0x0100054B,0x054B},{0x0100054C,0x054C},
	{0x0100054D,0x054D},{0x0100054E,0x054E},{0x0100054F,0x054F},{0x01000550,0x0550},
	{0x01000551,0x0551},{0x01000552,0x0552},{0x01000553,0x0553},{0x01000554,0x0554},
	{0x01000555,0x0555},{0x01000556,0x0556},{0x0100055A,0x055A},{0x0100055B,0x055B},
	{0x0100055B,0x055B},{0x0100055C,0x055C},{0x0100055C,0x055C},{0x0100055D,0x055D},
	{0x0100055D,0x055D},{0x0100055E,0x055E},{0x0100055E,0x055E},{0x01000561,0x0561},
	{0x01000562,0x0562},{0x01000563,0x0563},{0x01000564,0x0564},{0x01000565,0x0565},
	{0x01000566,0x0566},{0x01000567,0x0567},{0x01000568,0x0568},{0x01000569,0x0569},
	{0x0100056A,0x056A},{0x0100056B,0x056B},{0x0100056C,0x056C},{0x0100056D,0x056D},
	{0x0100056E,0x056E},{0x0100056F,0x056F},{0x01000570,0x0570},{0x01000571,0x0571},
	{0x01000572,0x0572},{0x01000573,0x0573},{0x01000574,0x0574},{0x01000575,0x0575},
	{0x01000576,0x0576},{0x01000577,0x0577},{0x01000578,0x0578},{0x01000579,0x0579},
	{0x0100057A,0x057A},{0x0100057B,0x057B},{0x0100057C,0x057C},{0x0100057D,0x057D},
	{0x0100057E,0x057E},{0x0100057F,0x057F},{0x01000580,0x0580},{0x01000581,0x0581},
	{0x01000582,0x0582},{0x01000583,0x0583},{0x01000584,0x0584},{0x01000585,0x0585},
	{0x01000586,0x0586},{0x01000587,0x0587},{0x01000589,0x0589},{0x01000589,0x0589},
	{0x0100058A,0x058A},{0x0100058A,0x058A},{0x01000653,0x0653},{0x01000654,0x0654},
	{0x01000655,0x0655},{0x01000660,0x0660},{0x01000661,0x0661},{0x01000662,0x0662},
	{0x01000663,0x0663},{0x01000664,0x0664},{0x01000665,0x0665},{0x01000666,0x0666},
	{0x01000667,0x0667},{0x01000668,0x0668},{0x01000669,0x0669},{0x0100066A,0x066A},
	{0x01000670,0x0670},{0x01000679,0x0679},{0x0100067E,0x067E},{0x01000686,0x0686},
	{0x01000688,0x0688},{0x01000691,0x0691},{0x01000698,0x0698},{0x010006A4,0x06A4},
	{0x010006A9,0x06A9},{0x010006AF,0x06AF},{0x010006BA,0x06BA},{0x010006BE,0x06BE},
	{0x010006C1,0x06C1},{0x010006CC,0x06CC},{0x010006CC,0x06CC},{0x010006D2,0x06D2},
	{0x010006D4,0x06D4},{0x010006F0,0x06F0},{0x010006F1,0x06F1},{0x010006F2,0x06F2},
	{0x010006F3,0x06F3},{0x010006F4,0x06F4},{0x010006F5,0x06F5},{0x010006F6,0x06F6},
	{0x010006F7,0x06F7},{0x010006F8,0x06F8},{0x010006F9,0x06F9},{0x010010D0,0x10D0},
	{0x010010D1,0x10D1},{0x010010D2,0x10D2},{0x010010D3,0x10D3},{0x010010D4,0x10D4},
	{0x010010D5,0x10D5},{0x010010D6,0x10D6},{0x010010D7,0x10D7},{0x010010D8,0x10D8},
	{0x010010D9,0x10D9},{0x010010DA,0x10DA},{0x010010DB,0x10DB},{0x010010DC,0x10DC},
	{0x010010DD,0x10DD},{0x010010DE,0x10DE},{0x010010DF,0x10DF},{0x010010E0,0x10E0},
	{0x010010E1,0x10E1},{0x010010E2,0x10E2},{0x010010E3,0x10E3},{0x010010E4,0x10E4},
	{0x010010E5,0x10E5},{0x010010E6,0x10E6},{0x010010E7,0x10E7},{0x010010E8,0x10E8},
	{0x010010E9,0x10E9},{0x010010EA,0x10EA},{0x010010EB,0x10EB},{0x010010EC,0x10EC},
	{0x010010ED,0x10ED},{0x010010EE,0x10EE},{0x010010EF,0x10EF},{0x010010F0,0x10F0},
	{0x010010F1,0x10F1},{0x010010F2,0x10F2},{0x010010F3,0x10F3},{0x010010F4,0x10F4},
	{0x010010F5,0x10F5},{0x010010F6,0x10F6},{0x01001E02,0x1E02},{0x01001E03,0x1E03},
	{0x01001E0A,0x1E0A},{0x01001E0B,0x1E0B},{0x01001E1E,0x1E1E},{0x01001E1F,0x1E1F},
	{0x01001E36,0x1E36},{0x01001E37,0x1E37},{0x01001E40,0x1E40},{0x01001E41,0x1E41},
	{0x01001E56,0x1E56},{0x01001E57,0x1E57},{0x01001E60,0x1E60},{0x01001E61,0x1E61},
	{0x01001E6A,0x1E6A},{0x01001E6B,0x1E6B},{0x01001E80,0x1E80},{0x01001E81,0x1E81},
	{0x01001E82,0x1E82},{0x01001E83,0x1E83},{0x01001E84,0x1E84},{0x01001E85,0x1E85},
	{0x01001E8A,0x1E8A},{0x01001E8B,0x1E8B},{0x01001EA0,0x1EA0},{0x01001EA1,0x1EA1},
	{0x01001EA2,0x1EA2},{0x01001EA3,0x1EA3},{0x01001EA4,0x1EA4},{0x01001EA5,0x1EA5},
	{0x01001EA6,0x1EA6},{0x01001EA7,0x1EA7},{0x01001EA8,0x1EA8},{0x01001EA9,0x1EA9},
	{0x01001EAA,0x1EAA},{0x01001EAB,0x1EAB},{0x01001EAC,0x1EAC},{0x01001EAD,0x1EAD},
	{0x01001EAE,0x1EAE},{0x01001EAF,0x1EAF},{0x01001EB0,0x1EB0},{0x01001EB1,0x1EB1},
	{0x01001EB2,0x1EB2},{0x01001EB3,0x1EB3},{0x01001EB4,0x1EB4},{0x01001EB5,0x1EB5},
	{0x01001EB6,0x1EB6},{0x01001EB7,0x1EB7},{0x01001EB8,0x1EB8},{0x01001EB9,0x1EB9},
	{0x01001EBA,0x1EBA},{0x01001EBB,0x1EBB},{0x01001EBC,0x1EBC},{0x01001EBD,0x1EBD},
	{0x01001EBE,0x1EBE},{0x01001EBF,0x1EBF},{0x01001EC0,0x1EC0},{0x01001EC1,0x1EC1},
	{0x01001EC2,0x1EC2},{0x01001EC3,0x1EC3},{0x01001EC4,0x1EC4},{0x01001EC5,0x1EC5},
	{0x01001EC6,0x1EC6},{0x01001EC7,0x1EC7},{0x01001EC8,0x1EC8},{0x01001EC9,0x1EC9},
	{0x01001ECA,0x1ECA},{0x01001ECB,0x1ECB},{0x01001ECC,0x1ECC},{0x01001ECD,0x1ECD},
	{0x01001ECE,0x1ECE},{0x01001ECF,0x1ECF},{0x01001ED0,0x1ED0},{0x01001ED1,0x1ED1},
	{0x01001ED2,0x1ED2},{0x01001ED3,0x1ED3},{0x01001ED4,0x1ED4},{0x01001ED5,0x1ED5},
	{0x01001ED6,0x1ED6},{0x01001ED7,0x1ED7},{0x01001ED8,0x1ED8},{0x01001ED9,0x1ED9},
	{0x01001EDA,0x1EDA},{0x01001EDB,0x1EDB},{0x01001EDC,0x1EDC},{0x01001EDD,0x1EDD},
	{0x01001EDE,0x1EDE},{0x01001EDF,0x1EDF},{0x01001EE0,0x1EE0},{0x01001EE1,0x1EE1},
	{0x01001EE2,0x1EE2},{0x01001EE3,0x1EE3},{0x01001EE4,0x1EE4},{0x01001EE5,0x1EE5},
	{0x01001EE6,0x1EE6},{0x01001EE7,0x1EE7},{0x01001EE8,0x1EE8},{0x01001EE9,0x1EE9},
	{0x01001EEA,0x1EEA},{0x01001EEB,0x1EEB},{0x01001EEC,0x1EEC},{0x01001EED,0x1EED},
	{0x01001EEE,0x1EEE},{0x01001EEF,0x1EEF},{0x01001EF0,0x1EF0},{0x01001EF1,0x1EF1},
	{0x01001EF2,0x1EF2},{0x01001EF3,0x1EF3},{0x01001EF4,0x1EF4},{0x01001EF5,0x1EF5},
	{0x01001EF6,0x1EF6},{0x01001EF7,0x1EF7},{0x01001EF8,0x1EF8},{0x01001EF9,0x1EF9},
	{0x01002070,0x2070},{0x01002074,0x2074},{0x01002075,0x2075},{0x01002076,0x2076},
	{0x01002077,0x2077},{0x01002078,0x2078},{0x01002079,0x2079},{0x01002080,0x2080},
	{0x01002081,0x2081},{0x01002082,0x2082},{0x01002083,0x2083},{0x01002084,0x2084},
	{0x01002085,0x2085},{0x01002086,0x2086},{0x01002087,0x2087},{0x01002088,0x2088},
	{0x01002089,0x2089},{0x010020A0,0x20A0},{0x010020A1,0x20A1},{0x010020A2,0x20A2},
	{0x010020A3,0x20A3},{0x010020A4,0x20A4},{0x010020A5,0x20A5},{0x010020A6,0x20A6},
	{0x010020A7,0x20A7},{0x010020A8,0x20A8},{0x010020A9,0x20A9},{0x010020AA,0x20AA},
	{0x010020AB,0x20AB},{0x01002202,0x2202},{0x01002205,0x2205},{0x01002208,0x2208},
	{0x01002209,0x2209},{0x0100220B,0x220B},{0x0100221A,0x221A},{0x0100221B,0x221B},
	{0x0100221C,0x221C},{0x0100222C,0x222C},{0x0100222D,0x222D},{0x01002235,0x2235},
	{0x01002247,0x2247},{0x01002248,0x2245},{0x01002262,0x2262},{0x01002263,0x2263},
	{0x01002800,0x2800},{0x01002801,0x2801},{0x01002802,0x2802},{0x01002803,0x2803},
	{0x01002804,0x2804},{0x01002805,0x2805},{0x01002806,0x2806},{0x01002807,0x2807},
	{0x01002808,0x2808},{0x01002809,0x2809},{0x0100280A,0x280A},{0x0100280B,0x280B},
	{0x0100280C,0x280C},{0x0100280D,0x280D},{0x0100280E,0x280E},{0x0100280F,0x280F},
	{0x01002810,0x2810},{0x01002811,0x2811},{0x01002812,0x2812},{0x01002813,0x2813},
	{0x01002814,0x2814},{0x01002815,0x2815},{0x01002816,0x2816},{0x01002817,0x2817},
	{0x01002818,0x2818},{0x01002819,0x2819},{0x0100281A,0x281A},{0x0100281B,0x281B},
	{0x0100281C,0x281C},{0x0100281D,0x281D},{0x0100281E,0x281E},{0x0100281F,0x281F},
	{0x01002820,0x2820},{0x01002821,0x2821},{0x01002822,0x2822},{0x01002823,0x2823},
	{0x01002824,0x2824},{0x01002825,0x2825},{0x01002826,0x2826},{0x01002827,0x2827},
	{0x01002828,0x2828},{0x01002829,0x2829},{0x0100282A,0x282A},{0x0100282B,0x282B},
	{0x0100282C,0x282C},{0x0100282D,0x282D},{0x0100282E,0x282E},{0x0100282F,0x282F},
	{0x01002830,0x2830},{0x01002831,0x2831},{0x01002832,0x2832},{0x01002833,0x2833},
	{0x01002834,0x2834},{0x01002835,0x2835},{0x01002836,0x2836},{0x01002837,0x2837},
	{0x01002838,0x2838},{0x01002839,0x2839},{0x0100283A,0x283A},{0x0100283B,0x283B},
	{0x0100283C,0x283C},{0x0100283D,0x283D},{0x0100283E,0x283E},{0x0100283F,0x283F},
	{0x01002840,0x2840},{0x01002841,0x2841},{0x01002842,0x2842},{0x01002843,0x2843},
	{0x01002844,0x2844},{0x01002845,0x2845},{0x01002846,0x2846},{0x01002847,0x2847},
	{0x01002848,0x2848},{0x01002849,0x2849},{0x0100284A,0x284A},{0x0100284B,0x284B},
	{0x0100284C,0x284C},{0x0100284D,0x284D},{0x0100284E,0x284E},{0x0100284F,0x284F},
	{0x01002850,0x2850},{0x01002851,0x2851},{0x01002852,0x2852},{0x01002853,0x2853},
	{0x01002854,0x2854},{0x01002855,0x2855},{0x01002856,0x2856},{0x01002857,0x2857},
	{0x01002858,0x2858},{0x01002859,0x2859},{0x0100285A,0x285A},{0x0100285B,0x285B},
	{0x0100285C,0x285C},{0x0100285D,0x285D},{0x0100285E,0x285E},{0x0100285F,0x285F},
	{0x01002860,0x2860},{0x01002861,0x2861},{0x01002862,0x2862},{0x01002863,0x2863},
	{0x01002864,0x2864},{0x01002865,0x2865},{0x01002866,0x2866},{0x01002867,0x2867},
	{0x01002868,0x2868},{0x01002869,0x2869},{0x0100286A,0x286A},{0x0100286B,0x286B},
	{0x0100286C,0x286C},{0x0100286D,0x286D},{0x0100286E,0x286E},{0x0100286F,0x286F},
	{0x01002870,0x2870},{0x01002871,0x2871},{0x01002872,0x2872},{0x01002873,0x2873},
	{0x01002874,0x2874},{0x01002875,0x2875},{0x01002876,0x2876},{0x01002877,0x2877},
	{0x01002878,0x2878},{0x01002879,0x2879},{0x0100287A,0x287A},{0x0100287B,0x287B},
	{0x0100287C,0x287C},{0x0100287D,0x287D},{0x0100287E,0x287E},{0x0100287F,0x287F},
	{0x01002880,0x2880},{0x01002881,0x2881},{0x01002882,0x2882},{0x01002883,0x2883},
	{0x01002884,0x2884},{0x01002885,0x2885},{0x01002886,0x2886},{0x01002887,0x2887},
	{0x01002888,0x2888},{0x01002889,0x2889},{0x0100288A,0x288A},{0x0100288B,0x288B},
	{0x0100288C,0x288C},{0x0100288D,0x288D},{0x0100288E,0x288E},{0x0100288F,0x288F},
	{0x01002890,0x2890},{0x01002891,0x2891},{0x01002892,0x2892},{0x01002893,0x2893},
	{0x01002894,0x2894},{0x01002895,0x2895},{0x01002896,0x2896},{0x01002897,0x2897},
	{0x01002898,0x2898},{0x01002899,0x2899},{0x0100289A,0x289A},{0x0100289B,0x289B},
	{0x0100289C,0x289C},{0x0100289D,0x289D},{0x0100289E,0x289E},{0x0100289F,0x289F},
	{0x010028A0,0x28A0},{0x010028A1,0x28A1},{0x010028A2,0x28A2},{0x010028A3,0x28A3},
	{0x010028A4,0x28A4},{0x010028A5,0x28A5},{0x010028A6,0x28A6},{0x010028A7,0x28A7},
	{0x010028A8,0x28A8},{0x010028A9,0x28A9},{0x010028AA,0x28AA},{0x010028AB,0x28AB},
	{0x010028AC,0x28AC},{0x010028AD,0x28AD},{0x010028AE,0x28AE},{0x010028AF,0x28AF},
	{0x010028B0,0x28B0},{0x010028B1,0x28B1},{0x010028B2,0x28B2},{0x010028B3,0x28B3},
	{0x010028B4,0x28B4},{0x010028B5,0x28B5},{0x010028B6,0x28B6},{0x010028B7,0x28B7},
	{0x010028B8,0x28B8},{0x010028B9,0x28B9},{0x010028BA,0x28BA},{0x010028BB,0x28BB},
	{0x010028BC,0x28BC},{0x010028BD,0x28BD},{0x010028BE,0x28BE},{0x010028BF,0x28BF},
	{0x010028C0,0x28C0},{0x010028C1,0x28C1},{0x010028C2,0x28C2},{0x010028C3,0x28C3},
	{0x010028C4,0x28C4},{0x010028C5,0x28C5},{0x010028C6,0x28C6},{0x010028C7,0x28C7},
	{0x010028C8,0x28C8},{0x010028C9,0x28C9},{0x010028CA,0x28CA},{0x010028CB,0x28CB},
	{0x010028CC,0x28CC},{0x010028CD,0x28CD},{0x010028CE,0x28CE},{0x010028CF,0x28CF},
	{0x010028D0,0x28D0},{0x010028D1,0x28D1},{0x010028D2,0x28D2},{0x010028D3,0x28D3},
	{0x010028D4,0x28D4},{0x010028D5,0x28D5},{0x010028D6,0x28D6},{0x010028D7,0x28D7},
	{0x010028D8,0x28D8},{0x010028D9,0x28D9},{0x010028DA,0x28DA},{0x010028DB,0x28DB},
	{0x010028DC,0x28DC},{0x010028DD,0x28DD},{0x010028DE,0x28DE},{0x010028DF,0x28DF},
	{0x010028E0,0x28E0},{0x010028E1,0x28E1},{0x010028E2,0x28E2},{0x010028E3,0x28E3},
	{0x010028E4,0x28E4},{0x010028E5,0x28E5},{0x010028E6,0x28E6},{0x010028E7,0x28E7},
	{0x010028E8,0x28E8},{0x010028E9,0x28E9},{0x010028EA,0x28EA},{0x010028EB,0x28EB},
	{0x010028EC,0x28EC},{0x010028ED,0x28ED},{0x010028EE,0x28EE},{0x010028EF,0x28EF},
	{0x010028F0,0x28F0},{0x010028F1,0x28F1},{0x010028F2,0x28F2},{0x010028F3,0x28F3},
	{0x010028F4,0x28F4},{0x010028F5,0x28F5},{0x010028F6,0x28F6},{0x010028F7,0x28F7},
	{0x010028F8,0x28F8},{0x010028F9,0x28F9},{0x010028FA,0x28FA},{0x010028FB,0x28FB},
	{0x010028FC,0x28FC},{0x010028FD,0x28FD},{0x010028FE,0x28FE},{0x010028FF,0x28FF}
};

static int
compar_ks2uni(const void *arg1,const void *arg2)
{
	struct ks2unitab_s *p1 = (struct ks2unitab_s *)arg1;
	struct ks2unitab_s *p2 = (struct ks2unitab_s *)arg2;
	if(p1->ks < p2->ks)
		return -1;
	else if(p1->ks > p2->ks)
		return 1;
	return 0;
}

static unsigned
x11_keysym2uni(unsigned ks)
{
	struct ks2unitab_s ref, *ptr;
	if(ks > 31 && ks < 256)
		return ks;
	ref.ks = ks;
	ptr = (struct ks2unitab_s *)bsearch(&ref,ks2unitab,NUM_XK_SYMS,
		sizeof(struct ks2unitab_s),compar_ks2uni);
	return ptr ? ptr->uni : 0;
}

static void
X11_make_key_codes(arch_window_p aw,XEvent *event,ui_key_event_p dst)
{
	unsigned nb, nub;
	unsigned char buf[68];
	wchar_t wbuf[68];
	KeySym ks = 0;
	GLuint saved_type;
	/*
	 * This is a hack to make XwcLookupString() work for KeyRelease
	 * event as for KeyPress (Xlib doc says this should _not_ be done).
	 */
	saved_type = event->type;
	event->type = KeyPress;
	if(aw->ic)
		nub = XwcLookupString(aw->ic, &event->xkey, wbuf, 64, &ks, NULL);
	else
		nub = 0;
	dst->utf8_len = Xutf8LookupString(aw->ic, &event->xkey,
		(char*)dst->utf8, AG_UTF8_MAXLEN, &ks, NULL);
	dst->utf8[(dst->utf8_len < AG_UTF8_MAXLEN) ?
		dst->utf8_len : AG_UTF8_MAXLEN-1] = 0;
	nb = XLookupString(&event->xkey, (char*)buf, 64, &ks, NULL);
	event->type = saved_type;
	dst->scan = aw->keycode_to_scan[event->xkey.keycode];
	if(nb > 0)
		dst->acode = buf[0];
	else
	{
		if(ks > 31 && ks < 256)
			dst->acode = ks;
		else
			dst->acode = 0;
	}
	if(nub > 0)
		dst->ucode = wbuf[0];
	else /* As a last resort, try to guess unicode from KeySym */
		dst->ucode = x11_keysym2uni(ks);
}

static void
X11_make_key_states(arch_window_p aw,GLuint x_state, ui_event_p ev)
{
	ev->state = ev->button = 0;

	if(x_state & Mod2Mask)
		ev->state |= UI_NUM_LOCK_MASK;
	if(x_state & LockMask)
		ev->state |= UI_CAPS_LOCK_MASK;
	if(x_state & Mod5Mask)
		ev->state |= UI_SCROLL_LOCK_MASK;
	if(x_state & ShiftMask)
		ev->state |= UI_SHIFT_MASK;
	if(x_state & ControlMask)
		ev->state |= UI_CONTROL_MASK;
	if(x_state & Mod1Mask)
		ev->state |= UI_ALT_MASK;
	if(x_state & Mod4Mask)
		ev->state |= UI_SUPER_MASK;

	if(x_state & Button1Mask)
		ev->button |= UI_LBUTTON_MASK;
	if(x_state & Button2Mask)
		ev->button |= UI_MBUTTON_MASK;
	if(x_state & Button3Mask)
		ev->button |= UI_RBUTTON_MASK;

	aw->last_key_state = ev->state;
	aw->last_button_state = ev->button;
}

static ui_event_type_t
X11_process_event(arch_window_p aw,XEvent *event,ui_event_p dst)
{
	x11_display_connection_p cn = aw->cn;
	ui_make_default_key_event(&dst->key);
	switch(event->type)
	{
		case KeyPress:
		case KeyRelease:

			dst->key.press = (event->type == KeyPress) ? GL_TRUE : GL_FALSE;
			if(!dst->key.press && aw->auto_repeat_disabled)
			{
				/*
				 * The Id's idea of XAutoRepeatOn/Off() is a
				 * bad one, since we have to change global X server state.
				 *
				 * There is a better way to ignore repeated keys;
				 * it was demonstrated in MJK's glut_event.c.
				 */
				XEvent head;
				if(XEventsQueued(cn->dpy, QueuedAfterReading))
				{
					XPeekEvent(cn->dpy, &head);
					if(head.type == KeyPress && 
						head.xkey.window == event->xkey.window &&
						head.xkey.keycode == event->xkey.keycode &&
						head.xkey.time == event->xkey.time)
					{
						XNextEvent(cn->dpy, &head);
						return ui_no_event;
					}
				}
			}
			dst->type = ui_key_event;
			dst->time =
				ui_make_message_time(&aw->start_time, event->xkey.time);
			dst->client_x = event->xkey.x;
			dst->client_y = event->xkey.y;
			dst->screen_x = event->xkey.x_root;
			dst->screen_y = event->xkey.y_root;
			X11_make_key_states(aw, event->xkey.state, dst);
			X11_make_key_codes(aw, event, &dst->key);
			if(dst->key.scan == sc_unknown)
				return ui_no_event;
			ui_set_configure_from_aw(aw, &dst->configure);
			aw->keyboard_state[dst->key.scan] = dst->key.press ? 1 : 0;
			break;

		case ButtonPress:
		case ButtonRelease:
	
			dst->key.press = (event->type == ButtonPress) ? GL_TRUE : GL_FALSE;
			switch(event->xbutton.button)
			{
				case Button1: dst->key.scan = sc_lbutton; break;
				case Button2: dst->key.scan = sc_mbutton; break;
				case Button3: dst->key.scan = sc_rbutton; break;
				case Button4:
					if(!dst->key.press) /* fake X-generated message */
						return ui_no_event;
					dst->key.scan = sc_wheelup;
					aw->keyboard_state[sc_wheeldown] = 0;
					break;
				case Button5:
					if(!dst->key.press) /* fake X-generated message */
						return ui_no_event;
					dst->key.scan = sc_wheeldown;
					aw->keyboard_state[sc_wheelup] = 0;
					break;
				default:
					return ui_no_event;
			}
			dst->type = ui_key_event;
			dst->time =
				ui_make_message_time(&aw->start_time, event->xkey.time);
			dst->client_x = event->xbutton.x;
			dst->client_y = event->xbutton.y;
			dst->screen_x = event->xbutton.x_root;
			dst->screen_y = event->xbutton.y_root;
			X11_make_key_states(aw, event->xbutton.state, dst);
			dst->key.acode = dst->key.ucode = 0;
			ui_set_configure_from_aw(aw, &dst->configure);
			aw->keyboard_state[dst->key.scan] = dst->key.press ? 1 : 0;
			break;

		case MotionNotify:

			dst->type = ui_pointer_motion_event;
			dst->time =
				ui_make_message_time(&aw->start_time, event->xkey.time);
			dst->client_x = event->xmotion.x;
			dst->client_y = event->xmotion.y;
			dst->screen_x = event->xmotion.x_root;
			dst->screen_y = event->xmotion.y_root;
			X11_make_key_states(aw, event->xkey.state, dst);
			ui_set_configure_from_aw(aw, &dst->configure);

			break;

		case FocusIn:
		case FocusOut:
		
			if(event->type == FocusIn)
			{
				aw->focused = GL_TRUE;
				dst->type = ui_window_focus_in_event;
				if(aw->saved_vidmode != -1)
					archWinSetVidMode(aw, aw->saved_vidmode);
			}
			else
			{
				aw->focused = GL_FALSE;
				dst->type = ui_window_focus_out_event;
				if(aw->current_vidmode != -1)
				{
					aw->saved_vidmode = aw->current_vidmode;
					archWinSetVidMode(aw, -1);
				}
			}
			dst->time = 0;
			dst->client_x = dst->client_y = dst->screen_x = dst->screen_y = 
				dst->client_dx = dst->client_dy = 0;
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			break;

		case ConfigureNotify:
		
			if(event->xconfigure.width != aw->width ||
				event->xconfigure.height != aw->height)
				dst->type = ui_window_size_event;
			else
				dst->type = ui_window_move_event;
			dst->time = 0;
			dst->client_x = dst->client_y = dst->screen_x = dst->screen_y = 
				dst->client_dx = dst->client_dy = 0;
			aw->x = event->xconfigure.x;
			aw->y = event->xconfigure.y;
			aw->width = event->xconfigure.width;
			aw->height = event->xconfigure.height;
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			break;

		case MapNotify:
		case UnmapNotify:
		
			if(event->type == MapNotify)
			{
				aw->visible = GL_TRUE;
				dst->type = ui_window_map_event;
			}
			else
			{
				aw->visible = GL_FALSE;
				dst->type = ui_window_unmap_event;
			}
			dst->time = 0;
			dst->client_x = dst->client_y = dst->screen_x = dst->screen_y = 
				dst->client_dx = dst->client_dy = 0;
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			break;

		case MappingNotify:

			XRefreshKeyboardMapping((XMappingEvent *)event);
			break;

		case ClientMessage:

			if(event->xclient.message_type != cn->wm_protocols_atom ||
					event->xclient.format != 32 ||
					(Atom)event->xclient.data.l[0] != cn->wm_delete_window_atom)
				return ui_no_event;
			dst->type = ui_wm_event;
			dst->time = 0;
			dst->client_x = dst->client_y = dst->screen_x = dst->screen_y = 
				dst->client_dx = dst->client_dy = 0;
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			dst->configure.wm_event = ui_wm_close;
			break;

	default:

		return ui_no_event;
	}

	return dst->type;
}

static ui_event_type_t
X11_post_process_event(arch_window_p aw,XEvent *event,ui_event_p dst)
{
	ui_event_type_t retval = X11_process_event(aw, event, dst);
	if(retval != ui_no_event && ui_default_wnd_proc(aw, dst))
		return retval;
	return ui_no_event;
}

static void
x11_poll_events(arch_window_p aw)
{
	ui_event_t ev;
	XEvent xev;
	long l, t, r, b, left, top, right, bottom;
	x11_display_connection_p cn = aw->cn;
	/*
	 * There seems to be no good method to lock viewport after
	 * mode switch. This is last resort.
	 */
#ifdef	ARCH_HAVE_XF86_VIDMODE
	if(cn->vidmode_active)
	{
		int x = 0, y = 0;
		XF86VidModeGetViewPort(cn->dpy, cn->scrnum, &x, &y);
		if(x || y)
		{
			XF86VidModeSetViewPort(cn->dpy, cn->scrnum, 0, 0);
			/* XWarpPointer(cn->dpy,None,None,0,0,0,0,-x,-y); */
			archWinWarpPointer(aw, aw->last_mx, aw->last_my);
		}
	}
#endif	/* ARCH_HAVE_XF86_VIDMODE */
	while(XCheckTypedWindowEvent(cn->dpy, aw->wnd, ClientMessage ,&xev))
	{
		if(X11_post_process_event(aw, &xev, &ev) != ui_no_event)
			DEQ_PUSH(event, aw->msg_que, &ev);
	}
	/*
	 * Remove all Expose events from queue and collapse them.
	 */
	left = top = 0x7FFFFFFF;
	right = bottom = -1;
	while(XCheckTypedWindowEvent(cn->dpy, aw->wnd, Expose, &xev))
	{
		l = xev.xexpose.x;
		t = xev.xexpose.y;
		r = l + xev.xexpose.width;
		b = t + xev.xexpose.height;
		if(left > l)
			left = l;
		if(top > t)
			top = t;
		if(right < r)
			right = r;
		if(bottom < b)
			bottom = b;
	}
	if(right != -1)
	{
		ev.type = ui_window_expose_event;
		ev.time = 0;
		ev.client_x = ev.client_y = ev.screen_x = ev.screen_y =
			ev.client_dx = ev.client_dy = 0;
		ui_make_default_key_event(&ev.key);
		ui_set_configure_from_aw(aw, &ev.configure);
		ev.configure.exposed_left = left;
		ev.configure.exposed_top = top;
		ev.configure.exposed_right = right;
		ev.configure.exposed_bottom = bottom;
		ev.state = aw->last_key_state;
		ev.button = aw->last_button_state;
		DEQ_PUSH_BACK(event, aw->msg_que, &ev);
	}
	while(XCheckWindowEvent(cn->dpy, aw->wnd, aw->event_mask, &xev))
	{
		if(X11_post_process_event(aw, &xev, &ev) != ui_no_event)
			DEQ_PUSH(event, aw->msg_que, &ev);
	}
}

/*
 * archWinWaitMessages()
 *
 * Based on interruptibleXNextEvent() from glut.
 */
static GLboolean
archWinWaitMessages(arch_window_p aw, GLuint timeout)
{
	fd_set fds;
	struct timeval	delay;
	GLuint rc;
	XFlush(aw->cn->dpy);
	if(XPending(aw->cn->dpy))
		return GL_TRUE;
	FD_ZERO(&fds);
	FD_SET(aw->cn->connection_fd, &fds);
	if(timeout != -1)
	{
		delay.tv_sec = timeout / 1000;
		delay.tv_usec = 1000 * (timeout % 1000);
		rc = select(aw->cn->connection_fd + 1, &fds, NULL, NULL, &delay);
	}
	else
		rc = select(aw->cn->connection_fd + 1, &fds, NULL, NULL, NULL);
	if(XPending(aw->cn->dpy))
		return GL_TRUE;
	return GL_FALSE;
}

static void
archWinPoll(arch_window_p aw, GLboolean block)
{
	x11_poll_events(aw);
	if(!block || aw->msg_que)
		return;
	archWinWaitMessages(aw,-1);
	x11_poll_events(aw);
}

static void
archWinGrabKeyboard(arch_window_p aw,GLboolean grab)
{
	if((grab && aw->cn->grab_keyboard) || (!grab && !aw->cn->grab_keyboard))
		return;
	if(grab)
	{
		XGrabKeyboard(aw->cn->dpy, aw->wnd, False,
			GrabModeAsync, GrabModeAsync, CurrentTime);
			aw->cn->grab_keyboard = GL_TRUE;
	}
	else
	{
		XUngrabKeyboard(aw->cn->dpy, CurrentTime);
		aw->cn->grab_keyboard = GL_FALSE;
	}
}

static void
archWinGrabMouse(arch_window_p aw,GLboolean grab)
{
	if(!aw->visible)
	{
		TRACE("Window invisible.");
		return;
	}
	if(grab)
	{
		if(!aw->ms_grab_count++)
		{
			XGrabPointer(aw->cn->dpy, aw->wnd, True, 0,
				GrabModeAsync, GrabModeAsync, aw->wnd, None, CurrentTime);
		}
	}
	else
	{
		if(!--aw->ms_grab_count)
		{
			XUngrabPointer(aw->cn->dpy, CurrentTime);
		}
	}
}

static GLboolean
archWinShow(arch_window_p aw,GLboolean show)
{
	if(show)
	{
		XMapRaised(aw->cn->dpy, aw->wnd);
		aw->visible = GL_TRUE;
		ui_win_loose_control(aw, GL_FALSE);
	}
	else
	{
		aw->visible = GL_FALSE;
		ui_win_loose_control(aw, GL_TRUE);
		XUnmapWindow(aw->cn->dpy, aw->wnd);
	}
	return GL_TRUE;
}

/*
 * Returns 'GL_TRUE' if cursor is visible at exit.
 */
static GLboolean
archWinShowPointer(arch_window_p aw,GLboolean show)
{
	GLboolean result;
	if(show)
	{
		aw->pointer_hide_count--;
		if(!aw->pointer_hide_count)
			XUndefineCursor(aw->cn->dpy,aw->wnd);
	}
	else
	{
		if(!aw->pointer_hide_count)
			XDefineCursor(aw->cn->dpy,aw->wnd,
				aw->cn->null_cursor);
		aw->pointer_hide_count++;
	}
	return result = (aw->pointer_hide_count > 0) ? GL_FALSE : GL_TRUE;
}

static GLboolean
archWinWarpPointer(arch_window_p aw,GLint x,GLint y)
{
	XWarpPointer(aw->cn->dpy, None, aw->wnd, 0, 0, 0, 0, x, y);
	return GL_TRUE;
}

static void
archWinClose(arch_window_p aw)
{
	while(aw->pointer_hide_count > 0)
		archWinShowPointer(aw, GL_TRUE);
	if(aw->wnd)
	{
		XEvent ev;
		ui_win_loose_control(aw, GL_TRUE);
		XUnmapWindow(aw->cn->dpy, aw->wnd);
		XDestroyWindow(aw->cn->dpy, aw->wnd);
		if(aw->ic)
			XDestroyIC(aw->ic);
		archWinWaitMessages(aw,1);
		while(XCheckWindowEvent(aw->cn->dpy, aw->wnd, aw->event_mask, &ev));
		
	}
	if(aw->cmap)
		XFreeColormap(aw->cn->dpy, aw->cmap);
	CLEAR_COMMON_WINDOW_FIELDS(aw);
	memFree(aw);
}

/*
 * "Smart" version: not working in rare cases.
 */

#ifdef	USE_GLX_GET_PROC_ADDRESS

void (*(*glx_get_proc_address)(const GLubyte *procname))(void) = NULL;

static void *libgl_handle = NULL;

static GLboolean
vgl_load_proc_init(char *proc_name,void **ptr_ret,GLboolean *terminate);

GLboolean (*vgl_load_proc)(char *proc_name,void **ptr_ret,GLboolean *terminate) = vgl_load_proc_init;

static GLboolean
vgl_load_proc_real(char *proc_name,void **ptr_ret,GLboolean *terminate)
{
	void *proc = (void*)glx_get_proc_address((const GLubyte *)proc_name);
	if(*ptr_ret && (*ptr_ret != proc))
	{
		/*
		 * This should never happen under X11 + GLX.
		 */
		TRACE("Different retval of glXGetProcAddress(): GLX broken ?");
		*terminate = GL_TRUE;
		return GL_FALSE;
	}
	if(proc)
	{
		*ptr_ret = (void*)proc;
		return GL_TRUE;
	}
	return GL_FALSE;
}

static GLboolean
vgl_load_proc_init(char *proc_name,void **ptr_ret,GLboolean *terminate)
{
	GLboolean result = GL_FALSE;
	if(!glx_get_proc_address)
	{
		if(!libgl_handle)
		{
			libgl_handle = dlopen("libGL.so",RTLD_LAZY);
			if(!libgl_handle)
			{
				TRACE_1ARG("dlopen(\"libGL.so\"): failed: \"%s\".", dlerror());
				return GL_FALSE;
			}
			*((void**)&glx_get_proc_address) = dlsym(libgl_handle, "glXGetProcAddress");
			if(!glx_get_proc_address)
			{
				/* Do NOT close libGL */
				TRACE("no glXGetProcAddress[ARB]() in libGL.so.");
				return GL_FALSE;
			}
		}
	
	}
	vgl_load_proc = vgl_load_proc_real;
	return vgl_load_proc(proc_name, ptr_ret, terminate);
}

#else	/* !USE_GLX_GET_PROC_ADDRESS */

/*
 * "Brute" version: no usage of glXGetProcAdress() at all.
 */

static void *libgl_handle = NULL; /* safely to be kept unprotected */

static GLboolean
vgl_load_proc_init(char *proc_name,void **ptr_ret,GLboolean *terminate);

GLboolean (*vgl_load_proc)(char *proc_name,void **ptr_ret,GLboolean *terminate) = vgl_load_proc_init;

static GLboolean
vgl_load_proc_real(char *proc_name,void **ptr_ret,GLboolean *terminate)
{
	void *proc = (void*)dlsym(libgl_handle, proc_name);
	if(proc)
	{
		*ptr_ret = (void*)proc;
		return GL_TRUE;
	}
	return GL_FALSE;
}

static GLboolean
vgl_load_proc_init(char *proc_name,void **ptr_ret,GLboolean *terminate)
{
	if(!libgl_handle) /* GL library not loaded yet */
	{
		/* In the worst case we load GL twice; acceptable. */
		libgl_handle = dlopen("libGL.so",RTLD_LAZY);
		if(!libgl_handle)
		{
			TRACE_1ARG("dlopen(\"libGL.so\"): failed: \"%s\".", dlerror());
			/* force again vgl_load_proc_init() call instead of vgl_load_proc() */
			return GL_FALSE;
		}
	}
	/* GL library loaded */
	vgl_load_proc = vgl_load_proc_real;
	return vgl_load_proc(proc_name, ptr_ret, terminate);
}

#endif	/* !USE_GLX_GET_PROC_ADDRESS */

void
agSwap()
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	XSync(ctx->wnd->cn->dpy,False);
	glXSwapBuffers(ctx->wnd->cn->dpy, ctx->wnd->wnd);
}

static void
x11_destroy_context(gl_context_p ctx)
{
	/* NOTE: we don't close libGL.so */
	x11_display_connection_p cn = ctx->wnd->cn;
	if(ctx->glx_context)
	{
		glXMakeCurrent(ctx->wnd->cn->dpy, None, NULL);
		glXDestroyContext(ctx->wnd->cn->dpy, ctx->glx_context);
	}
	if(ctx->visual_info)
		XFree(ctx->visual_info);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx);
	x11_close_connection(cn);
	memFree(ctx);
}

void
agShutdown(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	*get_thread_context_addr() = NULL;
	x11_destroy_context(ctx);
}

static gl_context_p
vglInit(
	char *caption,
	GLuint x, GLuint y, GLuint width, GLuint height,
	GLboolean fullscreen,
	GLboolean can_resize,
	GLboolean need_depth,
	GLboolean need_stencil,
	GLboolean need_dst_alpha)
{
	GLfloat aspect;
	gl_context_p ctx = NULL;
	GLuint bestmode;
	GLuint visinfo_class;
	GLint attrib[64], *aptr = attrib;
	x11_display_connection_p cn;
	cn = x11_open_connection();
	if(!cn)
		return NULL;
	ctx = alloc_gl_context();
	*aptr++ = GLX_RGBA;
	*aptr++ = GLX_DOUBLEBUFFER;
	*aptr++ = GLX_RED_SIZE;
	*aptr++ = 1;
	*aptr++ = GLX_GREEN_SIZE;
	*aptr++ = 1;
	*aptr++ = GLX_BLUE_SIZE;
	*aptr++ = 1;
	*aptr = None;
	if(need_depth)
	{
		*aptr++ = GLX_DEPTH_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	if(need_stencil)
	{
		*aptr++ = GLX_STENCIL_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	if(need_dst_alpha)
	{
		*aptr++ = GLX_ALPHA_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	ctx->visual_info = glXChooseVisual(cn->dpy, cn->scrnum, attrib);
	if(!ctx->visual_info)
	{
		TRACE("glXChooseVisual() failed.");
		goto err_exit;
	}
#if defined(__cplusplus) || defined(c_plusplus)
	visinfo_class = ctx->visual_info->c_class;
#else
	visinfo_class = ctx->visual_info->class;
#endif
	if(visinfo_class != DirectColor && visinfo_class != TrueColor)
	{
		TRACE("DirectColor or TrueColor visual required.");
		goto err_exit;
	}
	ctx->glx_context =
		glXCreateContext(cn->dpy, ctx->visual_info, NULL, True);
	if(!ctx->glx_context)
	{
		TRACE("glXCreateContext() failed.");
		goto err_exit;
	}
	ctx->wnd = archWinOpenNP(cn,ctx->visual_info,
		caption, x, y, width, height, !fullscreen, can_resize);
	if(!ctx->wnd)
		goto err_exit;
	width = ctx->wnd->width;
	height = ctx->wnd->height;
	archWinShow(ctx->wnd, GL_TRUE);
	XFlush(cn->dpy);
	if(fullscreen)
	{
		for(bestmode = 0; bestmode < ctx->wnd->cn->num_vidmodes; bestmode++)
		{
			vidmode_p mode = ctx->wnd->cn->vidmodes + bestmode;
			if((GLuint)mode->width == width && (GLuint)mode->height == height)
				break;
		}
		if(bestmode < ctx->wnd->cn->num_vidmodes)
			archWinSetVidMode(ctx->wnd, bestmode);
	}
	if(!glXMakeCurrent(cn->dpy, ctx->wnd->wnd, ctx->glx_context))
	{
		TRACE("glXMakeCurrent() failed.");
		goto err_exit;
	}
	ctx->vendor = strDup((char*)glGetString(GL_VENDOR));
	ctx->renderer = strDup((char*)glGetString(GL_RENDERER));
	ctx->version = strDup((char*)glGetString(GL_VERSION));
	ctx->extensions = strDup((char*)glGetString(GL_EXTENSIONS));
#ifdef	AG_VERBOSE
	printf("%s OpenGL on %s ver. %s.\n",
		ctx->vendor,ctx->renderer,ctx->version);
#endif	/* AG_VERBOSE */
	archWinPoll(ctx->wnd, GL_FALSE);
	aspect = (GLfloat)ctx->wnd->width / (GLfloat)ctx->wnd->height;
	glViewport(0, 0, ctx->wnd->width, ctx->wnd->height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-aspect, aspect, -1, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	return ctx;

err_exit:;
	x11_destroy_context(ctx);
	return NULL;
}

void
agWindowTitle(char *title)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	if(!title)
		return;
	strncpy(ctx->wnd->title, title, MAX_WINDOW_TITLE);
	XStoreName(ctx->wnd->cn->dpy, ctx->wnd->wnd, title);
}

GLuint
agTickCount(void)
{
	gl_context_p ctx;
	struct timezone tz = {0,0};
	struct timeval tv;
	unsigned tick;
	gettimeofday((struct timeval*)&tv,&tz);
	tick = tv.tv_sec * 1000 + tv.tv_usec / 1000;
	IF_NO_GLCONTEXT(ctx)
		return tick;
	return tick - ctx->wnd->start_time;
}

GLvoid
agSleep(GLuint milliseconds)
{
	struct timeval delay;
	delay.tv_sec = milliseconds / 1000;
	delay.tv_usec = 1000 * (milliseconds % 1000);
	select(0,NULL,NULL,NULL,&delay);
}

#endif	/* __UNIX__ */

/*
 * OGL loading stuff (portable).
 */

#define	LGLP(name) \
	(flag = (GLboolean)(flag && vgl_load_proc(#name,(void**)(&name),terminate)))

#define	RGLP(name) \
	if(name != NULL) \
		*terminate = GL_TRUE; \
	name = NULL

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.2.
 */

void (VGLEXTAPIENTRY *glBlendColor) (GLclampf, GLclampf, GLclampf, GLclampf);
void (VGLEXTAPIENTRY *glBlendEquation) (GLenum);
void (VGLEXTAPIENTRY *glDrawRangeElements) (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glTexImage3D) (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glCopyTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei);
void (VGLEXTAPIENTRY *glColorTable) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glColorTableParameterfv) (GLenum, GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glColorTableParameteriv) (GLenum, GLenum, const GLint *);
void (VGLEXTAPIENTRY *glCopyColorTable) (GLenum, GLenum, GLint, GLint, GLsizei);
void (VGLEXTAPIENTRY *glGetColorTable) (GLenum, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetColorTableParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetColorTableParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glColorSubTable) (GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glCopyColorSubTable) (GLenum, GLsizei, GLint, GLint, GLsizei);
void (VGLEXTAPIENTRY *glConvolutionFilter1D) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glConvolutionFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glConvolutionParameterf) (GLenum, GLenum, GLfloat);
void (VGLEXTAPIENTRY *glConvolutionParameterfv) (GLenum, GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glConvolutionParameteri) (GLenum, GLenum, GLint);
void (VGLEXTAPIENTRY *glConvolutionParameteriv) (GLenum, GLenum, const GLint *);
void (VGLEXTAPIENTRY *glCopyConvolutionFilter1D) (GLenum, GLenum, GLint, GLint, GLsizei);
void (VGLEXTAPIENTRY *glCopyConvolutionFilter2D) (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei);
void (VGLEXTAPIENTRY *glGetConvolutionFilter) (GLenum, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetConvolutionParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetConvolutionParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetSeparableFilter) (GLenum, GLenum, GLenum, GLvoid *, GLvoid *, GLvoid *);
void (VGLEXTAPIENTRY *glSeparableFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *, const GLvoid *);
void (VGLEXTAPIENTRY *glGetHistogram) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetHistogramParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetHistogramParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetMinmax) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetMinmaxParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetMinmaxParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glHistogram) (GLenum, GLsizei, GLenum, GLboolean);
void (VGLEXTAPIENTRY *glMinmax) (GLenum, GLenum, GLboolean);
void (VGLEXTAPIENTRY *glResetHistogram) (GLenum);
void (VGLEXTAPIENTRY *glResetMinmax) (GLenum);

static GLboolean
vgl_load_gl_1_2(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
#ifndef	AG_NO_GL_12_IMAGING
	LGLP(glColorTable);
	LGLP(glColorTableParameterfv);
	LGLP(glColorTableParameteriv);
	LGLP(glCopyColorTable);
	LGLP(glGetColorTable);
	LGLP(glGetColorTableParameterfv);
	LGLP(glGetColorTableParameteriv);
	LGLP(glColorSubTable);
	LGLP(glCopyColorSubTable);
	LGLP(glConvolutionFilter1D);
	LGLP(glConvolutionFilter2D);
	LGLP(glConvolutionParameterf);
	LGLP(glConvolutionParameterfv);
	LGLP(glConvolutionParameteri);
	LGLP(glConvolutionParameteriv);
	LGLP(glCopyConvolutionFilter1D);
	LGLP(glCopyConvolutionFilter2D);
	LGLP(glGetConvolutionFilter);
	LGLP(glGetConvolutionParameterfv);
	LGLP(glGetConvolutionParameteriv);
	LGLP(glGetSeparableFilter);
	LGLP(glSeparableFilter2D);
	LGLP(glGetHistogram);
	LGLP(glGetHistogramParameterfv);
	LGLP(glGetHistogramParameteriv);
	LGLP(glGetMinmax);
	LGLP(glGetMinmaxParameterfv);
	LGLP(glGetMinmaxParameteriv);
	LGLP(glHistogram);
	LGLP(glMinmax);
	LGLP(glResetHistogram);
	LGLP(glResetMinmax);
	LGLP(glBlendColor);
	LGLP(glBlendEquation);
	if(!flag)
	{
		TRACE("vgl_load_gl_1_2(): Warning: unable to load GL_ARB_imaging.");
		flag = GL_TRUE;
	}
#endif	/* !AG_NO_GL_12_IMAGING */
	LGLP(glDrawRangeElements);
	LGLP(glTexImage3D);
	LGLP(glTexSubImage3D);
	LGLP(glCopyTexSubImage3D);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.2 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.3.
 */

void (VGLEXTAPIENTRY *glActiveTexture) (GLenum);
void (VGLEXTAPIENTRY *glClientActiveTexture) (GLenum);
void (VGLEXTAPIENTRY *glMultiTexCoord1d) (GLenum, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord1dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord1f) (GLenum, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord1fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord1i) (GLenum, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord1iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord1s) (GLenum, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord1sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glMultiTexCoord2d) (GLenum, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord2dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord2f) (GLenum, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord2fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord2i) (GLenum, GLint, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord2iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord2s) (GLenum, GLshort, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord2sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glMultiTexCoord3d) (GLenum, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord3dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord3f) (GLenum, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord3fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord3i) (GLenum, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord3iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord3s) (GLenum, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord3sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glMultiTexCoord4d) (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord4dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord4f) (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord4fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord4i) (GLenum, GLint, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord4iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord4s) (GLenum, GLshort, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord4sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glLoadTransposeMatrixf) (const GLfloat *);
void (VGLEXTAPIENTRY *glLoadTransposeMatrixd) (const GLdouble *);
void (VGLEXTAPIENTRY *glMultTransposeMatrixf) (const GLfloat *);
void (VGLEXTAPIENTRY *glMultTransposeMatrixd) (const GLdouble *);
void (VGLEXTAPIENTRY *glSampleCoverage) (GLclampf, GLboolean);
void (VGLEXTAPIENTRY *glCompressedTexImage3D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexImage2D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexImage1D) (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexSubImage2D) (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexSubImage1D) (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glGetCompressedTexImage) (GLenum, GLint, GLvoid *);

static GLboolean
vgl_load_gl_1_3(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
	LGLP(glActiveTexture);
	LGLP(glClientActiveTexture);
	LGLP(glMultiTexCoord1d);
	LGLP(glMultiTexCoord1dv);
	LGLP(glMultiTexCoord1f);
	LGLP(glMultiTexCoord1fv);
	LGLP(glMultiTexCoord1i);
	LGLP(glMultiTexCoord1iv);
	LGLP(glMultiTexCoord1s);
	LGLP(glMultiTexCoord1sv);
	LGLP(glMultiTexCoord2d);
	LGLP(glMultiTexCoord2dv);
	LGLP(glMultiTexCoord2f);
	LGLP(glMultiTexCoord2fv);
	LGLP(glMultiTexCoord2i);
	LGLP(glMultiTexCoord2iv);
	LGLP(glMultiTexCoord2s);
	LGLP(glMultiTexCoord2sv);
	LGLP(glMultiTexCoord3d);
	LGLP(glMultiTexCoord3dv);
	LGLP(glMultiTexCoord3f);
	LGLP(glMultiTexCoord3fv);
	LGLP(glMultiTexCoord3i);
	LGLP(glMultiTexCoord3iv);
	LGLP(glMultiTexCoord3s);
	LGLP(glMultiTexCoord3sv);
	LGLP(glMultiTexCoord4d);
	LGLP(glMultiTexCoord4dv);
	LGLP(glMultiTexCoord4f);
	LGLP(glMultiTexCoord4fv);
	LGLP(glMultiTexCoord4i);
	LGLP(glMultiTexCoord4iv);
	LGLP(glMultiTexCoord4s);
	LGLP(glMultiTexCoord4sv);
	LGLP(glLoadTransposeMatrixf);
	LGLP(glLoadTransposeMatrixd);
	LGLP(glMultTransposeMatrixf);
	LGLP(glMultTransposeMatrixd);
	LGLP(glSampleCoverage);
	LGLP(glCompressedTexImage3D);
	LGLP(glCompressedTexImage2D);
	LGLP(glCompressedTexImage1D);
	LGLP(glCompressedTexSubImage3D);
	LGLP(glCompressedTexSubImage2D);
	LGLP(glCompressedTexSubImage1D);
	LGLP(glGetCompressedTexImage);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.3 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.4.
 */

void (VGLEXTAPIENTRY *glBlendFuncSeparate) (GLenum, GLenum, GLenum, GLenum);
void (VGLEXTAPIENTRY *glFogCoordf) (GLfloat);
void (VGLEXTAPIENTRY *glFogCoordfv) (const GLfloat *);
void (VGLEXTAPIENTRY *glFogCoordd) (GLdouble);
void (VGLEXTAPIENTRY *glFogCoorddv) (const GLdouble *);
void (VGLEXTAPIENTRY *glFogCoordPointer) (GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glMultiDrawArrays) (GLenum, GLint *, GLsizei *, GLsizei);
void (VGLEXTAPIENTRY *glMultiDrawElements) (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
void (VGLEXTAPIENTRY *glPointParameterf) (GLenum, GLfloat);
void (VGLEXTAPIENTRY *glPointParameterfv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glPointParameteri) (GLenum, GLint);
void (VGLEXTAPIENTRY *glPointParameteriv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glSecondaryColor3b) (GLbyte, GLbyte, GLbyte);
void (VGLEXTAPIENTRY *glSecondaryColor3bv) (const GLbyte *);
void (VGLEXTAPIENTRY *glSecondaryColor3d) (GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glSecondaryColor3dv) (const GLdouble *);
void (VGLEXTAPIENTRY *glSecondaryColor3f) (GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glSecondaryColor3fv) (const GLfloat *);
void (VGLEXTAPIENTRY *glSecondaryColor3i) (GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glSecondaryColor3iv) (const GLint *);
void (VGLEXTAPIENTRY *glSecondaryColor3s) (GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glSecondaryColor3sv) (const GLshort *);
void (VGLEXTAPIENTRY *glSecondaryColor3ub) (GLubyte, GLubyte, GLubyte);
void (VGLEXTAPIENTRY *glSecondaryColor3ubv) (const GLubyte *);
void (VGLEXTAPIENTRY *glSecondaryColor3ui) (GLuint, GLuint, GLuint);
void (VGLEXTAPIENTRY *glSecondaryColor3uiv) (const GLuint *);
void (VGLEXTAPIENTRY *glSecondaryColor3us) (GLushort, GLushort, GLushort);
void (VGLEXTAPIENTRY *glSecondaryColor3usv) (const GLushort *);
void (VGLEXTAPIENTRY *glSecondaryColorPointer) (GLint, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glWindowPos2d) (GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glWindowPos2dv) (const GLdouble *);
void (VGLEXTAPIENTRY *glWindowPos2f) (GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glWindowPos2fv) (const GLfloat *);
void (VGLEXTAPIENTRY *glWindowPos2i) (GLint, GLint);
void (VGLEXTAPIENTRY *glWindowPos2iv) (const GLint *);
void (VGLEXTAPIENTRY *glWindowPos2s) (GLshort, GLshort);
void (VGLEXTAPIENTRY *glWindowPos2sv) (const GLshort *);
void (VGLEXTAPIENTRY *glWindowPos3d) (GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glWindowPos3dv) (const GLdouble *);
void (VGLEXTAPIENTRY *glWindowPos3f) (GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glWindowPos3fv) (const GLfloat *);
void (VGLEXTAPIENTRY *glWindowPos3i) (GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glWindowPos3iv) (const GLint *);
void (VGLEXTAPIENTRY *glWindowPos3s) (GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glWindowPos3sv) (const GLshort *);

static GLboolean
vgl_load_gl_1_4(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
	LGLP(glBlendFuncSeparate);
	LGLP(glFogCoordf);
	LGLP(glFogCoordfv);
	LGLP(glFogCoordd);
	LGLP(glFogCoorddv);
	LGLP(glFogCoordPointer);
	LGLP(glMultiDrawArrays);
	LGLP(glMultiDrawElements);
	LGLP(glPointParameterf);
	LGLP(glPointParameterfv);
	LGLP(glPointParameteri);
	LGLP(glPointParameteriv);
	LGLP(glSecondaryColor3b);
	LGLP(glSecondaryColor3bv);
	LGLP(glSecondaryColor3d);
	LGLP(glSecondaryColor3dv);
	LGLP(glSecondaryColor3f);
	LGLP(glSecondaryColor3fv);
	LGLP(glSecondaryColor3i);
	LGLP(glSecondaryColor3iv);
	LGLP(glSecondaryColor3s);
	LGLP(glSecondaryColor3sv);
	LGLP(glSecondaryColor3ub);
	LGLP(glSecondaryColor3ubv);
	LGLP(glSecondaryColor3ui);
	LGLP(glSecondaryColor3uiv);
	LGLP(glSecondaryColor3us);
	LGLP(glSecondaryColor3usv);
	LGLP(glSecondaryColorPointer);
	LGLP(glWindowPos2d);
	LGLP(glWindowPos2dv);
	LGLP(glWindowPos2f);
	LGLP(glWindowPos2fv);
	LGLP(glWindowPos2i);
	LGLP(glWindowPos2iv);
	LGLP(glWindowPos2s);
	LGLP(glWindowPos2sv);
	LGLP(glWindowPos3d);
	LGLP(glWindowPos3dv);
	LGLP(glWindowPos3f);
	LGLP(glWindowPos3fv);
	LGLP(glWindowPos3i);
	LGLP(glWindowPos3iv);
	LGLP(glWindowPos3s);
	LGLP(glWindowPos3sv);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.4 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.5.
 */

void (VGLEXTAPIENTRY *glGenQueries) (GLsizei, GLuint *);
void (VGLEXTAPIENTRY *glDeleteQueries) (GLsizei, const GLuint *);
GLboolean (VGLEXTAPIENTRY *glIsQuery) (GLuint);
void (VGLEXTAPIENTRY *glBeginQuery) (GLenum, GLuint);
void (VGLEXTAPIENTRY *glEndQuery) (GLenum);
void (VGLEXTAPIENTRY *glGetQueryiv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetQueryObjectiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetQueryObjectuiv) (GLuint, GLenum, GLuint *);
void (VGLEXTAPIENTRY *glBindBuffer) (GLenum, GLuint);
void (VGLEXTAPIENTRY *glDeleteBuffers) (GLsizei, const GLuint *);
void (VGLEXTAPIENTRY *glGenBuffers) (GLsizei, GLuint *);
GLboolean (VGLEXTAPIENTRY *glIsBuffer) (GLuint);
void (VGLEXTAPIENTRY *glBufferData) (GLenum, GLsizeiptr, const GLvoid *, GLenum);
void (VGLEXTAPIENTRY *glBufferSubData) (GLenum, GLintptr, GLsizeiptr, const GLvoid *);
void (VGLEXTAPIENTRY *glGetBufferSubData) (GLenum, GLintptr, GLsizeiptr, GLvoid *);
GLvoid* (VGLEXTAPIENTRY *glMapBuffer) (GLenum, GLenum);
GLboolean (VGLEXTAPIENTRY *glUnmapBuffer) (GLenum);
void (VGLEXTAPIENTRY *glGetBufferParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetBufferPointerv) (GLenum, GLenum, GLvoid* *);

static GLboolean
vgl_load_gl_1_5(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;

#ifndef	AG_NO_GL_15_QUIERIES
	LGLP(glGenQueries);
	LGLP(glDeleteQueries);
	LGLP(glIsQuery);
	LGLP(glBeginQuery);
	LGLP(glEndQuery);
	LGLP(glGetQueryiv);
	LGLP(glGetQueryObjectiv);
	LGLP(glGetQueryObjectuiv);
	if(!flag)
	{
		TRACE("vgl_load_gl_1_5(): Warning: unable to load ARB_occlusion_query.");
		flag = GL_TRUE;
	}
#endif	/* !AG_NO_GL_15_QUIERIES */

	LGLP(glBindBuffer);
	LGLP(glDeleteBuffers);
	LGLP(glGenBuffers);
	LGLP(glIsBuffer);
	LGLP(glBufferData);
	LGLP(glBufferSubData);
	LGLP(glGetBufferSubData);
	LGLP(glMapBuffer);
	LGLP(glUnmapBuffer);
	LGLP(glGetBufferParameteriv);
	LGLP(glGetBufferPointerv);

	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.5 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 2.0.
 */

void (VGLEXTAPIENTRY *glBlendEquationSeparate) (GLenum, GLenum);
void (VGLEXTAPIENTRY *glDrawBuffers) (GLsizei, const GLenum *);
void (VGLEXTAPIENTRY *glStencilOpSeparate) (GLenum, GLenum, GLenum, GLenum);
void (VGLEXTAPIENTRY *glStencilFuncSeparate) (GLenum, GLenum, GLint, GLuint);
void (VGLEXTAPIENTRY *glStencilMaskSeparate) (GLenum, GLuint);
void (VGLEXTAPIENTRY *glAttachShader) (GLuint, GLuint);
void (VGLEXTAPIENTRY *glBindAttribLocation) (GLuint, GLuint, const GLchar *);
void (VGLEXTAPIENTRY *glCompileShader) (GLuint);
GLuint (VGLEXTAPIENTRY *glCreateProgram) (void);
GLuint (VGLEXTAPIENTRY *glCreateShader) (GLenum);
void (VGLEXTAPIENTRY *glDeleteProgram) (GLuint);
void (VGLEXTAPIENTRY *glDeleteShader) (GLuint);
void (VGLEXTAPIENTRY *glDetachShader) (GLuint, GLuint);
void (VGLEXTAPIENTRY *glDisableVertexAttribArray) (GLuint);
void (VGLEXTAPIENTRY *glEnableVertexAttribArray) (GLuint);
void (VGLEXTAPIENTRY *glGetActiveAttrib) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
void (VGLEXTAPIENTRY *glGetActiveUniform) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
void (VGLEXTAPIENTRY *glGetAttachedShaders) (GLuint, GLsizei, GLsizei *, GLuint *);
GLint (VGLEXTAPIENTRY *glGetAttribLocation) (GLuint, const GLchar *);
void (VGLEXTAPIENTRY *glGetProgramiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetProgramInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
void (VGLEXTAPIENTRY *glGetShaderiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetShaderInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
void (VGLEXTAPIENTRY *glGetShaderSource) (GLuint, GLsizei, GLsizei *, GLchar *);
GLint (VGLEXTAPIENTRY *glGetUniformLocation) (GLuint, const GLchar *);
void (VGLEXTAPIENTRY *glGetUniformfv) (GLuint, GLint, GLfloat *);
void (VGLEXTAPIENTRY *glGetUniformiv) (GLuint, GLint, GLint *);
void (VGLEXTAPIENTRY *glGetVertexAttribdv) (GLuint, GLenum, GLdouble *);
void (VGLEXTAPIENTRY *glGetVertexAttribfv) (GLuint, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetVertexAttribiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetVertexAttribPointerv) (GLuint, GLenum, GLvoid* *);
GLboolean (VGLEXTAPIENTRY *glIsProgram) (GLuint);
GLboolean (VGLEXTAPIENTRY *glIsShader) (GLuint);
void (VGLEXTAPIENTRY *glLinkProgram) (GLuint);
void (VGLEXTAPIENTRY *glShaderSource) (GLuint, GLsizei, const GLchar* *, const GLint *);
void (VGLEXTAPIENTRY *glUseProgram) (GLuint);
void (VGLEXTAPIENTRY *glUniform1f) (GLint, GLfloat);
void (VGLEXTAPIENTRY *glUniform2f) (GLint, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glUniform3f) (GLint, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glUniform4f) (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glUniform1i) (GLint, GLint);
void (VGLEXTAPIENTRY *glUniform2i) (GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glUniform3i) (GLint, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glUniform4i) (GLint, GLint, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glUniform1fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform2fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform3fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform4fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform1iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniform2iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniform3iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniform4iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniformMatrix2fv) (GLint, GLsizei, GLboolean, const GLfloat *);
void (VGLEXTAPIENTRY *glUniformMatrix3fv) (GLint, GLsizei, GLboolean, const GLfloat *);
void (VGLEXTAPIENTRY *glUniformMatrix4fv) (GLint, GLsizei, GLboolean, const GLfloat *);
void (VGLEXTAPIENTRY *glValidateProgram) (GLuint);
void (VGLEXTAPIENTRY *glVertexAttrib1d) (GLuint, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib1dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib1f) (GLuint, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib1fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib1s) (GLuint, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib1sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib2d) (GLuint, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib2dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib2f) (GLuint, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib2fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib2s) (GLuint, GLshort, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib2sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib3d) (GLuint, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib3dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib3f) (GLuint, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib3fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib3s) (GLuint, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib3sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nbv) (GLuint, const GLbyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4Niv) (GLuint, const GLint *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nsv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nub) (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
void (VGLEXTAPIENTRY *glVertexAttrib4Nubv) (GLuint, const GLubyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nuiv) (GLuint, const GLuint *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nusv) (GLuint, const GLushort *);
void (VGLEXTAPIENTRY *glVertexAttrib4bv) (GLuint, const GLbyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4d) (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib4dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib4f) (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib4fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib4iv) (GLuint, const GLint *);
void (VGLEXTAPIENTRY *glVertexAttrib4s) (GLuint, GLshort, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib4sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib4ubv) (GLuint, const GLubyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4uiv) (GLuint, const GLuint *);
void (VGLEXTAPIENTRY *glVertexAttrib4usv) (GLuint, const GLushort *);
void (VGLEXTAPIENTRY *glVertexAttribPointer) (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);

void (VGLEXTAPIENTRY *glProgramParameteriEXT)(GLuint program, GLenum pname, GLint value);
void (VGLEXTAPIENTRY *glFramebufferTextureEXT)(GLenum target, GLenum attachment, GLuint texture, GLint level);
void (VGLEXTAPIENTRY *glFramebufferTextureLayerEXT)(GLenum target, GLenum attachment, GLuint texture, GLint level, GLint layer);
void (VGLEXTAPIENTRY *glFramebufferTextureFaceEXT)(GLenum target, GLenum attachment,
GLuint texture, GLint level, GLenum face);

static GLboolean
vgl_load_gl_2_0(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE; /* make compiler happy */
#ifndef	AG_NO_GL_20_GEOMETRY_SHADER
	if(strstr((char*)glGetString(GL_EXTENSIONS), "GL_EXT_geometry_shader4"))
	{
		LGLP(glProgramParameteriEXT);
		LGLP(glFramebufferTextureEXT); 
		LGLP(glFramebufferTextureLayerEXT);
		LGLP(glFramebufferTextureFaceEXT);

		if(!flag)
			TRACE("vgl_load_gl_2_0(): Warning: unable to load GL_EXT_geometry_shader4.");
		else
			TRACE("vgl_load_gl_2_0(): GL_EXT_geometry_shader4 loaded.");
	}
	else
		TRACE("vgl_load_gl_2_0(): Warning: GL_EXT_geometry_shader4 not supported.");
#endif	/* !AG_NO_GL_20_GEOMETRY_SHADER */
	/*
	 * Ignore state of 'flag': it's Ok if geometry shaders are not supported.
	 */
	flag = GL_TRUE;
	LGLP(glBlendEquationSeparate);
	LGLP(glDrawBuffers);
	LGLP(glStencilOpSeparate);
	LGLP(glStencilFuncSeparate);
	LGLP(glStencilMaskSeparate);
	LGLP(glAttachShader);
	LGLP(glBindAttribLocation);
	LGLP(glCompileShader);
	LGLP(glCreateProgram);
	LGLP(glCreateShader);
	LGLP(glDeleteProgram);
	LGLP(glDeleteShader);
	LGLP(glDetachShader);
	LGLP(glDisableVertexAttribArray);
	LGLP(glEnableVertexAttribArray);
	LGLP(glGetActiveAttrib);
	LGLP(glGetActiveUniform);
	LGLP(glGetAttachedShaders);
	LGLP(glGetAttribLocation);
	LGLP(glGetProgramiv);
	LGLP(glGetProgramInfoLog);
	LGLP(glGetShaderiv);
	LGLP(glGetShaderInfoLog);
	LGLP(glGetShaderSource);
	LGLP(glGetUniformLocation);
	LGLP(glGetUniformfv);
	LGLP(glGetUniformiv);
	LGLP(glGetVertexAttribdv);
	LGLP(glGetVertexAttribfv);
	LGLP(glGetVertexAttribiv);
	LGLP(glGetVertexAttribPointerv);
	LGLP(glIsProgram);
	LGLP(glIsShader);
	LGLP(glLinkProgram);
	LGLP(glShaderSource);
	LGLP(glUseProgram);
	LGLP(glUniform1f);
	LGLP(glUniform2f);
	LGLP(glUniform3f);
	LGLP(glUniform4f);
	LGLP(glUniform1i);
	LGLP(glUniform2i);
	LGLP(glUniform3i);
	LGLP(glUniform4i);
	LGLP(glUniform1fv);
	LGLP(glUniform2fv);
	LGLP(glUniform3fv);
	LGLP(glUniform4fv);
	LGLP(glUniform1iv);
	LGLP(glUniform2iv);
	LGLP(glUniform3iv);
	LGLP(glUniform4iv);
	LGLP(glUniformMatrix2fv);
	LGLP(glUniformMatrix3fv);
	LGLP(glUniformMatrix4fv);
	LGLP(glValidateProgram);
	LGLP(glVertexAttrib1d);
	LGLP(glVertexAttrib1dv);
	LGLP(glVertexAttrib1f);
	LGLP(glVertexAttrib1fv);
	LGLP(glVertexAttrib1s);
	LGLP(glVertexAttrib1sv);
	LGLP(glVertexAttrib2d);
	LGLP(glVertexAttrib2dv);
	LGLP(glVertexAttrib2f);
	LGLP(glVertexAttrib2fv);
	LGLP(glVertexAttrib2s);
	LGLP(glVertexAttrib2sv);
	LGLP(glVertexAttrib3d);
	LGLP(glVertexAttrib3dv);
	LGLP(glVertexAttrib3f);
	LGLP(glVertexAttrib3fv);
	LGLP(glVertexAttrib3s);
	LGLP(glVertexAttrib3sv);
	LGLP(glVertexAttrib4Nbv);
	LGLP(glVertexAttrib4Niv);
	LGLP(glVertexAttrib4Nsv);
	LGLP(glVertexAttrib4Nub);
	LGLP(glVertexAttrib4Nubv);
	LGLP(glVertexAttrib4Nuiv);
	LGLP(glVertexAttrib4Nusv);
	LGLP(glVertexAttrib4bv);
	LGLP(glVertexAttrib4d);
	LGLP(glVertexAttrib4dv);
	LGLP(glVertexAttrib4f);
	LGLP(glVertexAttrib4fv);
	LGLP(glVertexAttrib4iv);
	LGLP(glVertexAttrib4s);
	LGLP(glVertexAttrib4sv);
	LGLP(glVertexAttrib4ubv);
	LGLP(glVertexAttrib4uiv);
	LGLP(glVertexAttrib4usv);
	LGLP(glVertexAttribPointer);
	
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 2.0 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.2 and above stuff.
 */

static GLboolean
ag_load_gl(GLuint required_version)
{
	GLboolean pass = GL_TRUE, terminate = GL_FALSE;
	if(required_version > AG_REQUIRED_1_1)
		pass = (GLboolean)(pass && vgl_load_gl_1_2(&terminate));
	else
		return pass;
	if(terminate)
		goto err_drv;
	if(required_version > AG_REQUIRED_1_2)
		pass = (GLboolean)(pass && vgl_load_gl_1_3(&terminate));
	else
		return pass;
	if(terminate)
		goto err_drv;
	if(required_version > AG_REQUIRED_1_3)
		pass = (GLboolean)(pass && vgl_load_gl_1_4(&terminate));
	else
		return pass;
	if(terminate)
		goto err_drv;
	if(required_version > AG_REQUIRED_1_4)
		pass = (GLboolean)(pass && vgl_load_gl_1_5(&terminate));
	else
		return pass;
	if(terminate)
		goto err_drv;
	if(required_version > AG_REQUIRED_1_5)
		pass = (GLboolean)(pass && vgl_load_gl_2_0(&terminate));
	else
		return pass;
	if(terminate)
		goto err_drv;
	if(required_version > AG_REQUIRED_2_0)
	{
		TRACE("Requested GL version still unsupported by AntiGlut.");
		return GL_FALSE;
	}
	return pass;
err_drv:;
	TRACE("Unable to switch to another GL implementation after first init.");
	return GL_FALSE;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Event queue management.
 */

void
uiAutoRepeat(GLboolean enable)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	ctx->wnd->auto_repeat_disabled =
		(GLboolean)(enable ? GL_FALSE : GL_TRUE);
}

void
uiAutoWarp(GLboolean enable)
{
	gl_context_p ctx;
	arch_window_p aw;
	IF_NO_GLCONTEXT(ctx)
		return;
	aw = ctx->wnd;
	if(enable)
	{
		int cx, cy;
		if(aw->auto_warp_enabled)
			return;
		aw->auto_warp_enabled = GL_TRUE;
		aw->auto_warp_saved_x = aw->last_mx;
		aw->auto_warp_saved_y = aw->last_my;
		cx = aw->width >> 1;
		cy = aw->height >> 1;
		aw->last_vmx = 0;
		aw->last_vmy = 0;
		aw->last_mx = cx;
		aw->last_my = cy;
		archWinShowPointer(aw, GL_FALSE);
		if(archWinWarpPointer(aw, cx, cy))
		{
			archWinGrabMouse(aw, GL_TRUE);
			return;
		}
		aw->auto_warp_enabled = GL_FALSE;
		aw->last_mx = aw->last_vmx = aw->auto_warp_saved_x;
		aw->last_my = aw->last_vmy = aw->auto_warp_saved_y;
		archWinShowPointer(aw, GL_TRUE);
	}
	if(!aw->auto_warp_enabled)
		return;
	aw->auto_warp_enabled = GL_FALSE;
	aw->last_mx = aw->last_vmx = aw->auto_warp_saved_x;
	aw->last_my = aw->last_vmy = aw->auto_warp_saved_y;
	archWinGrabMouse(aw, GL_FALSE);
	archWinWarpPointer(aw, aw->auto_warp_saved_x, aw->auto_warp_saved_y);
	archWinShowPointer(aw, GL_TRUE);
}

GLboolean
uiPending()
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	archWinPoll(ctx->wnd,GL_FALSE);
	return (GLboolean)(ctx->wnd->msg_que ? GL_TRUE : GL_FALSE);
}

GLboolean
uiWait()
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	archWinPoll(ctx->wnd, GL_TRUE);
	return GL_TRUE;
}

ui_event_p
uiPeekEvent(unsigned offset_from_first_push)
{
	ui_event_p ptr = NULL; /* make GCC happy */
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return NULL;
	if(!eventDeqPeekBack(ctx->wnd->msg_que,offset_from_first_push,&ptr))
		return NULL;
	return ptr;
}

ui_event_p
uiPeekLastEvent(unsigned offset_from_last_push)
{
	ui_event_p ptr = NULL; /* make GCC happy */
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return NULL;
	if(!eventDeqPeek(ctx->wnd->msg_que,offset_from_last_push,&ptr))
		return NULL;
	return ptr;
}

GLboolean
uiRemoveEvent(unsigned offset_from_first_push,ui_event_p ev_return)
{
	ui_event_p ptr = NULL; /* make GCC happy */
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	if(!eventDeqPeekBack(ctx->wnd->msg_que,offset_from_first_push,&ptr))
		return GL_FALSE;
	if(ev_return)
		memcpy(ev_return,ptr,sizeof(ui_event_t));
	ctx->wnd->msg_que = eventDeqRemoveBack(ctx->wnd->msg_que,
		offset_from_first_push);
	return GL_TRUE;
}

GLboolean
uiRemoveLastEvent(unsigned offset_from_last_push,ui_event_p ev_return)
{
	ui_event_p ptr = NULL; /* make GCC happy */
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	if(!eventDeqPeek(ctx->wnd->msg_que,offset_from_last_push,&ptr))
		return GL_FALSE;
	if(ev_return)
		memcpy(ev_return,ptr,sizeof(ui_event_t));
	ctx->wnd->msg_que =
		eventDeqRemove(ctx->wnd->msg_que,offset_from_last_push);
	return GL_TRUE;
}

GLboolean
uiNextEvent(ui_event_p ev_return)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	archWinPoll(ctx->wnd,GL_TRUE);
	ctx->wnd->msg_que = eventDeqPopBack(ctx->wnd->msg_que,ev_return);
	return GL_TRUE;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Initialization and shutting down.
 */

GLboolean
agWindowConfig(char *title,GLuint width,GLuint height,GLboolean fullscreen)
{
	struct thread_window_config_s *wc;
	if(*get_thread_context_addr())
		return GL_FALSE;
	wc = get_thread_window_config_addr();
	if(title)
	{
		strncpy(wc->title,title,512);
	}
	wc->title[511] = 0;
	wc->width = width;
	wc->height = height;
	wc->fullscreen = fullscreen;
	wc->left = wc->top = 0;
	wc->resizeable = GL_FALSE;
	return GL_TRUE;
}

GLboolean
agWindowedConfig(char *title,GLuint width,GLuint height,
	GLuint left,GLuint top)
{
	struct thread_window_config_s *wc;
	if(*get_thread_context_addr())
		return GL_FALSE;
	wc = get_thread_window_config_addr();
	if(title)
	{
		strncpy(wc->title,title,512);
	}
	wc->title[511] = 0;
	wc->width = width;
	wc->height = height;
	wc->fullscreen = GL_FALSE;
	wc->left = left;
	wc->top = top;
	wc->resizeable = GL_TRUE;
	return GL_TRUE;
}

GLboolean
agInit(GLuint required_version,GLboolean need_depth,
	GLboolean need_stencil,GLboolean need_dst_alpha)
{
	struct thread_window_config_s *wc;
	if(*get_thread_context_addr())
		agShutdown();
	wc = get_thread_window_config_addr();
	*get_thread_context_addr() = vglInit(
		wc->title,
		wc->left,wc->top,wc->width,wc->height,
		wc->fullscreen,
		wc->resizeable,
		need_depth,need_stencil,need_dst_alpha);
	if(!*get_thread_context_addr())
		return GL_FALSE;
	if(!ag_load_gl(required_version))
	{
		agShutdown();
		return GL_FALSE;
	}
	return GL_TRUE;
}

GLboolean
agGetWindowSize(GLuint *width,GLuint *height)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	if(width)
		*width = ctx->wnd->width;
	if(height)
		*height = ctx->wnd->height;
	return GL_TRUE;
}

GLboolean
agSavePPM(const char *outname)
{
	unsigned int r;
	unsigned char *buffer, *row;
	FILE *out;
	GLuint width = get_thread_window_config_addr()->width;
	GLuint height = get_thread_window_config_addr()->height;
	buffer = (unsigned char*)malloc(3 * width * height);
	if(!buffer)
		return GL_FALSE;
	glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, buffer);
	if(glGetError() != GL_NO_ERROR)
	{
		free(buffer);
		fprintf(stderr, "agSavePPM(): glReadPixels() failed.\n");
		return GL_FALSE;
	}
	out = fopen(outname, "wb");
	if(!out)
	{
		fprintf(stderr, "agSavePPM(): unable to open \"%s\".\n", outname);
		free(buffer);
		return GL_FALSE;
	}
	fprintf(out, "P6\n");
	fprintf(out, "%u %u\n255\n", width, height);
	/* the image was stored bottom - to - top */
	row = buffer + 3 * width * height;
	for(r = 0;r < height;++r, row -= 3 * width)
		fwrite(row, 3, width, out);
	fclose(out);
	free(buffer);
	return GL_TRUE;
}
